/*
 * Copyright (c) 1994 The University of Utah and
 * the Center for Software Science (CSS).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSS ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSS DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSS requests users of this software to return to css-dist@cs.utah.edu any
 * improvements that they make and grant CSS redistribution rights.
 *
 *      Author: Bryan Ford, University of Utah CSS
 */
#ifndef _MACH_IPC_H_
#define _MACH_IPC_H_

#include <mach/kern_return.h>
#include <mach/machine/ipc.h>
#include <mach/machine/vm_types.h>

typedef integer_t ipc_value_t;
typedef integer_t ipc_key_t;

typedef struct ipc_block
{
	void *data;
	vm_size_t size;
} ipc_block_t;

typedef struct ipc_msg
{
	/* Request/return code */
	ipc_value_t code;

	/* Contents of the message.  */
	ipc_value_t values[IPC_MAX_VALUES];
	ipc_block_t blocks[IPC_MAX_BLOCKS];
	ipc_key_t keys[IPC_MAX_KEYS];

} ipc_msg_t;


typedef kern_return_t ipc_return_t;

/* IPC error codes - XXX */
#define IPC_NULL		-1000	/* called a null key */
#define IPC_BAD_ID		-1001	/* bad request message ID */


/* Machine-independent IPC invocation routines.
   These should generally be inlined for performance.  */

ipc_return_t ipc_call(ipc_key_t invoke_key, ipc_msg_t *send, ipc_msg_t *recv);
ipc_return_t ipc_return(ipc_key_t invoke_key, ipc_msg_t *send, ipc_msg_t *recv);
ipc_return_t ipc_fork(ipc_key_t invoke_key, ipc_msg_t *send);


#endif /* _MACH_IPC_H_ */
