/* 
 * Copyright (c) 1988-1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *	Utah $Hdr: grf.c 1.41 94/12/14$
 */

/*
 * Graphics display driver for HP 300/400/700/800 machines.
 * This is the hardware-independent portion of the driver.
 * Hardware access is through the machine dependent grf switch routines.
 */

#include "grf.h"
#if NGRF > 0

#ifdef MACH_KERNEL
#define	HPUXCOMPAT		/* XXX should go elsewhere */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <kern/current.h>
#include <device/tty.h>
#include <device/errno.h>
#include <device/io_req.h>
#include <machine/machspl.h>
#include <vm/vm_kern.h>
#include <vm/pmap.h>
#ifdef hp700
#include <mach_glue.h>
#endif
#include <grfioctl.h>
#include <grfvar.h>
#else
#include <sys/param.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/mmap.h>
#include <sys/malloc.h>
#include <hpdev/grfioctl.h>
#include <hpdev/grfvar.h>
#endif

#include <machine/cpu.h>

#ifdef	MACH_KERNEL
#define current_pid()		current_thread()
#define	ishpuxproc(p)		(1)
#else
#define	current_pid()		u.u_procp->p_pid
#define	ishpuxproc(p)		((p)->p_flag & SHPUX)
#endif

#include "ite.h"
#if NITE == 0
#define	iteon(u,f)
#define	iteoff(u,f)
#endif

struct grf_softc grf_softc[NGRF];

#ifdef MAPMEM
int grfexit();
struct mapmemops grfops = { (int (*)())0, (int (*)())0, grfexit, grfexit };
#ifdef HPUXCOMPAT
struct mapmemops grflckops = { (int (*)())0, (int (*)())0, grfexit, grfexit };
#ifdef hp300
struct mapmemops grfiomops = { (int (*)())0, (int (*)())0, grfexit, grfexit };
#endif
#endif
#endif

#ifdef DEBUG
int grfdebug = 0;
#define GDB_DEVNO	0x01
#define GDB_MMAP	0x02
#define GDB_IOMAP	0x04
#define GDB_LOCK	0x08
#endif

/*ARGSUSED*/
grfopen(dev, flags)
	dev_t dev;
{
	int unit = GRFUNIT(dev);
	register struct grf_softc *gp = &grf_softc[unit];
	int error = 0;

	if (unit >= NGRF || (gp->g_flags & GF_ALIVE) == 0)
		return(ENXIO);
	if ((gp->g_flags & (GF_OPEN|GF_EXCLUDE)) == (GF_OPEN|GF_EXCLUDE))
		return(EBUSY);
#ifdef HPUXCOMPAT
	/*
	 * XXX: cannot handle both HPUX and BSD processes at the same time
	 */
	if (ishpuxproc(u.u_procp))
		if (gp->g_flags & GF_BSDOPEN)
			return(EBUSY);
		else
			gp->g_flags |= GF_HPUXOPEN;
	else
		if (gp->g_flags & GF_HPUXOPEN)
			return(EBUSY);
		else
			gp->g_flags |= GF_BSDOPEN;
#endif
	/*
	 * First open.
	 * XXX: always put in graphics mode.
	 */
	error = 0;
	if ((gp->g_flags & GF_OPEN) == 0) {
		gp->g_flags |= GF_OPEN;
		error = grfon(dev);
	}
	return(error);
}

/*ARGSUSED*/
grfclose(dev, flags)
	dev_t dev;
{
	register struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];

	(void) grfoff(dev);
#ifdef HPUXCOMPAT
	(void) grfunlock(gp);
#endif
	gp->g_flags &= GF_ALIVE;
	return(0);
}

#ifdef	MACH_KERNEL
io_return_t
grfgetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	int error;

	error = grfioctl(dev, flavor, data, 0);
	*count = ((flavor >> 16) & 0x7F) / sizeof(int);
	return(error);
}

io_return_t
grfsetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;
	unsigned int	count;
{
	return(grfioctl(dev, flavor, data, 0));
}
#endif

/*ARGSUSED*/
grfioctl(dev, cmd, data, flag)
	dev_t dev;
	caddr_t data;
{
	register struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	int error;

#ifdef HPUXCOMPAT
	if (ishpuxproc(u.u_procp))
		return(hpuxgrfioctl(dev, cmd, data, flag));
#endif
	error = 0;
	switch (cmd) {

	case GRFIOCGINFO:
		bcopy((caddr_t)&gp->g_display, data, sizeof(struct grfinfo));
		break;

	case GRFIOCON:
		error = grfon(dev);
		break;

	case GRFIOCOFF:
		error = grfoff(dev);
		break;

#ifdef MAPMEM
	case GRFIOCMAP:
		error = grfmmap(dev, (caddr_t *)data);
		break;

	case GRFIOCUNMAP:
		error = grfunmmap(dev, *(caddr_t *)data);
		break;
#endif

	default:
		error = EINVAL;
		break;

	}
	return(error);
}

/*ARGSUSED*/
grfselect(dev, rw)
	dev_t dev;
{
	if (rw == FREAD)
		return(0);
	return(1);
}

/*ARGSUSED*/
grfmap(dev, off, prot)
	dev_t dev;
{
	return(grfaddr(&grf_softc[GRFUNIT(dev)], off));
}

grfon(dev)
	dev_t dev;
{
	int unit = GRFUNIT(dev);
	struct grf_softc *gp = &grf_softc[unit];

	/*
	 * XXX: iteoff call relies on devices being in same order
	 * as ITEs and the fact that iteoff only uses the minor part
	 * of the dev arg.
	 */
	iteoff(unit, 3);
	return((*gp->g_sw->gd_mode)(gp,
				    (dev&GRFOVDEV) ? GM_GRFOVON : GM_GRFON,
				    (caddr_t)0));
}

grfoff(dev)
	dev_t dev;
{
	int unit = GRFUNIT(dev);
	struct grf_softc *gp = &grf_softc[unit];
	int error;

#ifdef MAPMEM
	(void) grfunmmap(dev, (caddr_t)0);
#endif
	error = (*gp->g_sw->gd_mode)(gp,
				     (dev&GRFOVDEV) ? GM_GRFOVOFF : GM_GRFOFF,
				     (caddr_t)0);
	/* XXX: see comment for iteoff above */
	iteon(unit, 2);
	return(error);
}

grfaddr(gp, off)
	struct grf_softc *gp;
	register int off;
{
#ifdef MAPMEM
#ifdef hp300
	register struct grfinfo *gi = &gp->g_display;

	/* control registers */
	if (off >= 0 && off < gi->gd_regsize)
		return(((u_int)gi->gd_regaddr + off) >> PGSHIFT);

	/* frame buffer */
	if (off >= gi->gd_regsize && off < gi->gd_regsize+gi->gd_fbsize) {
		off -= gi->gd_regsize;
		return(((u_int)gi->gd_fbaddr + off) >> PGSHIFT);
	}
#else
	return(0);
#endif
#endif
	/* bogus */
	return(-1);
}

grfmmap(dev, addrp)
	dev_t dev;
	caddr_t *addrp;
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	int len;
#ifdef MAPMEM
	register struct mapmem *mp;
	int flags, grfmapin();
#endif

#ifdef DEBUG
	if (grfdebug & GDB_MMAP)
		printf("grfmmap(%d): addr %x\n", current_pid(), *addrp);
#endif
#ifdef hp300
	len = gp->g_display.gd_regsize + gp->g_display.gd_fbsize;
#else
#ifdef hp700
	*addrp = (caddr_t)((int)gp->g_display.gd_regaddr & 0xFC000000);
	len = 0x2000000;
#else
	*addrp = gp->g_display.gd_regaddr + 0x8000;
	len = (gp->g_display.gd_fbaddr - *addrp) + 0x800000;
#endif
#endif
#ifdef MAPMEM
#ifdef hp300
	flags = MM_RW|MM_CI|MM_NOCORE;
#else
	flags = MM_RW|MM_IO|MM_NOCORE;
#endif
	mp = mmalloc(minor(dev), addrp, len, flags, &grfops);
	if (mp == MMNIL)
		return(u.u_error);
	if (!mmmapin(mp, grfmapin)) {
		mmfree(mp);
		return(u.u_error);
	}
	if ((*gp->g_sw->gd_mode)(gp, GM_MAP, *addrp)) {
		mmmapout(mp);
		mmfree(mp);
		u.u_error = EINVAL;
		return(u.u_error);
	}
#endif
#ifdef	MACH_KERNEL
	if (!pmap_iomap((vm_offset_t)*addrp))
		return(EINVAL);
#endif
	return(0);
}

grfunmmap(dev, addr)
	dev_t dev;
	caddr_t addr;
{
#ifdef MAPMEM
	register struct mapmem *mp, **mpp;
	int found, unit = minor(dev);
#endif
#ifdef	MACH_KERNEL
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
#endif

#ifdef DEBUG
	if (grfdebug & GDB_MMAP)
		printf("grfunmmap(%d): id %d addr %x\n",
		       current_pid(), minor(dev), addr);
#endif
#ifdef MAPMEM
#ifndef hp300
	/* GCUNMAP has no argument on the PA */
	addr = 0;
#endif
	found = 0;
	mpp = &u.u_mmap;
	for (mp = *mpp; mp; mp = *mpp) {
		if (mp->mm_ops != &grfops || mp->mm_id != unit) {
			mpp = &mp->mm_next;
			continue;
		}
		if (addr &&
		    (addr < mp->mm_uva || addr >= mp->mm_uva+mp->mm_size)) {
			mpp = &mp->mm_next;
			continue;
		}
		grfexit(mp);
		found++;
	}
	if (found == 0)
		return(EINVAL);
#endif
#ifdef	MACH_KERNEL
#ifdef hp700
	addr = (caddr_t)((int)gp->g_display.gd_regaddr & 0xFC000000);
#else
	addr = gp->g_display.gd_regaddr + 0x8000;
#endif
	pmap_iounmap((vm_offset_t)addr);
#endif
	return(0);
}

/*
 * HP-UX compatibility routines
 */
#ifdef HPUXCOMPAT

#ifndef MACH_KERNEL
#include <hpux/hpux.h>
#endif

/*ARGSUSED*/
hpuxgrfioctl(dev, cmd, data, flag)
	dev_t dev;
	caddr_t data;
{
	register struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	int error;

	error = 0;
	switch (cmd) {

	case GCID:
		*(int *)data = gp->g_display.gd_id;
		break;

	case GCON:
		error = grfon(dev);
		break;

	case GCOFF:
		error = grfoff(dev);
		break;

	case GCLOCK:
	case GCLOCK_MINIMUM:
#ifndef hp300
	case GCFASTLOCK:
#endif
		error = grflock(gp, 1);
		break;

	case GCUNLOCK:
	case GCUNLOCK_MINIMUM:
		error = grfunlock(gp);
		break;

	case GCAON:
	case GCAOFF:
		break;

	/* GCSTATIC is implied by our implementation */
	case GCSTATIC_CMAP:
	case GCVARIABLE_CMAP:
		break;

	/* map in control regs and frame buffer */
	case GCMAP:
		error = grfmmap(dev, (caddr_t *)data);
		break;

	case GCUNMAP:
		error = grfunmmap(dev, *(caddr_t *)data);
#ifdef MAPMEM
		/* XXX: HP-UX uses GCUNMAP to get rid of GCSLOT memory */
		if (error)
			error = grflckunmmap(dev, *(caddr_t *)data);
#endif
		break;

#ifdef MAPMEM
	case GCSLOT:
	{
		struct grf_slot *sp = (struct grf_slot *)data;

		sp->slot = grffindpid(gp);
		if (sp->slot)
			error = grflckmmap(dev, (caddr_t *)&sp->addr);
		else
			error = EINVAL;		/* XXX */
		break;
	}

#ifdef hp300
	/*
	 * XXX: only used right now to map in rbox control registers
	 * Will be replaced in the future with a real IOMAP interface.
	 */
	case IOMAPMAP:
		error = iommap(dev, (caddr_t *)data);
		if (!error)
			u.u_r.r_val1 = *(int *)data;	/* XXX: this sux */
		break;

	case IOMAPUNMAP:
		error = iounmmap(dev, *(caddr_t *)data);
		break;
#endif
#endif

	case GCDESCRIBE:
		error = (*gp->g_sw->gd_mode)(gp, GM_DESCRIBE, data);
		break;

	default:
		error = EINVAL;
		break;
	}
	return(error);
}

grflock(gp, block)
	register struct grf_softc *gp;
	int block;
{
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grflock(%d): dev %x flags %x\n",
		       current_pid(), gp-grf_softc, gp->g_flags);
#endif
#ifdef	MACH_KERNEL
	/* XXX implement me XXX */
#else
#ifdef MAPMEM
	if (gp->g_pid) {
#ifndef hp300
		/* XXX make sure kernel address is valid */
		mmvalidate((caddr_t)gp->g_lock, 1);
#endif
#ifdef DEBUG
		if (grfdebug & GDB_LOCK)
			printf(" lockpslot %d lockslot %d lock[lockslot] %d\n",
			       gp->g_lock->gl_lockslot, gp->g_lockpslot,
			       gp->g_lock->gl_locks[gp->g_lockpslot]);
#endif
		gp->g_lock->gl_lockslot = 0;
		if (gp->g_lock->gl_locks[gp->g_lockpslot] == 0) {
			gp->g_lockp = NULL;
			gp->g_lockpslot = 0;
		}
	}
#endif
	if (gp->g_lockp) {
		if (gp->g_lockp == u.u_procp)
			return(EBUSY);
		if (!block)
			return(OEAGAIN);
		do {
			gp->g_flags |= GF_WANTED;
			sleep((caddr_t)&gp->g_flags, PZERO+1);
		} while (gp->g_lockp);
#ifdef MAPMEM
#ifndef hp300
		/* XXX make sure kernel address is still valid */
		if (gp->g_pid)
			mmvalidate((caddr_t)gp->g_lock, 1);
#endif
#endif
	}
	gp->g_lockp = u.u_procp;
#ifdef MAPMEM
	if (gp->g_pid) {
		int slot = grffindpid(gp);

#ifdef DEBUG
		if (grfdebug & GDB_LOCK)
			printf("  slot %d\n", slot);
#endif
		gp->g_lockpslot = gp->g_lock->gl_lockslot = slot;
		gp->g_lock->gl_locks[slot] = 1;
	}
#endif
#endif
	return(0);
}

grfunlock(gp)
	register struct grf_softc *gp;
{
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grfunlock(%d): dev %x flags %x\n",
		       current_pid(), gp-grf_softc, gp->g_flags);
#endif
#ifdef	MACH_KERNEL
	/* XXX implement me XXX */
#else
	if (gp->g_lockp != u.u_procp)
		return(EBUSY);
#ifdef MAPMEM
	if (gp->g_pid) {
#ifndef hp300
		/* XXX make sure kernel address is valid */
		mmvalidate((caddr_t)gp->g_lock, 1);
#endif
#ifdef DEBUG
		if (grfdebug & GDB_LOCK)
			printf(" lockpslot %d lockslot %d lock[lockslot] %d\n",
			       gp->g_lock->gl_lockslot, gp->g_lockpslot,
			       gp->g_lock->gl_locks[gp->g_lockpslot]);
#endif
		gp->g_lock->gl_locks[gp->g_lockpslot] = 0;
		gp->g_lockpslot = gp->g_lock->gl_lockslot = 0;
	}
#endif
	if (gp->g_flags & GF_WANTED) {
		wakeup((caddr_t)&gp->g_flags); 
		gp->g_flags &= ~GF_WANTED;
	}
	gp->g_lockp = NULL;
#endif
	return(0);
}

#ifdef hp300
#include <hpdev/grfreg.h>
#endif

/*
 * Convert a BSD style minor devno to HPUX style.
 * We cannot just create HPUX style nodes as they require 24 bits
 * of minor device number and we only have 8.
 * XXX: This may give the wrong result for remote stats of other
 * machines where device 10 exists.
 */
grfdevno(dev)
	dev_t dev;
{
	int unit = GRFUNIT(dev);
	struct grf_softc *gp = &grf_softc[unit];
	int newdev;

	if (unit >= NGRF || (gp->g_flags&GF_ALIVE) == 0)
#ifdef MACH_KERNEL
		return (major(dev) << 24) | minor(dev);
#else
		return(bsdtohpuxdev(dev));
#endif
#ifdef hp300
	/* magic major number */
	newdev = 12 << 24;
	/* now construct minor number */
	if (gp->g_display.gd_regaddr != (caddr_t)GRFIADDR) {
		int sc = patosc(gp->g_display.gd_regaddr);
		newdev |= (sc << 16) | 0x200;
	}
	if (dev & GRFIMDEV)
		newdev |= 0x02;
	else if (dev & GRFOVDEV)
		newdev |= 0x01;
#else
#ifdef hp700
	newdev = (12 << 24) | 0x100000 | (unit << 8);
#else
	newdev = (14 << 24) | (unit << 8);
#endif
#endif
#ifdef DEBUG
	if (grfdebug & GDB_DEVNO)
		printf("grfdevno: dev %x newdev %x\n", dev, newdev);
#endif
	return(newdev);
}

#endif	/* HPUXCOMPAT */

/*
 * Mapped memory interface routines
 */
#ifdef MAPMEM

grfmapin(mp, off)
	struct mapmem *mp;
{
	return(grfaddr(&grf_softc[GRFUNIT(mp->mm_id)], off));
}

grfexit(mp)
	struct mapmem *mp;
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(mp->mm_id)];

#ifdef DEBUG
	if (grfdebug & GDB_MMAP)
		printf("grfexit(%d): id %d %x@%x\n",
		       current_pid(), mp->mm_id, mp->mm_size, mp->mm_uva);
#endif
#ifdef HPUXCOMPAT
	(void) grfunlock(gp);
	grfrmpid(gp);
#endif
	(void) (*gp->g_sw->gd_mode)(gp, GM_UNMAP, 0);
	mmmapout(mp);
	mmfree(mp);
}

#ifdef HPUXCOMPAT
#ifdef hp300
iommap(dev, addrp)
	dev_t dev;
	caddr_t *addrp;
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	register struct mapmem *mp;
	int len, grfmapin();

#ifdef DEBUG
	if (grfdebug & (GDB_MMAP|GDB_IOMAP))
		printf("iommap(%d): addr %x\n", current_pid(), *addrp);
#endif
	len = gp->g_display.gd_regsize;
	mp = mmalloc(minor(dev), addrp, len, MM_RW|MM_CI|MM_NOCORE, &grfiomops);
	if (mp == MMNIL)
		return(u.u_error);
	if (!mmmapin(mp, grfmapin)) {
		mmfree(mp);
		return(u.u_error);
	}
	return(0);
}

iounmmap(dev, addr)
	dev_t dev;
	caddr_t addr;
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	register struct mapmem *mp, **mpp;
	int found, len, unit = minor(dev);

#ifdef DEBUG
	if (grfdebug & (GDB_MMAP|GDB_IOMAP))
		printf("iounmmap(%d): id %d addr %x\n",
		       current_pid(), unit, addr);
#endif
	found = 0;
	len = gp->g_display.gd_regsize;
	mpp = &u.u_mmap;
	for (mp = *mpp; mp; mp = *mpp) {
		if (mp->mm_ops != &grfiomops || mp->mm_id != unit) {
			mpp = &mp->mm_next;
			continue;
		}
		if (addr &&
		    (addr < mp->mm_uva || addr >= mp->mm_uva+mp->mm_size ||
		    len != mp->mm_size)) {
			mpp = &mp->mm_next;
			continue;
		}
		grfexit(mp);
		found++;
	}
	return(found ? 0 : EINVAL);
}
#endif

/*
 * Processes involved in framebuffer mapping via GCSLOT are recorded in
 * an array of pids.  The first element is used to record the last slot used
 * (for faster lookups).  The remaining elements record up to GRFMAXLCK-1
 * process ids.  Returns a slot number between 1 and GRFMAXLCK or 0 if no
 * slot is available. 
 */
grffindpid(gp)
	struct grf_softc *gp;
{
	register short pid, *sp;
	register int i, limit;
	int ni;

	if (gp->g_pid == NULL) {
		gp->g_pid = (short *)
			malloc(GRFMAXLCK * sizeof(short), M_DEVBUF, M_WAITOK);
		bzero((caddr_t)gp->g_pid, GRFMAXLCK * sizeof(short));
	}
	pid = current_pid();
	ni = limit = gp->g_pid[0];
	for (i = 1, sp = &gp->g_pid[1]; i <= limit; i++, sp++) {
		if (*sp == pid)
			goto done;
		if (*sp == 0)
			ni = i;
	}
	i = ni;
	if (i < limit) {
		gp->g_pid[i] = pid;
		goto done;
	}
	if (++i == GRFMAXLCK)
		return(0);
	gp->g_pid[0] = i;
	gp->g_pid[i] = pid;
done:
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grffindpid(%d): slot %d of %d\n",
		       pid, i, gp->g_pid[0]);
#endif
	return(i);
}

grfrmpid(gp)
	struct grf_softc *gp;
{
	register short pid, *sp;
	register int limit, i;
	int mi;

	if (gp->g_pid == NULL || (limit = gp->g_pid[0]) == 0)
		return;
	pid = current_pid();
	limit = gp->g_pid[0];
	mi = 0;
	for (i = 1, sp = &gp->g_pid[1]; i <= limit; i++, sp++) {
		if (*sp == pid)
			*sp = 0;
		else if (*sp)
			mi = i;
	}
	i = mi;
	if (i < limit)
		gp->g_pid[0] = i;
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grfrmpid(%d): slot %d of %d\n",
		       pid, sp-gp->g_pid, gp->g_pid[0]);
#endif
}

#ifndef hp300
#define kvtop(va)	ltor(0, va)
#define cialloc(sz)	malloc(sz, M_DEVBUF, M_WAITOK)
#endif

/*ARGSUSED*/
grflckmapin(mp, off)
	struct mapmem *mp;
{
	u_int pa = kvtop((u_int)grf_softc[GRFUNIT(mp->mm_id)].g_lock);

#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grflckmapin(%d): va %x pa %x\n", current_pid(),
		       grf_softc[GRFUNIT(mp->mm_id)].g_lock, pa);
#endif
	return(pa >> PGSHIFT);
}

grflckmmap(dev, addrp)
	dev_t dev;
	caddr_t *addrp;
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	register struct mapmem *mp;
	int grflckmapin();

#ifdef DEBUG
	if (grfdebug & (GDB_MMAP|GDB_LOCK))
		printf("grflckmmap(%d): addr %x\n", current_pid(), *addrp);
#endif
	if (gp->g_lock == NULL) {
		gp->g_lock = (struct grf_lockpage *) cialloc(NBPG);
		if (gp->g_lock == NULL)
			return(ENOMEM);
	}
#ifndef hp300
	else
		/* XXX make sure kernel address is valid */
		mmvalidate((caddr_t)gp->g_lock, 1);
#endif
	mp = mmalloc(minor(dev), addrp, NBPG, MM_RW|MM_CI, &grflckops);
	if (mp == MMNIL)
		return(u.u_error);
	if (!mmmapin(mp, grflckmapin)) {
		mmfree(mp);
		return(u.u_error);
	}
	return(0);
}

grflckunmmap(dev, addr)
	dev_t dev;
	caddr_t addr;
{
	register struct mapmem *mp;
	int unit = minor(dev);

#ifdef DEBUG
	if (grfdebug & (GDB_MMAP|GDB_LOCK))
		printf("grflckunmmap(%d): id %d addr %x\n",
		       current_pid(), unit, addr);
#endif
	for (mp = u.u_mmap; mp; mp = mp->mm_next)
		if (mp->mm_ops == &grflckops && mp->mm_id == unit &&
		    mp->mm_uva == addr) {
			grfexit(mp);
			return(0);
		}
	return(EINVAL);
}

#endif	/* HPUXCOMPAT */
#endif	/* MAPMEM */
#endif	/* NGRF > 0 */
