/*
 * Support for Intel 82596 device (run in 586 compatibility mode)
 * on the HP 700 series machines.
 *
 * Derived from the Mach Intel 82586 driver for the PC.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: if_i596var.h 3.12 94/11/10$
 */
/* 
 *	Olivetti PC586 Mach Ethernet driver v1.0
 *	Copyright Ing. C. Olivetti & C. S.p.A. 1988, 1989
 *	All rights reserved.
 *
 */ 
/*
  Copyright 1988, 1989 by Olivetti Advanced Technology Center, Inc.,
Cupertino, California.

		All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appears in all
copies and that both the copyright notice and this permission notice
appear in supporting documentation, and that the name of Olivetti
not be used in advertising or publicity pertaining to distribution
of the software without specific, written prior permission.

  OLIVETTI DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
IN NO EVENT SHALL OLIVETTI BE LIABLE FOR ANY SPECIAL, INDIRECT, OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUR OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  Copyright 1988, 1989 by Intel Corporation, Santa Clara, California.

		All Rights Reserved

Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appears in all
copies and that both the copyright notice and this permission notice
appear in supporting documentation, and that the name of Intel
not be used in advertising or publicity pertaining to distribution
of the software without specific, written prior permission.

INTEL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
IN NO EVENT SHALL INTEL BE LIABLE FOR ANY SPECIAL, INDIRECT, OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifdef MACH_KERNEL
#include <if_i596reg.h>		/* chip/board specific defines	*/
#else
#include <hpsgc/if_i596reg.h>	/* chip/board specific defines	*/
#endif

#define	STATUS_TRIES	15000
#define	ETHER_ADD_SIZE	6	/* size of a MAC address 		*/
#define ETHER_PCK_SIZE	1500	/* maximum size of an ethernet packet 	*/

/*
 * 	Board Specific Defines:
 */

#define LAN_PROM_ADDR	0xF0810000
#define OFFSET_PROM	0x0000

/*
 * Make sure everything is cache aligned (32 bytes).
 * The original offsets apparently had to fit in 16k hence the determination
 * of N_FD, N_RBD, N_TBD.  Do we have the same constraints?  I don't know,
 * but we stay within the 16k to be safe.
 *
 *	RU:   23 at 22 (32) bytes			[ 0x0000 - 0x02DF ]
 *	RBD:  23 at 10 (32) + 23 at 540 (544) bytes	[ 0x02E0 - 0x369F ]
 *	CU:   256 bytes					[ 0x36A0 - 0x379F ]
 *	TBD:  13 at 8 (32) bytes			[ 0x37A0 - 0x393F ]
 *	TBUF: 1594 (1600) bytes				[ 0x3940 - 0x3F7F ]
 *	SCB:  16 (64) bytes [1]				[ 0x3F80 - 0x3FBF ]
 *	ISCP: 8 (32) bytes				[ 0x3FC0 - 0x3FDF ]
 *	SCP:  10 (32) bytes				[ 0x3FE0 - 0x3FFF ]
 *
 * [1] 64 instead of 32 as the status/command fields must be split
 *     across two cache lines.
 */
#define OFFSET_SCP      0x3FE0
#define OFFSET_ISCP     0x3FC0
#define OFFSET_SCB      (0x3F80+0x1E)	/* stat/cmd straddle c-line boundary */
#define OFFSET_RU	0x0000
#define OFFSET_RBD	0x02E0
#define OFFSET_CU	0x36A0

#define OFFSET_TBD      0x37A0
#define OFFSET_TBUF     0x3940

#define N_FD              23
#define N_RBD             23
#define N_TBD		  13

#define purgescbcmd(scb)	pcacheline(KERNSPC, (caddr_t)(scb)+2)
#define purgescbstat(scb) 	pcacheline(KERNSPC, (caddr_t)(scb)-30)
#define purgescball(scb)	purgescbstat(scb), purgescbcmd(scb)

#define flushscbcmd(scb)	fcacheline(KERNSPC, (caddr_t)(scb)+2)
#define flushscbstat(scb) 	fcacheline(KERNSPC, (caddr_t)(scb)-30)
#define flushscball(scb)	flushscbstat(scb), flushscbcmd(scb)

#define RCVBUFSIZE	540
#define DL_DEAD         0xffff

#define CMD_0           0     
#define CMD_1           0xffff

#define PC586NULL	0xffff			/* pc586 NULL for lists	*/

#define	DSF_LOCK	1
#define DSF_RUNNING	2

#define MOD_ENAL 1
#define MOD_PROM 2

/*
 * Driver (not board) specific defines and structures:
 */

#ifdef	MULTICAST
/*
 * If you require this pc586 driver to accept more than PC_MCASTLEN
 * Ethernet Multicast Addresses, change the #define below.
 */
#define PC_MCASTLEN     6
typedef struct  {
	u_short	ac_status;
	u_short	ac_command;
	u_short	ac_link_offset;
	/* multicast-specific data follows */
	u_short	mc_count;
	u_char	mc_addr[PC_MCASTLEN*ETHER_ADD_SIZE];
} multicast_t;
#endif

typedef	struct	{
	rbd_t	r;
	char	rbuffer[RCVBUFSIZE];
	char	rbd_pad[4];
} ru_t;

typedef struct { 
#ifdef	MACH_KERNEL
	struct	ifnet	ds_if;		/* generic interface header */
	u_char	ds_addr[6];		/* Ethernet hardware address */
#else
	struct	arpcom	pc586_ac;
#define	ds_if	pc586_ac.ac_if
#define	ds_addr	pc586_ac.ac_enaddr
#endif
	int	flags;
        int     seated;
#ifdef WATCHDOG
        int     timer;
#endif
        int     open;
        fd_t    *begin_fd;
	fd_t    *end_fd;
	rbd_t   *end_rbd;
	char    *prom;
	char	*sram;
#ifndef	IFF_OACTIVE
	int     tbusy;
#endif
	struct	core_lan *hwaddr;
#ifndef MACH_KERNEL
	caddr_t sc_bpf;
	short	sc_iflags;
#endif
#ifdef	MACH_KERNEL
	short	mode;
#endif
#ifdef	MULTICAST
	multicast_t sc_mcast;
	u_short	sc_mcrefcnt[PC_MCASTLEN];
#endif
} pc_softc_t;

#define	bcopy16(s,d,l)	bcopy((char *)(s), (char *)(d), (int)(l))

#ifdef __GNUC__
/*
 * All we *really* want to do is flush/purge stuff from the data cache.
 * The system pcacheline() and fcacheline() purge both I & D caches;
 * just inline what we want for speed and efficiency.
 */
#define pcacheline(spc, off) \
({ \
	register u_int _laddr_r = (u_int)off; \
	asm __volatile ("pdc %0(0)" : : "r" (_laddr_r)); \
	asm __volatile ("sync"); \
})
#define fcacheline(spc, off) \
({ \
	register u_int _laddr_r = (u_int)off; \
	asm __volatile ("fdc %0(0)" : : "r" (_laddr_r)); \
	asm __volatile ("sync"); \
})
#endif
