/* 
 * Copyright (c) 1992, 1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: mach_glue.h 1.8 94/12/14$
 */

/*
 * Glue to make OSF-1 or BSD drivers work in Mach 3.0
 * with minimal damage to driver code.
 */

#define DBG	1	/* XXX */

#define _NO_PROTO	/* XXX for gcc -traditional that we use */

#define	KERNSPC	0
#ifndef	NBPG
#define	NBPG	4096
#endif

#include <vm/vm_kern.h>

/* BSD stuff */
#define	sleep(event, pri) \
{ \
	thread_wait((int)(event), TRUE); \
}
#define wakeup(event) \
{ \
	thread_wakeup((int)(event)); \
}

#define ssmv(m)		ssm(m)

#define uprintf		printf
#define LOG_WARNING	1

#define ENODEV		D_NO_SUCH_DEVICE
#define EROFS		D_READ_ONLY
#define	EMFILE		D_NO_MEMORY
#define	EWOULDBLOCK	D_WOULD_BLOCK
#define	EBADF		D_READ_ONLY
#define	FREAD		D_READ
#define	FWRITE		D_WRITE
#define	FNDELAY		D_NODELAY

#define	MAX(a,b)	((a) > (b) ? (a) : (b))
#define	MIN(a,b)	((a) < (b) ? (a) : (b))
#define	DELAY(n)	delay(n)

#define CTRL(c)		(c&037)

/* Old Mach stuff */
#define	ZGET(zone, ret, type)		\
	ret = (type) zalloc(zone)
#if 0
#define	ZFREE(zone, element)		\
	zfree(zone, element)
#endif

/* HP OSF stuff */
#define pmap_flush_range(p, v, s)	\
{ \
	if ((p)->space != KERNSPC) \
		panic("pmap_flush_range: not kernel"); \
	else if ((((int)(v) & 31) + (s)) <= 32) { \
		register u_int _laddr_r = (u_int)(v); \
		asm __volatile ("fdc %0(0)" : : "r" (_laddr_r)); \
		asm __volatile ("sync"); \
	} else \
		fdcache(KERNSPC, v, s); \
}
#define pmap_purge_range(p, v, s)	\
{ \
	if ((p)->space != KERNSPC) \
		panic("pmap_purge_range: not kernel"); \
	else \
		pdcache(KERNSPC, v, s); \
}

#ifdef USELEDS
extern int inledcontrol;
/*
 * LED control
 */
#define PANEL_LED_DISK		0x04
#define PANEL_LED_TOGGLE	2
#define panel_led(led, how) \
	if (how == PANEL_LED_TOGGLE && inledcontrol == 0) ledcontrol(0, 0, led)
#else
#define	panel_led(led, how)
#endif

/*
 * Skeletal structs for attach routines
 */
struct hp_businfo {
	int	bustype;
};

#define	MAXCSRS	1
typedef struct hp_ctlrinfo {
	int		   ctlrnum;
	struct hp_ctlrinfo *parent_ctlrinfop;
	struct hp_businfo  *ctlr_businfop;
	int                irq;
	int		   numcsrs;
	vm_offset_t        csrs[MAXCSRS];
	vm_offset_t	   vcsrs[MAXCSRS];
} hp_ctlrinfo_t;

#define BUSTYPE_NOSUCHBUS	0
#define BUSTYPE_SYS		1
#define BUSTYPE_EISA		2

/*
 * For EISA
 */
#define RESOLVER_ADD_HANDLER 1
#define RESOLVER_DEL_HANDLER 2

#define	IO_EISA_IACK_PHYS	0xfc01f000
#define	IO_EISA_IACK_LEN	0x1000

#define io_map(pa, sz, nm)	(pa)
#define io_unmap(va, nm)

/*
 * Missing disk stuff
 */
#define dkunit(d)	(minor(d)>>3)
#define dkpart(d)	(minor(d)&0x07)
#define dkminor(u,p)	(((u)<<3)|((p)&0x07))
#define b_cylin		b_resid

#define	BBSIZE		8192
#define	SBSIZE		8192

struct partinfo {
	struct disklabel *disklab;
	struct partition *part;
};

#define	MAXPARTITIONS	8	/* XXX for disk_status.h */

#define DIOCGPART	_IOW('d', 104, struct partinfo)	/* get partition */
#define DIOCFORMAT	_IO('d', 111)
