/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * MACH 3.0, BSD, partial block and unaligned transfer support
 * done by the Computer Systems Laboratory (CSL) at The University of Utah.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: scsi_dsk.c 1.46 94/12/03$
 */

#ifndef COMPAT_NOLABEL
#define COMPAT_NOLABEL	/* XXX for now */
#endif

#include "sd.h"
#define NSCSI	NSD	/* XXX */

#if NSCSI > 0

/*
 * SCSI disk, floppy & cdrom driver
 */

#ifdef	MACH_KERNEL
#include <mach_glue.h>
#include <sys/types.h>
#include <device/param.h>
#include <device/errno.h>
#include <device/buf.h>
#include <device/conf.h>
#include <device/disk_status.h>
#include <machine/machparam.h>
#include <machine/machspl.h>
#include <scsi.h>
#include <scsi_common.h>
#else
#include <sys/param.h>
#include <sys/file.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/disklabel.h>
#include <sys/dk.h>
#include <sys/syslog.h>
#include <ufs/ffs/fs.h>
#include <sys/stat.h>
#include <sys/malloc.h>
#include <hpsgc/bsd_glue.h>
#include <machine/spl.h>
#include <hpsgc/scsi.h>
#include <hpsgc/scsi_common.h>
#endif
#ifdef PDCDUMP
#include <sys/reboot.h>
#endif

#ifdef  DBG
#define private
int     sddebug = 0;
#define SD_DEBUG_RW		0x01
#define SD_DEBUG_INIT		0x02
#define SD_DEBUG_CALLS		0x04
#define SD_DEBUG_CAPACITY	0x08
#define SD_DEBUG_MODE		0x10
#define SD_DEBUG_OPEN		0x20
#define SD_DEBUG_PARTIAL	0x40
#define SD_DEBUG_GEOM		0x80
#else
#define private static
#endif

#ifdef	MACH_KERNEL
struct  scsi_device scsi_dasd[] = {
        /* bus, target, lun, unused */
/* bus 0 */
        { 0,    0,      0,      0 },
        { 0,    1,      0,      0 },
        { 0,    2,      0,      0 },
        { 0,    3,      0,      0 },
        { 0,    4,      0,      0 },
        { 0,    5,      0,      0 },
        { 0,    6,      0,      0 },
        { 0,    7,      0,      0 },
/* bus 1 */
        { 1,    0,      0,      0 },
        { 1,    1,      0,      0 },
        { 1,    2,      0,      0 },
        { 1,    3,      0,      0 },
        { 1,    4,      0,      0 },
        { 1,    5,      0,      0 },
        { 1,    6,      0,      0 },
        { 1,    7,      0,      0 },
/* bus 2 */
        { 2,    0,      0,      0 },
        { 2,    1,      0,      0 },
        { 2,    2,      0,      0 },
        { 2,    3,      0,      0 },
        { 2,    4,      0,      0 },
        { 2,    5,      0,      0 },
        { 2,    6,      0,      0 },
        { 2,    7,      0,      0 },
/* bus 3 */
        { 3,    0,      0,      0 },
        { 3,    1,      0,      0 },
        { 3,    2,      0,      0 },
        { 3,    3,      0,      0 },
        { 3,    4,      0,      0 },
        { 3,    5,      0,      0 },
        { 3,    6,      0,      0 },
        { 3,    7,      0,      0 },
};
#define MAXUNITS        (sizeof(scsi_dasd)/sizeof(scsi_dasd[0]))
#else
extern	struct scsi_device scsi_dasd[];
#define	MAXUNITS	NSD
#endif

#define BUS(u)          scsi_dasd[u].bus
#define TGT(u)          scsi_dasd[u].target

/* Convert major/minor device numbers to unit/partition */
#define SDUNIT(d) dkunit(d)
#define SDPART(d) dkpart(d)
#ifdef	MACH_KERNEL
#define SDPART0(d)	(minor(d)&-8)
#define SDLABPART(d)	SDPART0(d)
#else
#define SDPART0(d)	((d)&-8)
#define SDLABPART(d)	(SDPART0(d)|2)	/* XXX 'c' partition */
#endif

/* Convert unit/partition to major/minor device numbers */
#define SDMAKEDEV(u,p) dkminor(u,p)

/*
 * Per UNIT (i.e., per disk) structure.
 */
typedef enum { SHIFTLEFT, SHIFTRIGHT } sd_shift;

struct  sdinfo {
        unsigned int            sd_init         : 1;    /* Init status */
        unsigned int            sd_floppy       : 1;    /* floppy drive? */
        unsigned int            sd_readonly     : 1;    /* Disk unit is readonly */
        unsigned int            sd_write_cache  : 1;    /* Disk unit supports write cache */
        unsigned int            sd_bvolatile    : 1;    /* Disklabel/boot area writeable */
        unsigned int            sd_label_valid  : 1;    /* Disk label valid */
        unsigned int            sd_media_locked : 1;    /* Media locked */
	unsigned int		sd_probed       : 1;	/* Device detected */
        unsigned char           sd_open;                /* Discreet openers (block + character) */
        unsigned char           sd_bopen;               /* Discreet character device openers */
        unsigned char           sd_copen;               /* Discreet block device openers */
        unsigned char           sd_btosshift;           /* DEV_BSIZE block# to sector# shift left factor */
        unsigned char           sd_bctoscshift;         /* Byte count to sector count shift right factor */
        sd_shift                sd_btosdirection;       /* Direction to shift bttos */
#ifdef	MACH_KERNEL
	unsigned char           sd_dkn;                 /* Dk stats */
#else
	char			sd_dkn;			/* Dk stats */
#endif
        struct  buf             sd_q;                   /* Head of buffer queue */
        struct  buf             sd_raw;                 /* Buf for raw I/O */
        struct  disklabel       sd_label;               /* Disklabel */
        struct  scsi_inquiry    sd_inquiry;             /* Inquiry structure */
} sdinfo[MAXUNITS];

/* Structure of drive specific data stored in disklabel */
struct  drivedata {
        unsigned        awre:1;                 /* Automatic Write REallocation */
        unsigned        arre:1;                 /* Automatic Read REallocation */
        unsigned        tb:1;                   /* Transfer Block */
        unsigned        rc:1;                   /* Read Continuous */
        unsigned        eer:1;                  /* Enable Early Recovery */
        unsigned        per:1;                  /* Post ERror */
        unsigned        dte:1;                  /* Disable Transfer on Error */
        unsigned        dcr:1;                  /* Disable CoRrection */
        unsigned        read_retry_count:8;
        unsigned        correction_span:8;
        unsigned        head_offset_count:8;
        unsigned        data_strobe_offset_count:8;
};

#ifndef	MACH_KERNEL
extern	int dkn;
#endif
extern	int dcache_line_size;

private int
sderror(cmd, info)
struct  scsi_cmd        *cmd;
struct  sdinfo          *info;
{
	int     x;
	char    *err;
	struct  rw_10_cmd *cp = (struct rw_10_cmd *) &cmd->cdb;
	struct  scsi_request_sense *sense;

#ifdef  DBG
	if (sddebug & SD_DEBUG_CALLS)
		printf("sderror(0x%x, 0x%x)\n", cmd, info);
#endif
        switch (cmd->scsi_status) {
        case SCSI_CHECK_CONDITION:
                if (!(sense = cmd->sense)) {
                        panic("sderror: no sense data");
                }
                else if (cmd->scsi_sense_status != SCSI_GOOD) {
                        printf("sderror: bad sense status (%d) bus=%d, target=%d, lun=%d\n", cmd->scsi_sense_status,
			cmd->device.bus, cmd->device.target, cmd->device.lun);
                        panic("sderror: bad sense status");
                }

                /* log sense data */
                log(LOG_WARNING, "sderror: CHECK CONDITION\n");
                log(LOG_WARNING, "sderror: valid = %d, error code = 0x%x, segment = 0x%x\n",
                        sense->valid, sense->error_code, sense->segment_number);
                log(LOG_WARNING, "sderror: filemark = %d, eom = %d, ili = %d, sense key = 0x%x\n",
                        sense->filemark, sense->end_of_media, sense->incorrect_length, sense->sense_key);
                if (sense->sense_key <= scsi_sense_key_count) log(LOG_WARNING, "sderror: sense_key = %s\n", scsi_sense_key_text[sense->sense_key]);
                if (sense->valid) log(LOG_WARNING, "sderror: info = 0x%x 0x%x 0x%x 0x%x\n",
                        sense->information_bytes[0], sense->information_bytes[1],
                        sense->information_bytes[2], sense->information_bytes[3]);
                log(LOG_WARNING, "sderror: add. sense length %d, cmd info = 0x%x 0x%x 0x%x 0x%x\n",
                        sense->additional_sense_length, sense->command_information[0],
                        sense->command_information[1], sense->command_information[2],
                        sense->command_information[3]);
                log(LOG_WARNING, "sderror: add. sense code 0x%x, add. sense code qualifier 0x%x, FRU code = 0x%x\n",
                        sense->additional_sense_code, sense->additional_sense_code_qualifier,
                        sense->field_replaceable_unit_code);
		log(LOG_WARNING, "sderror: sense key specific: sksv = %d, cd = %d, bpv = %d, bit_pointer = %d, field_pointer = 0x%x\n",
			sense->sksv, sense->cd, sense->bpv, sense->bit_pointer, sense->field_pointer);
                if (sense->additional_sense_length) {
                        log(LOG_WARNING, "sderror: add sense");
                        for (x = 0; x < sizeof(sense->additional_sense_bytes); ++x) {
                                if ((x & 7) == 0)
                                        log(LOG_WARNING, "\nsderror: add sense");
                                log(LOG_WARNING, " 0x%x", sense->additional_sense_bytes[x]);
                        }
                        log(LOG_WARNING, "\n");
                }

                /* interpret sense data */
                if (sense->sense_key == SCSI_SENSE_RECOVERED_ERROR) {
                        if (cp->opcode == SCSI_READ_10 || cp->opcode == SCSI_WRITE_10) {
                                log(LOG_WARNING,
                                        "sderror: recovered error bus=%d, target=%d, lun=%d (%s, blk_addr 0x%x, nblks 0x%x)\n",
                                        cmd->device.bus, cmd->device.target, cmd->device.lun,
                                        ((cp->opcode == SCSI_READ_10) ? "read" : "write"), RW10_GET_BADDR(cp), RW10_GET_BLEN(cp));
                        }
                        else {
                                /*
                                 * Simply dump cdb
                                 */
                                log(LOG_WARNING,
                                        "sderror: recovered error bus=%d, target=%d, lun=%d (cdb=0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x)\n",
                                        cmd->device.bus, cmd->device.target, cmd->device.lun,
                                        cmd->cdb.command[0], cmd->cdb.command[1], cmd->cdb.command[2], cmd->cdb.command[3],
                                        cmd->cdb.command[4], cmd->cdb.command[5], cmd->cdb.command[6], cmd->cdb.command[7],
                                        cmd->cdb.command[8], cmd->cdb.command[9], cmd->cdb.command[10], cmd->cdb.command[11]);
                        }
                        return(0);
                }
		else {
                        if (cp->opcode == SCSI_READ_10 || cp->opcode == SCSI_WRITE_10) {
                                log(LOG_WARNING,
                                        "sderror: error bus=%d, target=%d, lun=%d (%s, blk_addr 0x%x, nblks 0x%x)\n",
                                        cmd->device.bus, cmd->device.target, cmd->device.lun,
                                        ((cp->opcode == SCSI_READ_10) ? "read" : "write"), RW10_GET_BADDR(cp), RW10_GET_BLEN(cp));
                        }
                        else {
                                /*
                                 * Simply dump cdb
                                 */
                                log(LOG_WARNING,
                                        "sderror: error bus=%d, target=%d, lun=%d (cdb=0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x)\n",
                                        cmd->device.bus, cmd->device.target, cmd->device.lun,
                                        cmd->cdb.command[0], cmd->cdb.command[1], cmd->cdb.command[2], cmd->cdb.command[3],
                                        cmd->cdb.command[4], cmd->cdb.command[5], cmd->cdb.command[6], cmd->cdb.command[7],
                                        cmd->cdb.command[8], cmd->cdb.command[9], cmd->cdb.command[10], cmd->cdb.command[11]);
                        }
                        return(1);
		}
                break;

        case SCSI_GOOD:
                err = "Good status";
                break;
        case SCSI_CONDITION_MET:
                err = "Condition met";
                break;
        case SCSI_BUSY:
                err = "Device busy";
                break;
        case SCSI_INTERMEDIATE_GOOD:
                err = "Intermediate good";
                break;
        case SCSI_INTERMEDIATE_CONDITION_MET:
                err = "Intermediate condition met";
                break;
        case SCSI_RESERVATION_CONFLICT:
                err = "Reservation conflict";
                break;
        case SCSI_QUEUE_FULL:
                err = "Queue full";
                break;
        case SCSI_DATA_DIRECTION_ERROR:
                err = "Data direction error";
                break;
        case SCSI_DMA_UNDERRUN:
                err = "DMA underrun";
                break;
        case SCSI_DMA_OVERRUN:
                err = "DMA overrun";
                break;
        case SCSI_DEVICE_NA:
                err = "Device not available";
                break;
        case SCSI_UNDEF_STAT:
                err = "Undefined";
                break;
        default:
                err = "Unknown";
                break;
        }
        printf("sderror: %s, status 0x%x\n", err, cmd->scsi_status);
        return(1);
}

#ifndef	MACH_KERNEL
#define strncmp	bcmp
#else
static int
strncmp(x, y, n)
        unsigned char *x, *y;
        int n;
{
        int i, t;
        
        for (i=0; i<n && !(t=*x-*y) && (*x!=0); i++,x++,y++);
        return (t);
}
#endif

/*
 * Process command completion.
 */
void
sddone(cmd)
struct  scsi_cmd        *cmd;
{
        struct  sdinfo          *info;
        struct  buf             *bp, *q;
        struct  scsi_device     sd;

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sddone(0x%x)\n", cmd);
#endif
        sd = cmd->device;
        bp = (struct buf *)cmd->driver_info;
        bp->b_resid = 0;
        info = &sdinfo[SDUNIT(bp->b_dev)];

        if (cmd->flags & SCSI_FLUSH) {
                /* we have had some fatal error on this device */
                bp->b_flags |= B_ERROR;
                bp->b_error = EIO;
                bp->b_resid = bp->b_bcount;
        }
        else switch(cmd->scsi_status) {

        case SCSI_GOOD:
                break;

        case SCSI_CHECK_CONDITION:
                /* check sense data */
                if (sderror(cmd, info)) {
                        bp->b_flags |= B_ERROR;
                        bp->b_error = EIO;
                        bp->b_resid = bp->b_bcount;
                        scsi_flush(cmd->device);
                }
                break;

        case SCSI_CONDITION_MET:
        case SCSI_BUSY:
        case SCSI_INTERMEDIATE_GOOD:
        case SCSI_INTERMEDIATE_CONDITION_MET:
        case SCSI_RESERVATION_CONFLICT:
        case SCSI_QUEUE_FULL:
        default:
                /* This driver doesn't do any of this nonsense, so just panic */
                printf("sddone: scsi_status 0x%x\n", cmd->scsi_status);
                panic("sddone: unrecognized scsi_status");
        }
#ifndef	MACH_KERNEL
        if (info->sd_probed && info->sd_dkn >= 0)
                dk_busy &= ~(1 << info->sd_dkn);
#endif

	if (cmd->sense) scsi_sense_free(cmd->sense);
        scsi_command_free(cmd);

        q = &info->sd_q;
#ifdef MACH_KERNEL
        q->b_actf = bp->av_forw;
        if (q->b_actl == bp) {
                q->b_actl = (struct buf *)0;
        }
#else
	/*
	 * Neither BSD nor this driver use `b_actl/b_actb'.
	 */
	q->b_actf = bp->b_actf;
#endif
        q->b_active = 0;

        biodone(bp);
        sdustart(info, sd);
}

#ifndef	MACH_KERNEL
/*
 * This routine is called for partial block transfers and non-aligned
 * transfers (the latter only being possible on devices with a block size
 * larger than DEV_BSIZE).  The operation is performed in three steps
 * using a locally allocated buffer:
 *	1. transfer any initial partial block
 *	2. transfer full blocks
 *	3. transfer any final partial block
 */
static void
sdlblkstrat(bp, bsize)
	register struct buf *bp;
	register int bsize;
{
	register struct buf *cbp = (struct buf *)malloc(sizeof(struct buf),
							M_DEVBUF, M_WAITOK);
	caddr_t cbuf = (caddr_t)malloc(bsize, M_DEVBUF, M_WAITOK);
	space_t csp;
	register int bn, resid;
	register caddr_t addr;
	void sdstrategy();

	bzero((caddr_t)cbp, sizeof(*cbp));
	cbp->b_dev = bp->b_dev;
	bn = bp->b_blkno;
	resid = bp->b_bcount;
	addr = bp->b_un.b_addr;
	if ((bp->b_flags & (B_PAGET|B_PHYS)) == B_PHYS)
		csp = bvtospace(bp, addr);
	else
		csp = KERNSPC;
#ifdef DBG
	if (sddebug & SD_DEBUG_PARTIAL)
		printf("sdlblkstrat: bp %x flags %x bn %x resid %x addr %x\n",
		       bp, bp->b_flags, bn, resid, addr);
#endif

	while (resid > 0) {
		register int boff = dbtob(bn) & (bsize - 1);
		register int count;

		if (boff || resid < bsize) {
			count = MIN(resid, bsize - boff);
			cbp->b_flags = B_BUSY | B_PHYS | B_READ;
			/*
			 * Force KERNSPC for translation of b_addr since we're
			 * doing IO into a local (kernel) buffer.
			 */
			cbp->b_proc = NULL;
			cbp->b_blkno = bn - btodb(boff);
			cbp->b_un.b_addr = cbuf;
			cbp->b_bcount = bsize;
#ifdef DBG
			if (sddebug & SD_DEBUG_PARTIAL)
				printf(" readahead: bn %x cnt %x off %x addr %x\n",
				       cbp->b_blkno, count, boff, addr);
#endif
			sdstrategy(cbp);
			biowait(cbp);
			if (cbp->b_flags & B_ERROR) {
				bp->b_flags |= B_ERROR;
				bp->b_error = cbp->b_error;
				break;
			}
			if (bp->b_flags & B_READ) {
#ifdef DBG
				if (sddebug & SD_DEBUG_PARTIAL)
					printf(" bcopy(%x.%x, %x.%x, %d)\n",
					       KERNSPC, &cbuf[boff],
					       csp, addr, count);
#endif
				lbcopy(KERNSPC, &cbuf[boff], csp, addr, count);
				goto done;
			}
#ifdef DBG
			if (sddebug & SD_DEBUG_PARTIAL)
				printf(" bcopy(%x.%x, %x.%x, %d)\n",
				       csp, addr, KERNSPC, &cbuf[boff], count);
#endif
			lbcopy(csp, addr, KERNSPC, &cbuf[boff], count);
#ifdef DBG
			if (sddebug & SD_DEBUG_PARTIAL)
				printf(" writeback: bn %x cnt %x off %x addr %x\n",
				       cbp->b_blkno, count, boff, addr);
#endif
		} else {
			count = resid & ~(bsize - 1);
			/* 
			 * If page in/out of a page table, then we're doing
			 * io in kernel space.  Else we're doing io in the
			 * space specified by bp->b_proc.
			 */
			if (csp == KERNSPC)
				cbp->b_proc = 0;
			else
				cbp->b_proc = bp->b_proc;
			cbp->b_blkno = bn;
			cbp->b_un.b_addr = addr;
			cbp->b_bcount = count;
#ifdef DBG
			if (sddebug & SD_DEBUG_PARTIAL)
				printf(" fulltrans: bn %x cnt %x addr %x\n",
				       cbp->b_blkno, count, addr);
#endif
		}
		cbp->b_flags = B_BUSY | B_PHYS | (bp->b_flags & B_READ);
		sdstrategy(cbp);
		biowait(cbp);
		if (cbp->b_flags & B_ERROR) {
			bp->b_flags |= B_ERROR;
			bp->b_error = cbp->b_error;
			break;
		}
done:
		bn += btodb(count);
		resid -= count;
		addr += count;
#ifdef DBG
		if (sddebug & SD_DEBUG_PARTIAL)
			printf(" done: bn %x resid %x addr %x\n",
			       bn, resid, addr);
#endif
	}
	free(cbuf, M_DEVBUF);
	free(cbp, M_DEVBUF);
}
#endif

/*
 * Perform I/O.
 */
void
sdstrategy(bp)
struct  buf     *bp;
{
        struct  sdinfo          *info;
        struct  disklabel       *d;
        struct  partition       *p;
        struct  scsi_device     sd;
        daddr_t blkno;
        int     unit = SDUNIT(bp->b_dev);
        int     part = SDPART(bp->b_dev);
        int     error, s;

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdstrategy(0x%x)\n", bp);
#endif

        info = &sdinfo[unit];
        sd = scsi_dasd[unit];
        d = &info->sd_label;

        if (part >= d->d_npartitions) {
                /* Not a valid partition */
                error = ENODEV;
                goto bogus;
        }
        p = &d->d_partitions[part];

#ifdef	MACH_KERNEL
        /* only whole device sectors allowed */
        if (bp->b_bcount & (DEV_BSIZE-1)) {
                error = EINVAL;
                goto bogus;
        }
#endif

        /* Calculate the block# and make sure it's reasonable */
        blkno = bp->b_blkno;
        if (blkno >= p->p_size) {
                if (blkno == p->p_size) {
                        error = 0;
                }
                else {  /* OverRan Partition */
                        error = EINVAL;
                }
                goto bogus;
        }

        /* will this i/o go past partition end? */
	if ((blkno + btodb(bp->b_bcount)) >= p->p_size) {
                bp->b_bcount = dbtob(p->p_size - blkno);
        }
        blkno += p->p_offset;

        if ((bp->b_flags & B_READ) == B_WRITE)  {
                /* Make sure we're not writing on protected media OR in the label/bootblock area */
                if (info->sd_readonly ||
                    ((blkno < d->d_bbsize/DEV_BSIZE) && (!info->sd_bvolatile) && info->sd_label_valid)) {
                        error = EROFS;
                        goto bogus;
                }
        }
#ifndef	MACH_KERNEL
        /*
         * Non-aligned or partial-block transfers handled specially.
         * XXX none of the former yet.
         */
	if (info->sd_btosshift) {
		if (info->sd_btosdirection == SHIFTLEFT)
			s = (DEV_BSIZE >> info->sd_btosshift) - 1;
		else
			s = (DEV_BSIZE << info->sd_btosshift) - 1;
	} else
		s = DEV_BSIZE - 1;
        if ((dbtob(blkno) & s) || (bp->b_bcount & s)) {
                sdlblkstrat(bp, s+1);
                goto done;
        }
#endif

        bp->b_cylin = blkno;

        s = splbio();
        disksort(&info->sd_q, bp);
        sdustart(info, sd);
        splx(s);
        return;

bogus:
        bp->b_resid = bp->b_bcount;
        bp->b_error = error;
        if (error)
                bp->b_flags |= B_ERROR;
done:
        biodone(bp);
        return;
}

/*
 * Start unit operation.
 *
 * NOTE: Must be called at splbio().
 */
sdustart(info, sd)
struct  sdinfo          *info;
struct  scsi_device     sd;
{
        struct  buf     *q, *bp;
        struct  scsi_cmd *cmd;
        struct  rw_10_cmd *cp;
        int     s;
        daddr_t blkno;
        u_int   addr;

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdustart(0x%x)\n", info);
#endif
        q = &info->sd_q;
        if (q->b_active)
                return;

        bp = q->b_actf;
        if (bp) {
                q->b_active = 1;

                cmd = scsi_command_alloc();
                *cmd = nullcmd;
                cmd->device = sd;

                if (bp->b_flags & B_READ)
                        cmd->flags = SCSI_CALL | SCSI_DATA_IN;
                else
                        cmd->flags = SCSI_CALL | SCSI_DATA_OUT;
                cmd->data_len = bp->b_bcount;
                cmd->data_addr = (vm_offset_t)bp->b_un.b_addr;
#ifdef	MACH_KERNEL
		cmd->data_map = VM_MAP_NULL;
#else
                if (bp->b_proc)
			cmd->data_map = bp;
#ifdef DEBUG
		else if (bp->b_text)
			printf("sd%d: sdustart: bp=%x proc=%x text=%x flags=%x blkno=%x addr=%x bcount=%x vp=%x\n",
			       info->sd_dkn, bp, bp->b_proc, bp->b_text,
			       bp->b_flags, bp->b_blkno, bp->b_un.b_addr,
			       bp->b_bcount, bp->b_vp);
#endif
                if (bp->b_flags & B_PHYS)
                        cmd->flags |= SCSI_PHYSIO;
#endif
                cmd->io_done = sddone;
                cmd->driver_info = (vm_offset_t) bp;
                cp = (struct rw_10_cmd *) &cmd->cdb;
                if (bp->b_flags & B_READ) {
                        cp->opcode = SCSI_READ_10;
                }
                else {
                        cp->opcode = SCSI_WRITE_10;
/*
 * XXX apparently the existence of a write cache does not imply that
 * FUA is implemented.
 */
#if 0
                        /* force synchronous writes to media */
                        if (info->sd_write_cache && !(bp->b_flags & B_ASYNC))
                                cp->fua = 1;
#endif
                }

                blkno = bp->b_cylin;
                if (info->sd_btosdirection == SHIFTLEFT)
                        addr = blkno << info->sd_btosshift;
                else
                        addr = blkno >> info->sd_btosshift;
                RW10_SET_BADDR(cp, addr);
                RW10_SET_BLEN(cp, bp->b_bcount >> info->sd_bctoscshift);

#ifdef  DBG
                if (sddebug & SD_DEBUG_RW) {
                        printf("sdustart: bp=%x blkno=0x%x(%x) bcount=0x%x flags=%x cdbaddr=0x%x cdbcount=0x%x\n",
                               bp, bp->b_blkno, bp->b_cylin, bp->b_bcount,
                               bp->b_flags, addr,
			       bp->b_bcount>>info->sd_bctoscshift);

                }
#endif
                panel_led(PANEL_LED_DISK, PANEL_LED_TOGGLE);
                scsi_command(SCSI_QUEUE_TAIL, cmd);
#ifndef	MACH_KERNEL
                if (info->sd_probed && info->sd_dkn >= 0) {
                        dk_busy |= 1 << info->sd_dkn;
                        dk_xfer[info->sd_dkn]++;
                        dk_wds[info->sd_dkn] += bp->b_bcount >> 6;
                }
#endif
        }
        return;
}

/*
 * Read the disklabel from the disk.
 */
int
sdreadlabel(dev, d, info)
dev_t   dev;
struct  disklabel       *d;
struct  sdinfo          *info;
{
        int     secperunit = d->d_secperunit;   /* Never more than this */
        int     secsize = d->d_secsize;         /* Do not alter this */
        int     error = 0;

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdreadlabel(0x%x, 0x%x)\n", dev, d);
#endif

	/* XXX ufs_disksubr forces us to use the 'a' partition */
        if (readdisklabel(SDPART0(dev), sdstrategy, d))
                error = EIO;
        else
                info->sd_label_valid = 1;

        if (d->d_secperunit > secperunit)
                d->d_secperunit = secperunit;

        d->d_secsize = secsize;
#ifdef	MACH_KERNEL
	/*
	 * XXX allow us to use 'c' partition.
	 * dislabel falls within the 8k unused area of a filesystem
	 * so we don't have to worry about clobbering it.
	 */
	d->d_bbsize = 0;
#endif

        /* Since we may have altered it, need new checksum */
        d->d_checksum = 0;
        d->d_checksum = dkcksum(d);
        return(error);
}

#ifndef	MACH_KERNEL
#define	uprintf	printf	/* XXX for boot time calls */
#endif

/*
 * Drive initialization.
 */
int
sdinit(dev, probing)
dev_t dev;
int probing;
{
        struct  sdinfo *info;
        struct  scsi_cmd                *cmd;
        struct  scsi_inquiry            *inquiry;
        struct  scsi_request_sense      *sense;
        struct  scsi_mode1_header       *header;
        struct  scsi_mode_block         *block;
	struct	mode_sense_6_cmd	*msp;
        struct  scsi_mode_error_recovery_page   *error_page, changeable_error_page;
        struct  scsi_mode_cache_control_page    *cache_page, changeable_cache_page;
        struct  scsi_mode_flex_disk_param_page  *flex_page,  changeable_flex_page;
        struct  scsi_device             sd;
	vm_offset_t			data_buffer;
	vm_size_t			size;
        int                             unit, i, error = 0, retries = 0;

#define RETURN(e)		{ error = (e); goto done; }

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdinit(0x%x)\n", dev);
#endif

        unit = SDUNIT(dev);
        info = &sdinfo[unit];

        info->sd_floppy = 0;
        info->sd_readonly = 0;
        info->sd_write_cache = 0;
        info->sd_label_valid = 0;

	size = MAX(MAX(sizeof(*sense), sizeof(*inquiry)),
		   SCSI_MODE_SENSE_MAXLEN);
	size = (size + (dcache_line_size-1)) & ~(dcache_line_size-1);

	data_buffer = 0;
        kmem_alloc(kernel_map, &data_buffer, size);

        sd = scsi_dasd[unit];

        cmd = scsi_command_alloc();
        *cmd = nullcmd;

	/* Request the sense in order to clear any unit attentions */
#ifdef DBG
        if (sddebug & SD_DEBUG_INIT) {
                printf("sdinit: sd(%d,%d): initial request sense\n",
		       BUS(unit), TGT(unit));
        }
#endif
	sense = (struct scsi_request_sense *)data_buffer;
	SCSI_CMD6_BUF(cmd, SCSI_REQUEST_SENSE, *sense);
	cmd->device = sd;
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	scsi_command(SCSI_QUEUE_TAIL, cmd);

	if (cmd->scsi_status == SCSI_DEVICE_NA) {
                if (cmd->sense)
			panic("sdinit: sense data on non-available device!");
                RETURN(EIO);
	}

again:
        /* do a TEST UNIT READY operation to find drive availability */
#ifdef DBG
        if (sddebug & SD_DEBUG_INIT) {
                printf("sdinit: test unit ready\n");
        }
#endif
        SCSI_CMD6(cmd, SCSI_TEST_UNIT_READY);
        cmd->device = sd;
        cmd->flags = SCSI_WAIT;
        scsi_command(SCSI_QUEUE_TAIL, cmd);

        if (cmd->scsi_status == SCSI_DEVICE_NA) {
                if (cmd->sense)
                        panic("sdinit: sense data on non-available device!");
                RETURN(ENXIO);
        }
        else if (cmd->scsi_status != SCSI_GOOD) {
       		if (cmd->sense && cmd->sense->sense_key == SCSI_SENSE_UNIT_ATTENTION) {
       	                /* first one sometimes fails due to unit attn... just retry */
                        scsi_sense_free(cmd->sense);
                        SCSI_CMD6(cmd, SCSI_TEST_UNIT_READY);
                        cmd->device = sd;
                        cmd->flags = SCSI_WAIT;
                        scsi_command(SCSI_QUEUE_TAIL, cmd);
                }
        	if (cmd->scsi_status != SCSI_GOOD &&
		    (probing || sderror(cmd, info))) {
			/*
			 * A quick check to see if drive is ready.
			 * Give it a few tries before giving up.
			 */
                        if (cmd->sense &&
			    cmd->sense->sense_key == SCSI_SENSE_NOT_READY) {
				if (retries++ < 3) {
					scsi_sense_free(cmd->sense);
					DELAY(1000000);
					goto again;
				}
				if (!probing)
					printf("sd[%d,%d]: unit not ready\n",
					       BUS(unit), TGT(unit));
                        }
#ifdef  DBG
                        else if (cmd->sense) {
                                printf("sd[%d,%d]: failed test unit ready\n",
				       BUS(unit), TGT(unit));
                        }
#endif
			if (!probing)
				RETURN(EIO);
		}
        }

        /* do an Inquiry to see if this device is what we expect */
        inquiry = (struct scsi_inquiry *) data_buffer;
        SCSI_CMD6_BUF(cmd, SCSI_INQUIRY, *inquiry);
        cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
        cmd->device = sd;
        scsi_command(SCSI_QUEUE_TAIL, cmd);
        if (cmd->scsi_status != SCSI_GOOD && (probing || sderror(cmd, info))) {
#ifdef	DBG
                /* a quick check to see if drive is ready */
		if (!probing) {
			if (cmd->sense &&
			    cmd->sense->sense_key == SCSI_SENSE_NOT_READY) {
				printf("sd(%d,%d): 2nd unit not ready\n",
				       BUS(unit), TGT(unit));
			}
			else if (cmd->sense) {
				printf("sd(%d,%d): failed to respond to inquiry\n",
				       BUS(unit), TGT(unit));
			}
		}
#endif
                RETURN(EIO);
        }

        /* copy to info */
        info->sd_inquiry = *inquiry;
        inquiry = &info->sd_inquiry;
#ifdef  DBG
        if (sddebug & SD_DEBUG_INIT) {
                printf("sdinit: inquiry removable_media=0x%x iso_version=0x%x ecma_version=0x%x ansi_version=0x%x\n",
                        inquiry->removable_media, inquiry->iso_version, inquiry->ecma_version, inquiry->ansi_version);
                printf("sdinit: inquiry supports aen=0x%x relative_address=0x%x 32bit_xfers=0x%x 16bit_xfers=0x%x\n",
                        inquiry->supports_async_event_notification, inquiry->supports_relative_address,
			inquiry->supports_32bit_transfers, inquiry->supports_16bit_transfers);
                printf("sdinit: inquiry sync=0x%x linked cmds=0x%x cacheing=0x%x cmd queueing=0x%x\n",
                        inquiry->supports_sync_transfers, inquiry->supports_linked_operations,
			inquiry->supports_cacheing, inquiry->supports_command_queuing);
                printf("sdinit: peripheral_device_type=0x%x\n", inquiry->peripheral_device_type);
		printf("sdinit:\n");
		printf("\tvendor identification=");
		for (i = 0; i < sizeof(inquiry->vendor_identification); i++)
			printf("%c", inquiry->vendor_identification[i]);
		printf("\n");
		printf("\tproduct identification=");
		for (i = 0; i < sizeof(inquiry->product_identification); i++)
			printf("%c", inquiry->product_identification[i]);
		printf("\n");
		printf("\tproduct revision=");
		for (i = 0; i < sizeof(inquiry->product_revision); i++)
			printf("%c", inquiry->product_revision[i]);
		printf("\n");
        }
#endif
        if ((inquiry->peripheral_device_type != SCSI_DIRECT_ACCESS) &&
            (inquiry->peripheral_device_type != SCSI_OPTICAL_DEVICE) &&
            (inquiry->peripheral_device_type != SCSI_ROM_DEVICE)) {
#ifdef DBG
		if (!probing)
			uprintf("sd(%d,%d): not a direct access device\n",
				BUS(unit), TGT(unit));
#endif
                RETURN (EIO);
        }

	if (!info->sd_probed) {
		info->sd_probed = 1;
#ifndef	MACH_KERNEL
        	info->sd_dkn = dkn++;
#endif
	}

	/*
	 * For boot time probing, this is all we do.
	 * Print out a suitable description and return.
	 */
	if (probing) {
		char idstr[32];

		printf("sd%d: ", unit);
		switch (inquiry->ansi_version) {
		case 1:
		case 2:
			bcopy(inquiry->vendor_identification, idstr, 28);
			for (i = 27; i > 23; --i)
				if (idstr[i] != ' ')
					break;
			idstr[i+1] = 0;
			for (i = 23; i > 7; --i)
				if (idstr[i] != ' ')
					break;
			idstr[i+1] = 0;
			for (i = 7; i >= 0; --i)
				if (idstr[i] != ' ')
					break;
			idstr[i+1] = 0;
			printf("%s %s rev %s", idstr, &idstr[8], &idstr[24]);
			if (inquiry->ansi_version == 2)
				printf(" (SCSI-2)");
			break;
		default:
			printf("type 0x%x, qual 0x%x, ver %d",
			       inquiry->peripheral_device_type,
			       inquiry->device_type_qualifier,
			       inquiry->ansi_version);
			break;
		}
		printf("\n");
#ifdef	DBG
		if (sddebug & SD_DEBUG_GEOM) {
			sdshowgeometry(unit);
		}
#endif
		RETURN(0);
	}

	/*
	 * This drive rev has problems with mode select/sense so avoid them
	 */
        if ((strncmp("QUANTUM ", inquiry->vendor_identification, sizeof(inquiry->vendor_identification)) == 0) &&
		(strncmp("508D", inquiry->product_revision, sizeof(inquiry->product_revision)) == 0)) {
		info->sd_write_cache = 1;
		error = sdinitlabel(dev);
		goto done;
        }

        /* RESERVE the drive */
#ifdef DBG
        if (sddebug & SD_DEBUG_INIT) {
                printf("sdinit: reserving drive\n");
        }
#endif
        SCSI_CMD6(cmd, SCSI_RESERVE);
        cmd->flags = SCSI_WAIT;
        cmd->device = sd;
        scsi_command(SCSI_QUEUE_TAIL, cmd);
        if (cmd->scsi_status != SCSI_GOOD && sderror(cmd, info)) {
#ifdef	DBG
       		/* a quick check to see if drive is ready */
       		if (cmd->sense && cmd->sense->sense_key == SCSI_SENSE_NOT_READY) {
       			printf("sd(%d,%d): 3rd unit not ready\n", BUS(unit), TGT(unit));
       		}
       		else if (cmd->sense) {
       			printf("sd(%d,%d): failed to reserve unit\n", BUS(unit), TGT(unit));
       		}
#endif
       		RETURN(EIO);
        }

        /* do a START UNIT to spin up drive */
#ifdef DBG
        if (sddebug & SD_DEBUG_INIT) {
                printf("sdinit: start unit\n");
        }
#endif
        SCSI_CMD6(cmd, SCSI_START_UNIT);
	((struct start_stop_unit_cmd *)(&cmd->cdb))->start = 1; 
        cmd->device = sd;
        cmd->flags = SCSI_WAIT;
        scsi_command(SCSI_QUEUE_TAIL, cmd);
       	if (cmd->scsi_status != SCSI_GOOD && sderror(cmd, info)) {
#ifdef	DBG
                /* a quick check to see if drive is ready */
                if (cmd->sense && cmd->sense->sense_key == SCSI_SENSE_NOT_READY) {
                        printf("sd(%d,%d): 4th unit not ready\n", BUS(unit), TGT(unit));
                }
                else if (cmd->sense) {
                        printf("sd(%d,%d): failed to start unit\n", BUS(unit), TGT(unit));
                }
#endif
                RETURN(EIO);
        }

        /* set up for mode sense/select */
        header = (struct scsi_mode1_header *) data_buffer;
        block = (struct scsi_mode_block *) ((char *) header + sizeof(struct scsi_mode1_header));

        if (inquiry->removable_media) {         /* Check for Floppy */
                /* do a mode sense to determine floppy geometry */
#ifdef DBG
                if (sddebug & SD_DEBUG_INIT) {
                        printf("sdinit: mode sense for removable media\n");
                }
#endif
                SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
                cmd->device = sd;
                cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
		msp = (struct mode_sense_6_cmd *) &cmd->cdb;
                msp->pc = SCSI_MODE_SENSE_PC_CURRENT;
                msp->page_code = SCSI_MODE_FLEX_DISK_PARAM_PAGE;
                cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
                msp->allocation_length = cmd->data_len;
                cmd->data_addr = (vm_offset_t)header;
                scsi_command(SCSI_QUEUE_TAIL, cmd);
                if (cmd->scsi_status != SCSI_GOOD) {
			if (cmd->sense) scsi_sense_free(cmd->sense);
			cmd->sense = 0;
                } else {
                        info->sd_floppy = 1;
                        flex_page = (struct scsi_mode_flex_disk_param_page *)
                                ((unsigned char *) header + sizeof(*header) + header->block_descriptor_length);

			if (!header->block_descriptor_length)
				panic("sdinit: not block descriptor");

                        header->medium_type = 2;        /* MFM format (high density code) */
                        block->density_code = 2;        /* ditto */
                        block->number_of_blocks = 0;    /* Reset to zero */
                        block->block_length = 512;      /* Always 512 byte blocks */

                        /* do a mode select to set flex disk parameters */
#ifdef DBG
                        if (sddebug & SD_DEBUG_INIT) {
                                printf("sdinit: mode select to set flex\n");
                        }
#endif
                        SCSI_CMD6(cmd, SCSI_MODE_SELECT_6);
                        cmd->device = sd;
                        cmd->flags = SCSI_DATA_OUT | SCSI_WAIT;
                        cmd->data_len = sizeof(*header) + header->block_descriptor_length + SCSI_MODE_PAGE_HEADER_SIZE + flex_page->page_parameter_length;
                        ((struct mode_select_6_cmd *)(&cmd->cdb))->parameter_list_length = cmd->data_len;
                        cmd->data_addr = (vm_offset_t)header;

                        header->data_length = 0;
			/* DPOFUA is reserved on SELECT for DA/CDROM/OPTICAL */
			header->device_specific_params &= ~0x10;
                        flex_page->parameters_savable = 0;
                        flex_page->sectors_per_track = flex_page->sectors_per_track * flex_page->bytes_per_sector / 512; 
                        flex_page->bytes_per_sector = 512;

                        scsi_command(SCSI_QUEUE_TAIL, cmd);
                        if (cmd->scsi_status != SCSI_GOOD) {
                                if (sderror(cmd, info)) {
                                        uprintf("sd(%d,%d): failed to select mode (flex)\n", major(dev), minor(dev));
                                }
                                RETURN (EIO);
                        }
                }
        }

	if (error = sdinitlabel(dev))
		goto done;

        /* CD-ROM does not use bad block modes or write cache */
        if (inquiry->peripheral_device_type == SCSI_ROM_DEVICE) {
                info->sd_readonly = 1;
        } else {
		struct	drivedata	*dd;

                /* Temporary hack until all drives have new disk label */
		dd = (struct drivedata *)0;
		for (i = 0; i < NDDATA; i++) {
			if (info->sd_label.d_drivedata[i] != 0) {
				dd = (struct drivedata *)
					info->sd_label.d_drivedata;
				break;
			}			
		}

                /* do a mode sense to determine bad block modes */
#ifdef DBG
                if (sddebug & SD_DEBUG_INIT) {
                        printf("sdinit: mode sense for bad block modes\n");
                }
#endif
                SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
                cmd->device = sd;
                cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
		msp = (struct mode_sense_6_cmd *) &cmd->cdb;
                msp->pc = SCSI_MODE_SENSE_PC_CHANGEABLE;
                msp->page_code = SCSI_MODE_ERROR_RECOVERY_PAGE;
                cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
                msp->allocation_length = cmd->data_len;
                cmd->data_addr = (vm_offset_t)header;
                scsi_command(SCSI_QUEUE_TAIL, cmd);
                if (cmd->scsi_status != SCSI_GOOD) {
                        if (sderror(cmd, info)) {
#ifdef DBG
                                if (sddebug & SD_DEBUG_INIT)
                                        printf("sd(%d,%d): failed to sense mode (err)\n", BUS(unit), TGT(unit));
#endif
                        }
                        RETURN (EIO);
                }

                error_page = (struct scsi_mode_error_recovery_page *)
                        ((unsigned char *)header + sizeof(*header) + header->block_descriptor_length);

		/* save changeable parameters */
		changeable_error_page = *error_page;

		/* make sure we see good stuff */
#ifdef DBG
                if (sddebug & SD_DEBUG_INIT) {
                        printf("sdinit: mode sense for good stuff\n");
                }
#endif
		bzero(error_page, sizeof(*error_page));
                SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
                cmd->device = sd;
                cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
                msp->pc = SCSI_MODE_SENSE_PC_CURRENT;
                msp->page_code = SCSI_MODE_ERROR_RECOVERY_PAGE;
                cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
                msp->allocation_length = cmd->data_len;
                cmd->data_addr = (vm_offset_t)header;
                scsi_command(SCSI_QUEUE_TAIL, cmd);
                if (cmd->scsi_status != SCSI_GOOD) {
                        if (sderror(cmd, info)) {
#ifdef	DBG
                                if (sddebug & SD_DEBUG_INIT)
                                        printf("sd(%d,%d): failed to sense mode (err)\n", BUS(unit), TGT(unit));
#endif
                        }
                        RETURN (EIO);
                }

                if (header->device_specific_params & 0x80)
                        info->sd_readonly = 1;

                /* do a mode select to set bad block mode */
#ifdef DBG
                if (sddebug & SD_DEBUG_INIT) {
                        printf("sdinit: mode select to set bad block mode\n");
                }
#endif
                SCSI_CMD6(cmd, SCSI_MODE_SELECT_6);
                cmd->device = sd;
                cmd->flags = SCSI_DATA_OUT | SCSI_WAIT;
                cmd->data_len = sizeof(*header) + header->block_descriptor_length + SCSI_MODE_PAGE_HEADER_SIZE + error_page->page_parameter_length;
                ((struct mode_select_6_cmd *)(&cmd->cdb))->parameter_list_length = cmd->data_len;
                cmd->data_addr = (vm_offset_t)header;

                header->data_length = 0;
		/* DPOFUA is reserved on SELECT for DA/CDROM/OPTICAL */
		header->device_specific_params &= ~0x10;
                error_page->parameters_savable = 0;

                if (dd) {
		        /* Use data in drive label */
                        if (changeable_error_page.awre) error_page->awre = dd->awre;
                        if (changeable_error_page.arre) error_page->arre = dd->arre;
                        if (changeable_error_page.tb)   error_page->tb   = dd->tb;
                        if (changeable_error_page.rc)   error_page->rc   = dd->rc;
                        if (changeable_error_page.eer)  error_page->eer  = dd->eer;
                        if (changeable_error_page.per)  error_page->per  = dd->per;
                        if (changeable_error_page.dte)  error_page->dte  = dd->dte;
                        if (changeable_error_page.dcr)  error_page->dcr  = dd->dcr;
                        if (changeable_error_page.read_retry_count)          error_page->read_retry_count         = dd->read_retry_count;
                        if (changeable_error_page.correction_span)           error_page->correction_span          = dd->correction_span;
                        if (changeable_error_page.head_offset_count)         error_page->head_offset_count        = dd->head_offset_count;
                        if (changeable_error_page.data_strobe_offset_count)  error_page->data_strobe_offset_count = dd->data_strobe_offset_count;
		}
		else {
                        if (changeable_error_page.awre) error_page->awre = 1;
                        if (changeable_error_page.arre) error_page->arre = 1;
                        if (changeable_error_page.tb)   error_page->tb  = 0;
                        if (changeable_error_page.per)  error_page->per = 1;
                        if (changeable_error_page.dte)  error_page->dte = 0;
                        if (changeable_error_page.dcr)  error_page->dcr = 0;
                        if (changeable_error_page.read_retry_count) error_page->read_retry_count = 3;
                        if (info->sd_floppy) {
                                header->medium_type = 2;        /* MFM format (high density code) */
                                block->density_code = 2;        /* ditto */
                                block->number_of_blocks = 0;    /* Reset to zero */
                                block->block_length = 512;      /* Always 512 byte blocks */
                        }
                }

                scsi_command(SCSI_QUEUE_TAIL, cmd);
                if (cmd->scsi_status != SCSI_GOOD) {
                        if (sderror(cmd, info)) {
                                unsigned char   *ptr = (unsigned char *) cmd->data_addr;
#ifdef	DBG
                                if (sddebug & SD_DEBUG_INIT) {
                                        printf("sd(%d,%d): failed to select mode (err)\n", BUS(unit), TGT(unit));
                                        printf("DATA =  0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",  ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5], ptr[6], ptr[7]);
                                        printf("DATA =  0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",  ptr[8], ptr[9], ptr[10], ptr[11], ptr[12], ptr[13], ptr[14], ptr[15]);
                                        printf("DATA =  0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",  ptr[16], ptr[17], ptr[18], ptr[19], ptr[20], ptr[21], ptr[22], ptr[23]);
                                }
#endif
                        }
                        RETURN (EIO);
                }

		/* Assume write cache is present */
		info->sd_write_cache = TRUE;

		/* do a mode sense to determine if the drive supports write cache control */
#ifdef DBG
                if (sddebug & SD_DEBUG_INIT) {
                        printf("sdinit: mode sense for write cache\n");
                }
#endif
		SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
		cmd->device = sd;
		cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
		msp = (struct mode_sense_6_cmd *) &cmd->cdb;
                msp->pc = SCSI_MODE_SENSE_PC_CHANGEABLE;
		msp->page_code = SCSI_MODE_CACHE_CONTROL_PAGE;
		cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
		msp->allocation_length = cmd->data_len;
		cmd->data_addr = (vm_offset_t)header;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD) {
#ifdef	DBG
			if (sddebug & (SD_DEBUG_MODE|SD_DEBUG_INIT)) {
				(void) sderror(cmd, info);
			}
#endif
			if (cmd->scsi_status == SCSI_CHECK_CONDITION && cmd->sense->sense_key == SCSI_SENSE_ILLEGAL_REQUEST) {
				info->sd_write_cache = FALSE;
				scsi_sense_free(cmd->sense);
				cmd->sense = 0;
			}
		 	else if (sderror(cmd, info)) {
#ifdef	DBG
                                if (sddebug & SD_DEBUG_INIT)
                                        printf("sd(%d,%d): failed to sense changeable mode (cache)\n", BUS(unit), TGT(unit));
#endif
				RETURN (EIO);
			}
		}

	        cache_page = (struct scsi_mode_cache_control_page *)
			((unsigned char *)header + sizeof(*header) + header->block_descriptor_length);

		changeable_cache_page = *cache_page;

#ifdef DBG
                if (sddebug & SD_DEBUG_INIT) {
                        printf("sdinit: mode sense for write cache 2\n");
                }
#endif
		SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
		cmd->device = sd;
		cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
		msp->pc = SCSI_MODE_SENSE_PC_CURRENT;
		msp->page_code = SCSI_MODE_CACHE_CONTROL_PAGE;
		cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
		msp->allocation_length = cmd->data_len;
		cmd->data_addr = (vm_offset_t)header;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD) {
			if (cmd->scsi_status == SCSI_CHECK_CONDITION && cmd->sense->sense_key == SCSI_SENSE_ILLEGAL_REQUEST) {
				info->sd_write_cache = FALSE;
				scsi_sense_free(cmd->sense);
				cmd->sense = 0;
			}
		 	else if (sderror(cmd, info)) {
#ifdef DBG
                                if (sddebug & SD_DEBUG_INIT) {
                                        printf("sd(%d,%d): failed to sense current mode (cache)\n", BUS(unit), TGT(unit));
                                }
#endif
				RETURN (EIO);
			}
		}

		/* if the drive supports the cache control page */
		if (info->sd_write_cache) {
#ifdef DBG
                        if (sddebug & SD_DEBUG_INIT) {
                                printf("sdinit: mode select for cache control page\n");
                        }
#endif
			SCSI_CMD6(cmd, SCSI_MODE_SELECT_6);
			cmd->device = sd;
			cmd->flags = SCSI_DATA_OUT | SCSI_WAIT;
			cmd->data_len = sizeof(*header) + header->block_descriptor_length + SCSI_MODE_PAGE_HEADER_SIZE + cache_page->page_parameter_length;
			((struct mode_select_6_cmd *)(&cmd->cdb))->parameter_list_length = cmd->data_len;
			cmd->data_addr = (vm_offset_t)header;

			header->data_length = 0;
			/* DPOFUA is reserved on SELECT for DA/CDROM/OPTICAL */
			header->device_specific_params &= ~0x10;
			cache_page->parameters_savable = 0;

			if (changeable_cache_page.wce) {
				cache_page->wce = 1;
				scsi_command(SCSI_QUEUE_TAIL, cmd);
				if (cmd->scsi_status != SCSI_GOOD) {
#ifdef DBG
					if (sderror(cmd, info)) {
						printf("sd(%d,%d): failed to select mode (cache)\n", BUS(unit), TGT(unit));
					}
#endif
					RETURN (EIO);
				}
			}
			else if (!cache_page->wce) {
				info->sd_write_cache = FALSE;
			}
		}
#ifdef DBG
		if (!info->sd_write_cache)
			printf("sd(%d,%d): does not support write cache\n",
			       BUS(unit), TGT(unit));
#endif
        }

        /* if removeable media then lock in the media */
        if (inquiry->removable_media) {
#ifdef DBG
                if (sddebug & SD_DEBUG_INIT) {
                        printf("sdinit: prevent removal\n");
                }
#endif
                SCSI_CMD6(cmd, SCSI_PREVENT_REMOVAL);
                ((struct prevent_removal_cmd *)(&cmd->cdb))->prevent = 1;
                cmd->device = sd;
                cmd->flags = SCSI_WAIT;
                scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status == SCSI_GOOD)
			info->sd_media_locked = TRUE;
		else {
#ifdef DBG
			if (sddebug & SD_DEBUG_INIT)
				printf("sdinit: prevent removal failed");
			if (cmd->sense)
				if (sddebug & SD_DEBUG_INIT)
					printf(", sense key = 0x%x\n",
					       cmd->sense->sense_key);
#endif
			printf("sd%d: WARNING: media not locked in\n", unit);
		}
        }

#ifdef	DBG
	if (sddebug & SD_DEBUG_INIT) {
		printf("sdinit: sdinfo, floppy = %d, readonly = %d, write_cache = %d, label_valid = %d, media locked = %d\n",
			info->sd_floppy, info->sd_readonly, info->sd_write_cache, info->sd_label_valid, info->sd_media_locked);
	}
#endif

done:
	if (cmd->sense) scsi_sense_free(cmd->sense);
        scsi_command_free(cmd);
	kmem_free(kernel_map, data_buffer, size);

#ifndef	MACH_KERNEL
	if (error == 0)
		dk_wpms[info->sd_dkn] = 32 * (60 * DEV_BSIZE / 2); /* XXX */
#endif

#ifdef PDCDUMP
	/*
	 * XXX compute dump parameters for pdcdump
	 */
	{
		extern int dumpsize, dumpoffset;
		extern dev_t dumpdev;
		extern u_long bootdev;

		/*
		 * XXX PDC dump routine can only dump to the device we loaded
		 * the kernel from (bootdev).  Ensure that dumpdev is
		 * consistant with this, else we would compute a dump offset
		 * for one disk and then dump on the boot disk using that
		 * offset.
		 */
		if (scsi_dasd[SDUNIT(dumpdev)].bus != 0 ||
		    B_UNIT(bootdev) != scsi_dasd[SDUNIT(dumpdev)].target) {
			static int first = 1;

			if (first) {
				printf("WARNING: bad dump device (0x%x), must be on the boot device (0x%x)\n",
				       dumpdev,
				       makedev(major(dev), B_UNIT(bootdev)));
				first = 0;
			}
			dumpoffset = -1;
			probing = 1;
		}
		if (!probing && dumpoffset < 0 &&
		    error == 0 && SDPART0((int)dumpdev) == SDPART0((int)dev)) {
			struct disklabel *d = &info->sd_label;
			int part, size;

			part = SDPART(dumpdev);
			if (part < d->d_npartitions) {
				size = btoc(dbtob(d->d_partitions[part].p_size));
				if (dumpsize > size)
					dumpsize = size;
#ifndef MACH_KERNEL
				/*
				 * BSD: leave some slack at the beginning of
				 * the swap partition for the kernel's use.
				 * Otherwise the dump always gets corrupted.
				 */
				if (dumpsize == size) {
					extern long dumplo;

					dumpsize -= btoc(1*1024*1024);
					if (dumplo == 0)
						dumplo += btodb(1*1024*1024);
				}
#endif
				size = d->d_partitions[part].p_offset +
					d->d_partitions[part].p_size;
				dumpoffset = size - btodb(ctob(dumpsize));
			}
#ifdef DGB
			if (sddebug & SD_DEBUG_INIT)
				printf("sdinit: dev=0x%x dumpdev=0x%x dumpsize=0x%x dumpoffset=0x%x\n",
				       dev, dumpdev, dumpsize, dumpoffset);
#endif
#ifdef MACH_KERNEL
			dumpconf();
#endif
		}
	}
#endif
        return(error);

#undef	RETURN
}

#ifdef COMPAT_NOLABEL
struct sdpartition {
	char *sdp_name;
	struct partition sdp_part[MAXPARTITIONS];
} sddfltpart[] = {
	/* SEAGATE ST43400N 2.8Gb -- Robert special */
	"SEAGATE :2777Mb",
	{
		{ 0,       0	  }, /* a */
		{ 0,       0	  }, /* b */
		{ 5684196, 0	  }, /* c */
#ifdef OOPS
		{ 132072,  2776124}, /* d - swap */
		{ 2776000, 124    }, /* e - fs */
		{ 2776000, 2908196}, /* f - fs */
#else
		{ 0,       0	  }, /* d */
		{ 0,       0	  }, /* e */
		{ 0,       0	  }, /* f */
#endif
		{ 0,       0	  }, /* g */
		{ 0,       0	  }, /* h */
	},
	/* HP 97560 1.3Gb */
	"HP      :1292Mb",
	{
#ifdef UTAH750
		{ 49152,   4096	  }, /* a */
		{ 65536,   53248  }, /* b */
		{ 2647080, 0	  }, /* c */
		{ 132072,  53248  }, /* d */
		{ 1414186, 184320 }, /* e */
		{ 1048574, 1598506}, /* f */
		{ 365612,  184320 }, /* g */
		{ 2097148, 549932 }  /* h */
#else
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 2647080, 0      },
		{ 496684,  53248  },
		{ 1048574, 549932 },
		{ 1048574, 1598506},
		{ 2528296, 118784 },
		{ 2097148, 549932 }
#endif
	},
	/* Micropolis 1528 1280mb */
	"MICROP  :1280Mb",
	{
#if 1
		/* the more traditional style */
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 2621676, 0	  },
		{ 524288,  53248  },
		{ 131072,  577536 },
		{ 1913068, 708608 },
		{ 2502892, 118784 },
		{ 2044140, 577536 }
#else
		/* in the style of HP97560 above */
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 2621676, 0	  },
		{ 471280,  53248  },
		{ 1048574, 524528 },
		{ 1048574, 1573102},
		{ 2502892, 118784 },
		{ 2097148, 524528 }
#endif
	},
	/* Micropolis 1588-15 */
	"MICROP  :639Mb",
	{
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 1307340, 0      },
		{ 131072,  53248  },
		{ 262144,  184320 },
		{ 860876,  446464 },
		{ 1188556, 118784 },
		{ 1123020, 184320 }
	},
	/* Micropolis 1588T 660mb */
	"MICROP  :636Mb",
	{
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 1304052, 0      },
		{ 131072,  53248  },
		{ 262144,  184320 },
		{ 857588,  446464 },
		{ 1185268, 118784 },
		{ 1119732, 184320 }
	},
	/* HP 97548U 660mb */
	"HP      :633Mb",
	{
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 1296512, 0      },
		{ 131072,  53248  },
		{ 262144,  184320 },
		{ 850048,  446464 },
		{ 1177728, 118784 },
		{ 1112192, 184320 }
	},
	/* Quantum PD425S 400mb */
	"QUANTUM :406Mb",
	{
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 833150,  0      },
		{ 131072,  53248  },
		{ 184320,  184320 },
		{ 464510,  368640 },
		{ 714366,  118784 },
		{ 648830,  184320 }
	},
	/* HP C2235 400mb */
	"HP      :402Mb",
	{
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 825012,  0      },
		{ 131072,  53248  },
		{ 184320,  184320 },
		{ 456372,  368640 },
		{ 706228,  118784 },
		{ 640692,  184320 }
	},
	/* HP 2212S 335mb */
	"HP      :316Mb",
	{
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 648256,  0      },
		{ 131072,  53248  },
		{ 184320,  184320 },
		{ 279616,  368640 },
		{ 529472,  118784 },
		{ 463936,  184320 }
	},
	/* Quantum PD210S 200mb */
	"QUANTUM :199Mb",
	{
		{ 49152,   4096   },
		{ 65536,   53248  },
		{ 408574,  0      },
		{ 131072,  53248  },
		{ 184320,  184320 },
		{ 39934,   368640 },
		{ 289790,  118784 },
		{ 224254,  184320 }
	}
};
#define NSDDISKTYPES (sizeof(sddfltpart)/sizeof(sddfltpart[0]))
#endif

/*
 * Must have inquiry data before doing this
 */
sdinitlabel(dev)
dev_t dev;
{
        struct  sdinfo *info;
        struct  disklabel *d;
        struct  scsi_cmd                *cmd;
        struct  scsi_read_capacity      *capacity;
        struct  scsi_device             sd;
        int                             unit, i, error = 0;
	vm_size_t			size;

        unit = SDUNIT(dev);
        info = &sdinfo[unit];
        sd = scsi_dasd[unit];

        cmd = scsi_command_alloc();
        *cmd = nullcmd;

	size = (sizeof(*capacity)+(dcache_line_size-1))&~(dcache_line_size-1);
	capacity = 0;
	kmem_alloc(kernel_map, (vm_offset_t *)&capacity, size);

        /* find out the size of the disk */
#ifdef DBG
        if (sddebug & SD_DEBUG_INIT) {
                printf("sdinit: read capacity\n");
        }
#endif
        SCSI_CMD10(cmd, SCSI_READ_CAPACITY);
        cmd->device = sd;
        cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
        cmd->data_len = sizeof(*capacity);
        cmd->data_addr = (vm_offset_t) capacity;
        scsi_command(SCSI_QUEUE_TAIL, cmd);
        if (cmd->scsi_status != SCSI_GOOD) {
                if (sderror(cmd, info)) {
#ifdef DBG
                        if (sddebug & SD_DEBUG_INIT)
                                printf("sd(%d,%d): failed to provide disk size\n", BUS(unit), TGT(unit));
#endif
                }
		if (cmd->sense) scsi_sense_free(cmd->sense);
		scsi_command_free(cmd);
		kmem_free(kernel_map, (vm_offset_t)capacity, size);
		return(EIO);
        }

#ifdef  DBG
        if (sddebug & SD_DEBUG_CAPACITY)
                printf("sdinit: last block address = %d, block size = %d\n", capacity->logical_block_address, capacity->block_length);
#endif

        /* Clear out the disk label */
        bzero(&info->sd_label, sizeof(info->sd_label));

        /* Initialize the disklabel - from disk if possible, else a fake */
        d = &info->sd_label;
        d->d_secsize = DEV_BSIZE; /* Don't even *THINK* of using the REAL sector size - land mines everywhere */
        d->d_secperunit = ((capacity->logical_block_address+1)*capacity->block_length)/DEV_BSIZE;
        d->d_secpercyl = 1;     /* Just to avoid /0 in 'readdisklabel()' - what braindamage! */

        /* Shift factors - block-to-sector, byte count to sector count. */
        if (DEV_BSIZE >= capacity->block_length) {
                info->sd_btosdirection = SHIFTLEFT;
                for (info->sd_btosshift = 0, i = DEV_BSIZE/capacity->block_length; i > 1; i >>= 1) info->sd_btosshift++;
        }
        else {
                info->sd_btosdirection = SHIFTRIGHT;
                for (info->sd_btosshift = 0, i = capacity->block_length/DEV_BSIZE; i > 1; i >>= 1) info->sd_btosshift++;
        }
        for (info->sd_bctoscshift = 0, i = capacity->block_length; i > 1; i >>= 1) info->sd_bctoscshift++;

        /* Do not attempt to read disk label for CD-ROM */
        if ((info->sd_inquiry.peripheral_device_type == SCSI_ROM_DEVICE) ||
	    (sdreadlabel(dev, d, info))) {
                /*
                 * Build a fake label.
                 */
                struct  partition       *p;
                int     i;
		char    buf[10];
                char    *bp = buf;
                int     n;
                bzero(d, sizeof(*d));
                d->d_secperunit = ((capacity->logical_block_address+1)*capacity->block_length)/DEV_BSIZE;
                d->d_secsize = DEV_BSIZE;
                d->d_magic = d->d_magic2 = DISKMAGIC;
                d->d_type = DTYPE_SCSI;
                {
                        /*
                         * Build up a name of the form 'Vendor:nnnn Mb' from the known scsi values.
                         */
#define STRCPY(f, t) { bp = f; while(*bp) *(t)++ = *bp++; }
                        char    *to = d->d_typename;
                        int     n = (d->d_secsize * d->d_secperunit) / (1024 * 1024);
                        char    *dash = ":";
                        char    *mb = "Mb";
                        char    *bp;
			/* copy in vendor id.  its not nul terminated */
			bcopy(info->sd_inquiry.vendor_identification, to, sizeof(info->sd_inquiry.vendor_identification));
			to += sizeof(info->sd_inquiry.vendor_identification);
                        STRCPY(dash, to);
                        for(bp=buf;n;n/=10) *bp++ = "0123456789"[n%10];
                        while (bp > buf) *to++ = *--bp;
                        STRCPY(mb, to);
                        *to++ = '\0';
                }
#ifdef	MACH_KERNEL
                if (info->sd_inquiry.peripheral_device_type != SCSI_ROM_DEVICE &&
                    info->sd_floppy == 0) {
                        /* we do not expect a disk label for CD-ROM, dunk the warning */
			printf("sd%d: WARNING: no disk label\n", unit);
                }
#endif
                d->d_npartitions = MAXPARTITIONS;
                d->d_sbsize = SBSIZE;
                /*
                 * Simulate old behavior
                 */
                d->d_bbsize = 0;
#ifndef MACH_KERNEL
		d->d_nsectors = d->d_secperunit;
		d->d_ntracks = 1;
		d->d_ncylinders = 1;
		d->d_secpercyl = d->d_nsectors;
#endif
#ifdef COMPAT_NOLABEL
                /*
                 * Look up type in our table to get default partitions
                 */
                for (i = 0; i < NSDDISKTYPES; i++)
                        if (strcmp(sddfltpart[i].sdp_name, d->d_typename) == 0)
                                break;
                if (i < NSDDISKTYPES) {
                        struct sdpartition *sdp = &sddfltpart[i];

                        printf("sd%d: type '%s', using default partitions\n",
			       unit, d->d_typename);
                        for (i = 0; i < d->d_npartitions; i++) {
                                p = &d->d_partitions[i];
                                *p = sdp->sdp_part[i];
#ifndef MACH_KERNEL
				if (i == 1 || i == 3)
					p->p_fstype = FS_SWAP;
				else if (i == 2)
					p->p_fstype = FS_BOOT;
				else
#endif
                                p->p_fstype = FS_BSDFFS;
                                p->p_fsize = DEV_BSIZE;
                        }
                } else
#endif
		{
			if (info->sd_inquiry.peripheral_device_type != SCSI_ROM_DEVICE) {
				printf("sd%d: unrecognized type '%s', ",
				       unit, d->d_typename);
				printf("all partitions map entire %d %d-byte blocks\n",
				       d->d_secperunit, d->d_secsize);
			}
			for (i = 0; i < d->d_npartitions; i++) {
				p = &d->d_partitions[i];
				p->p_size = d->d_secperunit;
				p->p_offset = 0;
				p->p_fsize = DEV_BSIZE;
				p->p_fstype = FS_BSDFFS;
			}
		}
        }
	if (cmd->sense) scsi_sense_free(cmd->sense);
	scsi_command_free(cmd);
	kmem_free(kernel_map, (vm_offset_t)capacity, size);
	return(0);
}

#ifndef	MACH_KERNEL
#undef	uprintf
#endif

int
sdopen(dev, flag, fmt)
dev_t   dev;
int     flag;
int     fmt;
{
        struct  sdinfo  *info;
        struct  scsi_device sd;
        int     unit = SDUNIT(dev);
        int     part = SDPART(dev);
        int     error, s;
        unsigned long partmask;
        boolean_t inited;

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdopen(0x%x, 0x%x, 0x%x)\n", dev, flag, fmt);
#endif

        /* Make sure unit# is good, partition reasonable */
        if (unit >= MAXUNITS || part >= MAXPARTITIONS) {
                return (ENODEV);
        }

        error = 0;
        info = &sdinfo[unit];
	sd = scsi_dasd[unit];
#ifndef MACH_KERNEL
	/*
	 * XXX
	 * Make sure bus exists, otherwise we panic in scsi_command.
	 * Also make sure drive was found at static autoconf time (see
	 * glue.c), otherwise we might hang probing a non-existant EISA
	 * SCSI drive.
	 * XXX
	 */
	{
		extern struct scsi_controller *scsi_controller[];

		if (scsi_controller[sd.bus] == 0)
			return (ENODEV);
		if (!info->sd_probed)
			return (ENODEV);
	}
#endif

        s = splbio();

        inited = FALSE;
#ifdef  DBG
        if (sddebug & SD_DEBUG_OPEN)
                printf("sdopen[%d,%d]: sd_open=0x%x sd_probed=%d sd_label_valid=%d sd_removable_media=%d\n", unit, part, info->sd_open, info->sd_probed, info->sd_label_valid, info->sd_inquiry.removable_media);
#endif
	/*
	 * If the device is not open and is removable, not probed or has
	 * an invalid label then (re)read disklabel.
	 */
        if (!info->sd_open &&
	    (!info->sd_probed || !info->sd_label_valid || info->sd_inquiry.removable_media)) {

                /* avoid race with other openers */
                while (info->sd_init) {
                        sleep((caddr_t)info, PRIBIO);
                }

                /* initialize raw buf struct */
                bzero(&info->sd_raw, sizeof(info->sd_raw));
                info->sd_init = 1;              /* keep others out */
                error = sdinit(dev, 0);
                info->sd_init = 0;
                wakeup((caddr_t)info);
                if (error) {
                        goto open_done;
                }
                inited = TRUE;
        }

        /* Make sure partition exists */
        if (part >= info->sd_label.d_npartitions) {
                /* No such partition */
                error = ENODEV;
                goto open_done;
        }

        /* If he wants to write, make sure it's ok */
        if ((flag&FWRITE) && info->sd_readonly) {
                uprintf("sd(%d,%d): drive is write protected\n", major(dev), minor(dev));
                error = EROFS;

                if (inited) {
                        struct  scsi_cmd        *cmd = scsi_command_alloc();

                        /* if removeable media then unlock the media */
                        if (info->sd_inquiry.removable_media) {
                                SCSI_CMD6(cmd, SCSI_PREVENT_REMOVAL);
                                cmd->cdb.c6.length = 0;
                                cmd->device = sd;
                                cmd->flags = SCSI_WAIT;
                                scsi_command(SCSI_QUEUE_TAIL, cmd);
				if (cmd->scsi_status == SCSI_GOOD)
					info->sd_media_locked = FALSE;
                        }

                        /* RELEASE the drive */
                        SCSI_CMD6(cmd, SCSI_RELEASE);
                        cmd->flags = SCSI_WAIT;
                        cmd->device = sd;
                        scsi_command(SCSI_QUEUE_TAIL, cmd);

        		if (cmd->sense) scsi_sense_free(cmd->sense);
                        scsi_command_free(cmd);
                }

                goto open_done;
        }

        /* Track each partition for char and block devices */
        partmask =  (1 << part);
#ifndef MACH_KERNEL
        switch (fmt) {
        case S_IFCHR:
                info->sd_copen |= partmask;
                break;
        case S_IFBLK:
                info->sd_bopen |= partmask;
                break;
        }
#endif
        info->sd_open |= partmask;

open_done:
        splx(s);
        return(error);

}

int
sdclose(dev, flag, fmt)
dev_t   dev;
int     flag;
int     fmt;
{
        struct  sdinfo  *info;
        struct  scsi_device     sd;
        int     unit = SDUNIT(dev);
        int     part = SDPART(dev);
        int     error, s;
        u_int   partmask;

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdclose(0x%x, 0x%x, 0x%x)\n", dev, flag, fmt);
#endif

        /* Make sure unit# is good, partition reasonable */
        if (unit >= MAXUNITS || part >= MAXPARTITIONS) {
                return(ENODEV);
        }

        error = 0;
        info = &sdinfo[unit];
        sd = scsi_dasd[unit];

        s = splbio();

        if (!info->sd_open) {
                error = ENODEV;
                goto close_done;
        }

	/* don't check the partition cuz we could have changed the number of partitions */

        /* Track each partition for char and block devices */
        partmask =  (1 << part);
#ifndef	MACH_KERNEL
        switch (fmt) {
        case S_IFCHR:
                info->sd_copen &= ~partmask;
                break;
        case S_IFBLK:
                info->sd_bopen &= ~partmask;
                break;
        }
        info->sd_open = info->sd_bopen | info->sd_copen;
#else
	info->sd_open &= ~partmask;
#endif

        if (!info->sd_open) {

                struct  scsi_cmd        *cmd = scsi_command_alloc();

                /* if removeable media then unlock the media */
                if (info->sd_inquiry.removable_media) {
                        /* We don't care if we get an error */
                        SCSI_CMD6(cmd, SCSI_PREVENT_REMOVAL);
                        cmd->cdb.c6.length = 0;
                        cmd->device = sd;
                        cmd->flags = SCSI_WAIT;
                        scsi_command(SCSI_QUEUE_TAIL, cmd);
                }

                /* RELEASE the drive */
                SCSI_CMD6(cmd, SCSI_RELEASE);
                cmd->flags = SCSI_WAIT;
                cmd->device = sd;
                scsi_command(SCSI_QUEUE_TAIL, cmd);
                if ((cmd->scsi_status != SCSI_GOOD) && sderror(cmd, info)) {
                        panic("sdclose: unable to release device");
                }
		if (cmd->sense) scsi_sense_free(cmd->sense);
                scsi_command_free(cmd);
        }

close_done:
        splx(s);
        return(error);
}


int
sdread(dev, uio)
dev_t dev;
struct  uio     *uio;
{
	struct  buf     *bp;
        int     unit = SDUNIT(dev);
        int     part = SDPART(dev);
        int     retval;

#ifdef  DBG
	if (sddebug & SD_DEBUG_CALLS)
		printf("sdread(0x%x, 0x%x)\n", dev, uio);
#endif

        /* Make sure unit# is good, partition reasonable */
        if (unit >= MAXUNITS || part >= MAXPARTITIONS) {
                return(-1);
        }

	bp = &sdinfo[unit].sd_raw;
	retval = physio(sdstrategy, bp, dev, B_READ, scsi_minphys(scsi_dasd[unit]), uio);
	return(retval);
}


int
sdwrite(dev, uio)
dev_t dev;
struct	uio	*uio;
{
	struct  buf     *bp;
        int     unit = SDUNIT(dev);
        int     part = SDPART(dev);
        int     retval;

#ifdef  DBG
	if (sddebug & SD_DEBUG_CALLS)
		printf("sdwrite(0x%x, 0x%x)\n", dev, uio);
#endif
        /* Make sure unit# is good, partition reasonable */
        if (unit >= MAXUNITS || part >= MAXPARTITIONS) {
                return(-1);
        }

	bp = &sdinfo[unit].sd_raw;
	retval = physio(sdstrategy, bp, dev, B_WRITE, scsi_minphys(scsi_dasd[unit]), uio);
	return(retval);
}

#ifdef MACH_KERNEL
/*
 * Mach splits ioctls into those that write data into the kernel, and
 * those that read data out of the kernel (yep, very annoying).
 * Provide some glue around sdioctl().
 */

io_return_t
sdgetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	int error;

	error = sdioctl(dev, flavor, data, 0);
	*count = ((flavor >> 16) & IOCPARM_MASK) / sizeof(int);    /* XXX */
	return(error);
}

io_return_t
sdsetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;
	unsigned int	count;
{
	return(sdioctl(dev, flavor, data, FWRITE));
}
#endif

int
sdioctl(dev, op, arg, flag)
dev_t   dev;
int     op;
caddr_t arg;
int     flag;
{
        struct  scsi_device     sd;
        struct  scsi_cmd        *cmd;
        struct  sdinfo          *info;
        struct  disklabel       *d;
        int     unit = SDUNIT(dev);
        int     part = SDPART(dev);
        int     error = 0;
	u_long	multimask;

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdioctl(0x%x, 0x%x, 0x%x, 0x%x)\n", dev, op, arg, flag);
#endif

        /* Make sure unit# is good, partition reasonable */
        if (unit >= MAXUNITS || part >= MAXPARTITIONS) {
                return(ENODEV);
        }

        sd = scsi_dasd[unit];
        info = &sdinfo[unit];
        d = &info->sd_label;

	/* we will accept blank disklabels for the format command */
	if ((op != DIOCWFORMAT) && (part >= d->d_npartitions)) {
		return(ENODEV);
	}

	/*
	 * Determine if we have exclusive access to the disk so we can relax checking
	 * for disruptive disklabel changes and allow formatting of the disk.
	 */
	multimask = 0;
	if (info->sd_label_valid) {
	        u_int partmask = (1 << part);
		if (~partmask & info->sd_open)
			/* this disk is opened by someone else also */
			multimask = info->sd_open;
#ifndef MACH_KERNEL
		else if ((partmask & info->sd_bopen) && (partmask & info->sd_copen))
			/* our partition is opened twice, thus by someone else also */
			multimask = info->sd_open;
#endif
	}

        switch (op) {

                case DIOCGDINFO:
                        /* Get Disk Info - get disklabel for this unit */
                        *(struct disklabel *)arg = *d;
                        break;

#ifdef MACH_KERNEL
		/* XXX lame */
		case DIOCGDINFO - (0x10<<16):
			bcopy(d, arg, sizeof(struct disklabel) - 0x10);
			break;
#endif

                case DIOCGPART:
                        /* Get Partition - return pointer to disk label & partition */
                        /* Note: internal Only */
                        ((struct partinfo *)arg)->disklab = d;
                        ((struct partinfo *)arg)->part = &d->d_partitions[part];
                        break;

#ifdef MACH_KERNEL
		/* XXX lame */
		case DIOCSDINFO - (0x10<<16):
			bzero(&((struct disklabel *)arg)->d_partitions[MAXPARTITIONS], sizeof(struct partition));
			/* fall into... */
#endif
                case DIOCSDINFO:
                        /* Set disk info - accept new disk label - do not write to disk */
#ifndef MACH_KERNEL
			if (nsuser(u.u_procp->p_ucred, &u.u_acflag)) {
				error = EACCES;
				break;
			}
#endif
                        if ((flag & FWRITE) == 0)
                                error = EBADF;
                        else
                                error = setdisklabel(d, (struct disklabel *) arg, multimask);
                        break;

                case DIOCWLABEL:
                        /* Enable/disable writing into the label and bootstrap area of the disk */
#ifndef MACH_KERNEL
			if (nsuser(u.u_procp->p_ucred, &u.u_acflag)) {
				error = EACCES;
				break;
			}
#endif
                        if ((flag & FWRITE) == 0)
                                error = EBADF;
                        else
                                if (*(int *)arg)
                                        info->sd_bvolatile = 1;
                                else
                                        info->sd_bvolatile = 0;
                        break;

#ifdef MACH_KERNEL
		/* XXX lame */
		case DIOCWDINFO - (0x10<<16):
			bzero(&((struct disklabel *)arg)->d_partitions[MAXPARTITIONS], sizeof(struct partition));
			/* fall into... */
#endif
                case DIOCWDINFO:
                        /* Write disk info - accept new disk label - write to disk */
#ifndef MACH_KERNEL
			if (nsuser(u.u_procp->p_ucred, &u.u_acflag)) {
				error = EACCES;
				break;
			}
#endif
                        if ((flag & FWRITE) == 0)
                                error = EBADF;
                        else if ((error = setdisklabel(d, (struct disklabel *)arg, multimask)) == 0) {
                                int     bv = info->sd_bvolatile;
                                info->sd_bvolatile = 1;         /* Have to make sure it's unlocked */
                                error = writedisklabel(SDLABPART(dev), sdstrategy, d);
                                info->sd_bvolatile = bv;        /* Then return it to previous state */
                                info->sd_label_valid = 0;       /* Force reread */
                        }
                        break;

#ifdef	MACH_KERNEL
                case DIOCFORMAT:
                        cmd = scsi_command_alloc();
                        *cmd = nullcmd;
                        SCSI_CMD6(cmd, SCSI_FORMAT_UNIT);
                        cmd->flags = SCSI_WAIT;
                        cmd->device = sd;
                        scsi_command(SCSI_QUEUE_TAIL, cmd);
                        if (cmd->scsi_status != SCSI_GOOD) {
                                sderror(cmd,sd);
                                /* a quick check to see if drive is ready */
                                if (cmd->sense && cmd->sense->sense_key == SCSI_SENSE_NOT_READY) {
                                        uprintf("sd(%d,%d): unit not ready\n", major(dev), minor(dev));
                                }
                                else if (cmd->sense) {
                                        uprintf("sd(%d,%d): failed to format drive\n", major(dev), minor(dev));
                                }
				if (cmd->sense) scsi_sense_free(cmd->sense);
                                scsi_command_free(cmd);
                                return (EIO);
                        }
                        scsi_command_free(cmd);
                        break;
#endif

                default:
                        error = ENOTTY;
                        break;
        }
        return (error);
}


int
sdsize(dev)
dev_t   dev;
{
        struct  sdinfo  *info;
        int     unit = SDUNIT(dev);
        int     part = SDPART(dev);

#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdsize(0x%x)\n", dev);
#endif

        /* Make sure unit# is good, partition reasonable */
        if (unit >= MAXUNITS || part >= MAXPARTITIONS) {
                return(-1);
        }

        info = &sdinfo[unit];

        if (part >= info->sd_label.d_npartitions) {
                return(-1);
        }

        return(info->sd_label.d_partitions[part].p_size);
}


/*
 * sdmakedev - called when RB_ASKME is used from setconf()
 * Given a device address return a dev_t for it
 */
dev_t
sdmakedev(maj, ctrlr, unit, part)
int     maj;
int     ctrlr;
int     unit;
int     part;
{
        /* XXX Only works for ctlr = 0 */
#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdmakedev(0x%x, 0x%x, 0x%x, 0x%x)\n", maj, ctrlr, unit, part);
#endif
        return(makedev(maj, SDMAKEDEV(unit, part)));
}

/*
 * sdbreakdev - break up a dev_t into its parts.
 */
sdbreakdev(dev, maj, ctrlr, unit, part)
dev_t   dev;
int     *maj;
int     *ctrlr;
int     *unit;
int     *part;
{
#ifdef  DBG
        if (sddebug & SD_DEBUG_CALLS)
                printf("sdbreakdev(0x%x, 0x%x, 0x%x, 0x%x, 0x%x)\n", dev, maj, ctrlr, unit, part);
#endif
        *maj = major(dev);
        *ctrlr = scsi_dasd[*unit = SDUNIT(dev)].bus;
        *part = SDPART(dev);
        return;
}

sddump(bp)
        struct buf *bp;
{
        bp->b_flags |= B_ERROR;
        bp->b_error = EIO;
}

#ifdef DBG
sdshowgeometry(unit)
	int unit;
{
	vm_offset_t data_buffer;
	vm_size_t size;
        struct scsi_cmd *cmd;
	struct	mode_sense_6_cmd *msp;
        struct scsi_mode1_header *header;
	struct scsi_mode_notch_page *notch_page;
	struct scsi_mode_rigid_disk_geometry_page *rigid_page;
	int i, maxnotch, blocks, tblocks, hd, cyl, shd, scyl, ehd, ecyl;
	int tracks, ttracks, changelpn;

	size = MAX(MAX(sizeof(*rigid_page), sizeof(*notch_page)),
		   SCSI_MODE_SENSE_MAXLEN);
	size = (size + (dcache_line_size-1)) & ~(dcache_line_size-1);

	data_buffer = 0;
        kmem_alloc(kernel_map, &data_buffer, size);
        header = (struct scsi_mode1_header *) data_buffer;
	cmd = scsi_command_alloc();
	*cmd = nullcmd;

	SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
	cmd->device = scsi_dasd[unit];
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	msp = (struct mode_sense_6_cmd *) &cmd->cdb;
	msp->pc = SCSI_MODE_SENSE_PC_CURRENT;
	msp->page_code = SCSI_MODE_RIGID_DISK_GEOMETRY_PAGE;
	cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
	msp->allocation_length = cmd->data_len;
	cmd->data_addr = (vm_offset_t) header;
	scsi_command(SCSI_QUEUE_TAIL, cmd);
	if (cmd->scsi_status != SCSI_GOOD) {
		if (sderror(cmd, &sdinfo[unit])) {
			printf("sd(%d,%d): failed to sense mode (rigid)\n",
			       BUS(unit), TGT(unit));
		}
		goto done;
	}
	rigid_page = (struct scsi_mode_rigid_disk_geometry_page *)
		((unsigned char *)header + sizeof(*header) + header->block_descriptor_length);
	cyl = SCSI_24BIT_FIELD(rigid_page, number_of_cylinders);
	hd = rigid_page->number_of_heads;
	printf("sd%d: bus=%d, target=%d, ",
	       unit, scsi_dasd[unit].bus, scsi_dasd[unit].target);
	printf("cyl=%d, hds=%d, rpm=%d", cyl, hd, rigid_page->rotation_rate);

	/*
	 * First find out if we can change the LPN bit
	 */
	SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
	cmd->device = scsi_dasd[unit];
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	msp = (struct mode_sense_6_cmd *) &cmd->cdb;
	msp->pc = SCSI_MODE_SENSE_PC_CHANGEABLE;
	msp->page_code = SCSI_MODE_NOTCH_PAGE;
	cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
	msp->allocation_length = cmd->data_len;
	cmd->data_addr = (vm_offset_t) header;
	scsi_command(SCSI_QUEUE_TAIL, cmd);
	if (cmd->scsi_status != SCSI_GOOD) {
		if (cmd->sense) {
			scsi_sense_free(cmd->sense);
			cmd->sense = 0;
		}
		changelpn = 0;
	} else {
		notch_page = (struct scsi_mode_notch_page *)
			((unsigned char *)header + sizeof(*header) +
			 header->block_descriptor_length);
		changelpn = notch_page->logical_notch;
	}

	/*
	 * Now get the current notch info for the drive
	 */
	SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
	cmd->device = scsi_dasd[unit];
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	msp = (struct mode_sense_6_cmd *) &cmd->cdb;
	msp->pc = SCSI_MODE_SENSE_PC_CURRENT;
	msp->page_code = SCSI_MODE_NOTCH_PAGE;
	cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
	msp->allocation_length = cmd->data_len;
	cmd->data_addr = (vm_offset_t) header;
	scsi_command(SCSI_QUEUE_TAIL, cmd);
	if (cmd->scsi_status != SCSI_GOOD) {
		if (sderror(cmd, &sdinfo[unit])) {
			printf("sd(%d,%d): failed to sense mode (notch)\n",
			       BUS(unit), TGT(unit));
		}
		goto done;
	}
	notch_page = (struct scsi_mode_notch_page *)
		((unsigned char *)header + sizeof(*header) +
		 header->block_descriptor_length);

	if (!notch_page->notched_drive) {
		printf("\n");
		goto done;
	}
	maxnotch = notch_page->maximum_notches;
	printf(", notches=%d:\n", maxnotch);

	tblocks = ttracks = 0;
	for (i = 1; i <= maxnotch; i++) {
		if (changelpn) {
			/*
			 * Select the notch in logical (block) mode
			 */
			notch_page->active_notch = i;
			notch_page->logical_notch = 1;
			SCSI_CMD6(cmd, SCSI_MODE_SELECT_6);
			cmd->device = scsi_dasd[unit];
			cmd->flags = SCSI_DATA_OUT | SCSI_WAIT;
			cmd->data_len = sizeof(*header) +
				header->block_descriptor_length +
					SCSI_MODE_PAGE_HEADER_SIZE +
						notch_page->page_parameter_length;
			((struct mode_select_6_cmd *)(&cmd->cdb))->parameter_list_length = cmd->data_len;
			cmd->data_addr = (vm_offset_t)header;
			header->data_length = 0;
			header->device_specific_params &= ~0x10;
			notch_page->parameters_savable = 0;
			scsi_command(SCSI_QUEUE_TAIL, cmd);
			if (cmd->scsi_status != SCSI_GOOD) {
				if (sderror(cmd, &sdinfo[unit]))
					printf("    notch %d: select1 failed\n", i);
				if (cmd->sense) {
					scsi_sense_free(cmd->sense);
					cmd->sense = 0;
				}
			}

			/*
			 * Get info for that notch
			 */
			SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
			cmd->device = scsi_dasd[unit];
			cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
			msp = (struct mode_sense_6_cmd *) &cmd->cdb;
			msp->pc = SCSI_MODE_SENSE_PC_CURRENT;
			msp->page_code = SCSI_MODE_NOTCH_PAGE;
			cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
			msp->allocation_length = cmd->data_len;
			cmd->data_addr = (vm_offset_t) header;
			scsi_command(SCSI_QUEUE_TAIL, cmd);
			if (cmd->scsi_status != SCSI_GOOD) {
				if (sderror(cmd, &sdinfo[unit]))
				printf("    notch %d: sense1 failed\n", i+1);
				if (cmd->sense) {
					scsi_sense_free(cmd->sense);
					cmd->sense = 0;
				}
			}
			blocks = notch_page->ending_boundary -
				notch_page->starting_boundary + 1;
		} else
			blocks = 0;

		/*
		 * Reselect the notch in physical mode
		 */
		notch_page->active_notch = i;
		if (changelpn)
			notch_page->logical_notch = 0;
		SCSI_CMD6(cmd, SCSI_MODE_SELECT_6);
		cmd->device = scsi_dasd[unit];
		cmd->flags = SCSI_DATA_OUT | SCSI_WAIT;
		cmd->data_len = sizeof(*header) +
			header->block_descriptor_length +
				SCSI_MODE_PAGE_HEADER_SIZE +
					notch_page->page_parameter_length;
		((struct mode_select_6_cmd *)(&cmd->cdb))->parameter_list_length = cmd->data_len;
		cmd->data_addr = (vm_offset_t)header;
		header->data_length = 0;
		header->device_specific_params &= ~0x10;
		notch_page->parameters_savable = 0;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD) {
			if (sderror(cmd, &sdinfo[unit]))
				printf("    notch %d: select2 failed\n", i+1);
			if (cmd->sense) {
				scsi_sense_free(cmd->sense);
				cmd->sense = 0;
			}
		}

		/*
		 * Reacquire info for that notch
		 */
		SCSI_CMD6(cmd, SCSI_MODE_SENSE_6);
		cmd->device = scsi_dasd[unit];
		cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
		msp = (struct mode_sense_6_cmd *) &cmd->cdb;
		msp->pc = SCSI_MODE_SENSE_PC_CURRENT;
		msp->page_code = SCSI_MODE_NOTCH_PAGE;
		cmd->data_len = SCSI_MODE_SENSE_MAXLEN;
		msp->allocation_length = cmd->data_len;
		cmd->data_addr = (vm_offset_t) header;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD) {
			if (sderror(cmd, &sdinfo[unit]))
				printf("    notch %d: sense2 failed\n", i+1);
			if (cmd->sense) {
				scsi_sense_free(cmd->sense);
				cmd->sense = 0;
			}
		}
		printf("    notch %d: ", notch_page->active_notch);
		if (notch_page->logical_notch) {
			blocks = notch_page->ending_boundary -
				notch_page->starting_boundary + 1;
			printf("[%d-%d]: 0x%x/0x%x\n",
			       notch_page->starting_boundary,
			       notch_page->ending_boundary,
			       notch_page->pages_notched_hi,
			       notch_page->pages_notched_lo);
			tracks = 0;
		} else {
			scyl = notch_page->starting_boundary >> 8;
			shd = notch_page->starting_boundary & 0xFF;
			ecyl = notch_page->ending_boundary >> 8;
			ehd = notch_page->ending_boundary & 0xFF;
			printf("[%d/%d-%d/%d]: ", scyl, shd, ecyl, ehd);
			tracks = 0;
			if (shd != 0) {
				tracks += (hd - shd);
				shd = 0;
				scyl++;
			}
			if (ehd != (hd - 1)) {
				tracks += (ehd + 1);
				ehd = hd - 1;
				ecyl--;
			}
			tracks += (ecyl - scyl + 1) * hd;
			printf("%d tracks at %d blocks per track",
			       tracks, blocks / tracks);
			if (blocks % tracks)
				printf(" [not exact: %d blocks, %d tracks]",
				       blocks, tracks);
			printf("\n");
		}
		tblocks += blocks;
		ttracks += tracks;
	}
	printf("%d blocks in %d tracks\n", tblocks, ttracks);
	/*
	 * Re-select notch 0 (entire unit)
	 */
	notch_page->active_notch = 0;
	SCSI_CMD6(cmd, SCSI_MODE_SELECT_6);
	cmd->device = scsi_dasd[unit];
	cmd->flags = SCSI_DATA_OUT | SCSI_WAIT;
	cmd->data_len = sizeof(*header) +
		header->block_descriptor_length +
			SCSI_MODE_PAGE_HEADER_SIZE +
				notch_page->page_parameter_length;
	((struct mode_select_6_cmd *)(&cmd->cdb))->parameter_list_length =
		cmd->data_len;
	cmd->data_addr = (vm_offset_t)header;
	header->data_length = 0;
	header->device_specific_params &= ~0x10;
	notch_page->parameters_savable = 0;
	scsi_command(SCSI_QUEUE_TAIL, cmd);

done:
	if (cmd->sense)
		scsi_sense_free(cmd->sense);
	scsi_command_free(cmd);
}
#endif

#endif /* NSCSI > 0 */
