/*	$CHeader: kgdb_stub.c 1.6 93/04/08 10:41:39 $ */

/*
 * Mach Operating System
 * Copyright (c) 1993, CONVEX Computer Corporation.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CONVEX ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION,
 * AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 */
/*
 * Machine-independent stub for remote kernel gdb.
 *
 * Known problem areas are marked with "FIXME".
 */

#ifdef KGDB_LAN
#include <kern/task.h>
#include <mach/boolean.h>
#include <sys/reboot.h>
#include <machine/gdb_packet.h>
#include <machine/kgdb.h>
#include <machine/setjmp.h>
#include <norma/ipc_ether.h>

jmp_buf_t *kgdb_recover = 0;

boolean_t kgdb_initialized = FALSE;
boolean_t kgdb_debug_init = TRUE;	/* != 0 waits for host at bootstrap */
boolean_t kgdb_connected = FALSE;	/* do we know where responses go? */
boolean_t kgdb_active = FALSE;		/* is kgdb polling? */
boolean_t kgdb_stop = FALSE;		/* should we stop in ether intr()? */

int kgdb_enabled = 0;		/* temp hack for hpdev/dca.c */

/*
 * kgdb_debug_flags contains several bit fields for debugging:
 *
 *    1	kgdb_packet_input (dropped packets)
 *    2 kgdb_packet_input (packets kept)
 *    4 kgdb_get_request
 *    8 kgdb_send_packet
 *   16	kgdb_response
 *   32	kgdb_request
 *   64 kgdb_trap
 *  128	kgdb_bcopy (memory faults)
 */
unsigned int kgdb_debug_flags = 0;

static void
kgdb_response(gdb_response_t response, unsigned int data_size, char *data);

static void
kgdb_debug(unsigned int flag, char *format, int arg)
{
	if (kgdb_debug_flags & flag)
		printf(format, arg);
}

/*
 * This routine is called by the mux initialization code as soon as the
 * mux is configured.
 */
kgdb_mux_init()
{
	extern int bootargc;	/* model_dep.c */
	extern int boothowto;

	/* XXX for HP-UX boot program */
	if (bootargc > 1)
		boothowto |= RB_KDB;

	kgdb_net_init();
	kgdb_initialized = TRUE;
	if (boothowto & RB_KDB) {
		boothowto &= ~RB_KDB;

		if (kgdb_debug_init) {
			printf("kgdb waiting...");
			kgdb_break();
			printf("connected.\n");
		} else
			printf("kgdb enabled\n");
	}
}

/*
 * Do a (slow) bcopy.  If we called the normal kernel bcopy, we would
 * get into trouble if we ever tried to step through it.
 */
static void
kgdb_bcopy(char *source, char *destination, unsigned int size)    
{
	while (size-- > 0) {
		*destination++ = *source++;
	}
}
	
/*
 * Do a (slow) bcopy, catching any VM faults and returning an error.
 */
static boolean_t
kgdb_safe_bcopy(char *source, char *destination, unsigned int size)
{
	jmp_buf_t jmpbuf;
	jmp_buf_t *prev_jmpbuf;

	if (_setjmp(kgdb_recover = &jmpbuf)) {
		kgdb_debug(128, "kgdb_safe_bcopy: memory fault\n", 0);
		kgdb_recover = prev_jmpbuf;
		return FALSE;
	}

	while (size-- > 0) {
		*destination++ = *source++;
	}

	kgdb_recover = prev_jmpbuf;
	return TRUE;
}

#define htons(s)	(s)
#define ntohs(s)	(s)
#define htonl(l)	(l)
#define ntohl(l)	(l)

/* Local and remote address information (in network byte order) */

typedef struct {
	char ether[6];
	char ip[4];
	char port[2];
} net_addr;

static net_addr kgdb_local_addr;
static net_addr kgdb_foreign_addr;

static boolean_t 	    kgdb_request_ready = FALSE;
static boolean_t 	    kgdb_attach_seen = FALSE;
static struct gdb_request   kgdb_request_buffer;
static struct gdb_response  kgdb_response_buffer;
static struct gdb_response  kgdb_previous_response_buffer;
static unsigned int 	    kgdb_previous_xid = 0;
static unsigned int 	    kgdb_previous_response_size = 0;

/*
 * Handle an incoming packet from the network.  Called from network
 * driver interrupt handler.  The driver drops everything except
 * debugger packets before calling here.
 */
boolean_t
kgdb_packet_input(char *buffer, unsigned int size)
{
        netipc_ether_header_t ehp = (netipc_ether_header_t) buffer;
	netipc_udpip_header_t uhp = (netipc_udpip_header_t) (ehp + 1);
	char *dp = (char *) (uhp + 1);
	unsigned int request_size;
	gdb_request_t request = &kgdb_request_buffer;

	/*
	 * If we're not initialized or we are already working on a packet,
	 * drop this one.
	 */
	if (!kgdb_initialized || kgdb_request_ready) {
		kgdb_debug(1, "kgdb_packet_input: dropped (not initialized)\n",
			   0);
		return FALSE;
	}

	/*
	 * Copy packet to incoming buffer
	 */
	request_size = ntohs(uhp->udp_length) - 8;
	kgdb_bcopy(dp, (char *) &kgdb_request_buffer, request_size);

	/*
	 * If it's not for the microkernel stub (task != 0) and the
	 * microkernel stub is not active, pass it up to the task stub.
	 */
	if (ntohl(request->task_id) != 0) {
		if (kgdb_active) {
			return TRUE;	/* we're active; swallow it */
		}
		return FALSE;
	}


	if (ntohl(request->request) == 'a') {
	  kgdb_attach_seen =  TRUE;
	}

	/*
	 * Save all the address stuff for the response
	 */
	kgdb_bcopy((char *) ehp->e_src, kgdb_foreign_addr.ether, 6);
	kgdb_bcopy((char *) ehp->e_dest, kgdb_local_addr.ether, 6);
	kgdb_bcopy((char *) &uhp->ip_src, kgdb_foreign_addr.ip, 4);
	kgdb_bcopy((char *) &uhp->ip_dst, kgdb_local_addr.ip, 4);
	kgdb_bcopy((char *) &uhp->udp_source_port, kgdb_foreign_addr.port, 2);
	kgdb_bcopy((char *) &uhp->udp_dest_port, kgdb_local_addr.port, 2);
	
	if (kgdb_attach_seen) {
	  /*
	   * If this is just a request for status, drop it unless our
	   * status has changed, i.e., we've taken a breakpoint or something.
	   */
	  if (!kgdb_active && ntohl(request->request) == 'w') {
		kgdb_debug (1, "kgdb_packet_input: dropped (no status change)\n",
			    0);
		return TRUE;
	  }
	}
	else {
	  if (ntohl(request->request) == 'w') {
	      kgdb_response_buffer.xid = request->xid;
	      kgdb_response_buffer.return_code = -1;
	      kgdb_response(&kgdb_response_buffer, 0, 0);
	      return TRUE;
	  }
        }

	if (!kgdb_active) {
		kgdb_stop = TRUE;	/* make driver stop later */
	}

	kgdb_debug(2, "kgdb_packet_input: %d bytes received\n", request_size);

	kgdb_connected = TRUE;
	kgdb_request_ready = TRUE;

	return TRUE;
}

/*
 * Take care of all the grungy network packet details and send the packet
 * to the remote machine.
 */
static void
kgdb_send_packet(char *buffer, unsigned int size)
{
	static char packet_buffer[ETHERMTU];
	struct netipc_ether_header eh;
	struct netipc_udpip_header ih;
	netipc_ether_header_t ehp = (netipc_ether_header_t) packet_buffer;
	netipc_udpip_header_t uhp = (netipc_udpip_header_t) (ehp + 1);
	char *dp = (char *) (uhp + 1);
	int i, total_length, udp_length;
	unsigned long checksum;
	
	kgdb_debug(8, "kgdb_send_packet: %d bytes\n", size);

	udp_length = UDP_OVERHEAD + size;
	total_length = EHLEN + IP_OVERHEAD + udp_length;
	
	kgdb_bcopy(kgdb_foreign_addr.ether, (char *) eh.e_dest, 6);
	kgdb_bcopy(kgdb_local_addr.ether, (char *) eh.e_src, 6);
	eh.e_ptype = htons(ETHERTYPE_IP);
	
	ih.ip_version = IPVERSION;
	ih.ip_header_length = 5;
	ih.ip_type_of_service = 0;
	ih.ip_total_length = htons(udp_length + IP_OVERHEAD);
	ih.ip_id = htons(0);
	ih.ip_fragment_offset = htons(0);
	ih.ip_time_to_live = 0xff;
	ih.ip_protocol = UDP_PROTOCOL;
	ih.ip_checksum = 0;
	
	kgdb_bcopy(kgdb_local_addr.ip, (char *) &ih.ip_src, 4);
	kgdb_bcopy(kgdb_foreign_addr.ip, (char *) &ih.ip_dst, 4);
	
	for (checksum = i = 0; i < 10; i++) {
		checksum += ((unsigned short *) &ih)[i];
	}
	checksum = (checksum & 0xffff) + (checksum >> 16);
	checksum = (checksum & 0xffff) + (checksum >> 16);
	ih.ip_checksum = (~checksum & 0xffff);
	
	kgdb_bcopy(kgdb_local_addr.port, (char *) &ih.udp_source_port, 2);
	kgdb_bcopy(kgdb_foreign_addr.port, (char *) &ih.udp_dest_port, 2);
	ih.udp_length = htons(udp_length);
	ih.udp_checksum = 0;

	kgdb_bcopy((char *) &eh, (char *) ehp, sizeof(eh));
	kgdb_bcopy((char *) &ih, (char *) uhp, sizeof(ih));
	kgdb_bcopy(buffer, dp, size);
	
	while (total_length < 64) {
		packet_buffer[total_length] = 0;
		total_length++;
	}

	kgdb_net_xmt(packet_buffer, total_length);
}

/*
 * Wait (forever) until a request for debugging service is received.
 */
static gdb_request_t
kgdb_get_request()
{
	kgdb_debug(4, "kgdb_get_request:\n", 0);
	while (TRUE) {
		gdb_request_t request = &kgdb_request_buffer;
		
		while (!kgdb_request_ready) {
			kgdb_net_intr();
		}

		request->xid = ntohl(request->xid);
		request->type = ntohl(request->type);
		request->task_id = ntohl(request->task_id);
		request->thread_id = ntohl(request->thread_id);
		request->request = ntohl(request->request);
		request->address = ntohl(request->address);
		request->size = ntohl(request->size);

		if (request->type != 0) {
			kgdb_debug(4, "kgdb_get_request: non-request (%x) dropped\n",
			       request->type);
			continue;
		}

		/*
		 * Check to see whether this is a duplicate of the
		 * last request received.  If so, retransmit the
		 * previous response.
		 */
		if (request->xid == kgdb_previous_xid &&
		    kgdb_previous_response_size != 0) {
		    	kgdb_debug(4, "kgdb_get_request: xid %08x, resending last response\n",
				   request->xid);
			kgdb_send_packet((char *)
					 &kgdb_previous_response_buffer,
					kgdb_previous_response_size);
			delay(200000); /* don't flood the net! */
			continue;
		}
		
		kgdb_debug(4, "kgdb_get_request: received request %08x\n",
			   request->xid);
		kgdb_previous_xid = request->xid;
		kgdb_previous_response_size = 0;
		return(request);
	}
}

/*
 * Copy any desired data into the response and send it.
 */
static void
kgdb_response(gdb_response_t response, unsigned int data_size, char *data)
{
	unsigned int size;

	if (data_size > 0) {
		if (!kgdb_safe_bcopy(data, response->data, data_size)) {
			response->return_code = 1;
		}
	}

	kgdb_debug(16, "kgdb_response: response for request %08x\n",
		   response->xid);

	response->xid = htonl(response->xid);
	response->type = htonl(1); /* response */
	response->return_code = htonl(response->return_code);
	response->size = htonl(data_size);

	size = sizeof(*response) - sizeof(response->data) + data_size;
	kgdb_bcopy((char *) response, (char *) &kgdb_previous_response_buffer,
		   size);
	kgdb_previous_response_size = size;
	kgdb_send_packet((char *) response, size);
	kgdb_request_ready = FALSE; /* now we can handle another request */
}

static char kgdb_task_info_buffer[RESPONSE_DATA_SIZE];
static unsigned int kgdb_task_info_size;

static void
kgdb_task_info()
{
	processor_set_t pset;
	task_t task;
	unsigned int *dp;

	dp = (unsigned int *) kgdb_task_info_buffer;
	kgdb_task_info_size = 0;

	if (queue_first(&all_psets) == 0) {
		return;
	}
	queue_iterate(&all_psets, pset, processor_set_t, all_psets) {
		if (queue_first(&pset->tasks) == 0) {
			continue;
		}
		queue_iterate(&pset->tasks, task, task_t, pset_tasks) {
			*dp++ = (unsigned int) task;
			*dp++ = (unsigned int) task->map;
			*dp++ = 0/*XXXtask->user_stop_count*/;
			*dp++ = 0/*XXXtask->thread_count*/;
			kgdb_task_info_size += sizeof(unsigned int) * 4;
			if (kgdb_task_info_size >= RESPONSE_DATA_SIZE) {
				return;
			}
		}
	}
}

/*
 * Perform requested action and send a response.
 */
static boolean_t
kgdb_request(gdb_request_t request, gdb_response_t response, int type,
	     char *ssp)
{
	int signal = 5;

	kgdb_debug(32, "kgdb_request: request %08x ", request->xid);
	kgdb_debug(32, "(%c)\n", request->request);
	response->xid = request->xid;
	response->return_code = 0;
	switch (request->request) {
	case 'w':
	case 'a':
	case '?':
		kgdb_response(response, sizeof(signal), (char *) &signal);
		break;
	    
	case 'g':		/* read registers */
		kgdb_response(response, REGISTER_BYTES, ssp);
		break;
	    
	case 'G':		/* write registers */
		kgdb_bcopy(response->data, ssp, REGISTER_BYTES);
		kgdb_response(response, 0, 0);
		break;
	    
	case 'M':		/* write memory */
		if (request->size <= 512) {
			if (!kgdb_safe_bcopy(request->data,
					     (char *) request->address,
					     request->size)) {
				response->return_code = 1;
			}
			kgdb_flush_cache(request->address, request->size);
		} else {
			printf("bad size %d\n", request->size);
			response->return_code = 3;
		}
		kgdb_response(response, 0, 0);
		break;

	case 'm':		/* read memory */
		if (request->size <= 512) {
			kgdb_response(response, request->size,
				      (char *) request->address);
		} else {
		        printf("bad size %d\n", request->size);
			response->return_code = 3;
			kgdb_response(response, 0, 0);
		}
		break;
	    
	case 'c':		/* continue */
		kgdb_continue(type, ssp, request->address);
		kgdb_response(response, 0, 0);
		return TRUE;

	case 'C':		/* detach and continue */
		kgdb_continue(type, ssp, (vm_offset_t)0);
		kgdb_response(response, 0, 0);
		return TRUE;

	case 's':		/* single step */
		kgdb_single_step(type, ssp, request->address);
		kgdb_response(response, 0, 0);
		return TRUE;
	    
	case 'k':		/* kill kernel (reboot) */
		printf("kgdb: rebooting\n");
		kgdb_response(response, 0, 0);
		fastboot(0);	/* reboot; should never return */
		return TRUE;
	case 'T':
		kgdb_task_info();
		kgdb_response(response, kgdb_task_info_size,
			      kgdb_task_info_buffer);
		return FALSE;
	default:
		printf("kgdb_request: unknown request %c\n", request->request);
		response->return_code = 1;
		kgdb_response(response, 0, 0);
		break;
	}
	
	return FALSE;	/* not finished; wait for additional commands */
}

/*
 * This routine is called from the trap or interrupt handler when
 * a breakpoint instruction is encountered or a single step operation
 * completes. The argument is a pointer to a machine dependent 
 * saved_state structure that was built on the interrupt or kernel stack.
 */
void 
kgdb_trap(int type, char *ssp)
{
	jmp_buf_t jmpbuf;
	jmp_buf_t *prev_jmpbuf = kgdb_recover;
	gdb_response_t response = &kgdb_response_buffer;
	int signal = 5;

	if (!kgdb_initialized) {
		kgdb_recover = prev_jmpbuf;
		return;			/* not debugging */
	}

	kgdb_debug(64, "kgdb_trap(%d)\n", type);

	if (_setjmp(kgdb_recover = &jmpbuf)) {
		printf("kgdb_trap(%d): unexpected exception in debugger\n",
		       type);
		dump_ssp(ssp);
		delay(500000);
	}

	kgdb_stop = FALSE;
	kgdb_active = TRUE;
	kgdb_fixup_pc(type, ssp);	/* machine-dependent fix PC */

	if (!kgdb_connected) {
		printf("kgdb: waiting for debugger connection\n");
	}

	while (TRUE) {
		gdb_request_t request = kgdb_get_request();
		if (kgdb_request(request, response, type, ssp)) {
			break;
		}
	}

	kgdb_recover = prev_jmpbuf;
	kgdb_active = FALSE;
}
#endif
