/*
 * Mach Operating System
 * Copyright (c) 1990,1989 Carnegie Mellon University
 * Copyright (c) 1990,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: exec.c 1.11 94/12/03$
 */

/*
 * hppa-specific routines for loading a.out files.
 */

#include <mach/thread_status.h>
/*#include <mach_interface.h>*/
#include <mach_user_internal.h>

#include <file_io.h>
#include <loader_info.h>

#include <machine/exec.h>
#include <machine/act.h>

#define round_to_word(x) (((x) + sizeof(int)) & ~(sizeof(int) - 1))

/*
 * Machine-dependent portions of execve() for the hp800
 */

int ex_get_header(fp, lp)
	struct file *fp;
	register struct loader_info *lp;
{
	union exec_hdr x;
	register int	result;
	vm_size_t	resid;

	/*
	 * Read in all the necessary file headers at once.
	 */

	result = read_file(fp, 0, (vm_offset_t)&x, sizeof(x), &resid);
	if (result)
		return (result);
	if (resid)
		return (EX_NOT_EXECUTABLE);

	/*
	 * We load 800 BSD images, 700 OSF SOM images or ELF images
	 */

	if (x.som.fhdr.system_id != MID_HP800
	    && x.som.fhdr.system_id != MID_HPUX
	    && x.elf.ehdr.e_ident[0] != ELFMAG0
	    && x.elf.ehdr.e_ident[1] != ELFMAG1
	    && x.elf.ehdr.e_ident[2] != ELFMAG2
	    && x.elf.ehdr.e_ident[3] != ELFMAG3)
		return (EX_NOT_EXECUTABLE);

	if (x.som.fhdr.system_id != MID_HP800
	    && x.som.fhdr.system_id != MID_HPUX)
	  goto elf;

som:
	/*
	 * see if this is an OSF SOM images
	 */
	if (x.som.fhdr.system_id == MID_HPUX) {
		if (x.som.ehdr.som_auxhdr.type != 0xC)
			return (EX_NOT_EXECUTABLE);
		x.som.ehdr.exec_entry = x.som.ehdr.exec_flags;
	}

	switch ((int)x.som.fhdr.a_magic) {

	    case EXEC_MAGIC:
		lp->format = EX_READIN;
		lp->text_start  = 0;
		lp->text_size   = 0;
		lp->text_offset = 0;
		lp->data_start  = x.som.ehdr.exec_tmem;
		lp->data_size   = (x.som.ehdr.exec_dfile + x.som.ehdr.exec_dsize
				   - x.som.ehdr.exec_tfile);
		lp->data_offset = x.som.ehdr.exec_tfile;
		lp->bss_size    = x.som.ehdr.exec_bsize;
		break;

	    case SHARE_MAGIC:
		if (x.som.ehdr.exec_tsize == 0) {
			return (EX_NOT_EXECUTABLE);
		}
		lp->format = EX_SHAREABLE;
		lp->text_start  = x.som.ehdr.exec_tmem;
		lp->text_size   = x.som.ehdr.exec_tsize;
		lp->text_offset = x.som.ehdr.exec_tfile;
		lp->data_start  = x.som.ehdr.exec_dmem;
		lp->data_size   = x.som.ehdr.exec_dsize;
		lp->data_offset = x.som.ehdr.exec_dfile;
		lp->bss_size    = x.som.ehdr.exec_bsize;
		break;

	    case DEMAND_MAGIC:
		if (x.som.ehdr.exec_tsize == 0) {
			return (EX_NOT_EXECUTABLE);
		}
		lp->format = EX_PAGEABLE;
		lp->text_start  = x.som.ehdr.exec_tmem;
		lp->text_size   = x.som.ehdr.exec_tsize;
		lp->text_offset = x.som.ehdr.exec_tfile;
		lp->data_start  = x.som.ehdr.exec_dmem;
		lp->data_size   = x.som.ehdr.exec_dsize;
		lp->data_offset = x.som.ehdr.exec_dfile;
		lp->bss_size    = x.som.ehdr.exec_bsize;
		break;
	    default:
		return (EX_NOT_EXECUTABLE);
	}
	lp->entry_1 = x.som.ehdr.exec_entry;

	/*
	 * We'll use entry_2 to pass the default dp
	 */

	lp->entry_2 = x.som.fhdr.presumed_dp;

	return 0;

elf:
	{
	    unsigned int index;

	    lp->format = EX_SHAREABLE;
	    for (index = 0; index < x.elf.ehdr.e_phnum; index++) {

		/* Skip zero sized segments and those which are not
		   part of the executable image.  */
		if (x.elf.phdrs[index].p_type != PT_LOAD
		    || x.elf.phdrs[index].p_memsz == 0)
		    continue;

		/* If it's not writable, then it must be the text.  */
		if ((x.elf.phdrs[index].p_flags & PF_W) == 0) {
		    lp->text_start = x.elf.phdrs[index].p_vaddr;
		    lp->text_size = x.elf.phdrs[index].p_filesz;
		    lp->text_offset = x.elf.phdrs[index].p_offset;
		} else {
		    lp->data_start = x.elf.phdrs[index].p_vaddr;
		    lp->data_size = x.elf.phdrs[index].p_filesz;
		    lp->data_offset = x.elf.phdrs[index].p_offset;
		    lp->bss_size = (x.elf.phdrs[index].p_memsz
				    -x.elf.phdrs[index].p_filesz);
		}
	    }
	    lp->entry_1 = x.elf.ehdr.e_entry;
	    return 0;
	}
}

char *set_regs(user_task, user_thread, lp, arg_size)
	mach_port_t	user_task;
	mach_port_t	user_thread;
	struct loader_info *lp;
	int		arg_size;
{
	vm_offset_t	stack_start;
	vm_offset_t	stack_end;
	struct parisc_thread_state regs;
	unsigned int	reg_size;

	/*
	 * Allocate stack.
	 */

	stack_start = PARISC_STACK_START;
	stack_end   = PARISC_STACK_START + PARISC_STACK_SIZE;

	(void)vm_allocate(user_task,
			  &stack_start,
			  (vm_size_t)(stack_end - stack_start),
			  FALSE);

	reg_size = PARISC_THREAD_STATE_COUNT;

	(void)act_get_state_KERNEL(user_thread,
				PARISC_THREAD_STATE,
				(thread_state_t)&regs,
				&reg_size);

	regs.iioq_head = lp->entry_1;
	regs.iioq_tail = lp->entry_1 + 4;

	regs.dp = lp->entry_2;

	/*
	 * put the argument list at the beginning of the stack
	 */

	regs.sp = round_to_word(stack_start + arg_size);

	(void)act_set_state_KERNEL(user_thread,
				PARISC_THREAD_STATE,
				(thread_state_t)&regs,
				reg_size);

	return ((char *) stack_start);
}

void
set_regs_2(count, usp, user_task, user_thread)
	unsigned	count;
	unsigned int	*usp;
	mach_port_t	user_task;
	mach_port_t	user_thread;
{
	struct parisc_thread_state regs;
	unsigned int	reg_size;

	reg_size = PARISC_THREAD_STATE_COUNT;

	(void)act_get_state_KERNEL(user_thread,
				PARISC_THREAD_STATE,
				(thread_state_t)&regs,
				&reg_size);

	regs.arg0 = count;
	regs.arg1 = (unsigned) ++usp;
	regs.arg2 = 0;
	regs.arg3 = 0;	

	(void)act_set_state_KERNEL(user_thread,
				PARISC_THREAD_STATE,
				(thread_state_t)&regs,
				reg_size);
}
