/*
 * Copyright (c) 1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *      Utah $Hdr: msg.s 1.6 95/08/23$
 *      Author: Bryan Ford, University of Utah CSL
 */
/*
 *	Fast migrating mach_msg_trap path
 */

#include <machine/asm.h>
#include <assym.s>
#include <machine/psl.h>
#include <machine/act.h>

#define COMPLEX_REQUEST	1
#define COMPLEX_REPLY	1
#define ACT_FALLBACK	0

#define FR_SIZE		64	/* XXX */

	.space	$PRIVATE$
	.subspa	$DATA$

	.import cur_pcb,data
	.import need_ast,data
	.import istackptr,data /* XXX */
	.import kernel_port,data
#ifdef SYSCALLSTAT
	.import syscallstats,data
#endif

	.space	$TEXT$
	.subspa $CODE$

	.export	$mach_msg_trap_fast,entry

/*
 * On entry:
 *	ret1 = current PCB
 *	t3 = current activation
 *	t1 = system call code (=-25)
 *	arg0 = user-space message buffer
 *	arg1 = message options (must be MACH_MSG_OPTION_FAST for this path to work)
 *	arg2 = send_size
 *	arg3 = receive size
 *
 * Except for t1, these must not be changed as long as we may still fall off the fast path.
 * All other caller-save registers are available.
 */
$mach_msg_trap_fast

	/*
	 * During the first part of this routine,
	 * interruption rollback does nothing
	 * (except release locks on a multiprocessor);
	 * encapsulation rollback just kicks back to syscall entry.
	 */

#ifdef SYSCALLSTAT
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	12(r1),t2
	ldo	1(t2),t2
	stw	t2,12(r1)
#endif
	/*
	 * First check to make sure we're both sending and receiving,
	 * and no notifications or timeouts are requested.
	 */
#ifdef SYSCALLSTAT
	comib,=,n MACH_MSG_OPTION_MIGR,arg1,$foo1
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	16(r1),t2
	ldo	1(t2),t2
	stw	t2,16(r1)
	b,n	$abort
$foo1
#else
	comib,<>,n MACH_MSG_OPTION_MIGR,arg1,$abort
#endif

#ifdef MULTIPLE_SPACES_PER_TASK
	/*
	 * Get the user's data space pointer for explicit user space accesses.
	 */
	ldw	SS_SR5(ret1),t4
#else
	/*
	 * Get the user's data space pointer for explicit user space accesses.
	 * XXX sr4 - sr6 are all the same, so we only save sr4 in the saved
	 * state.  Hence, we must use it.
	 */
	ldw	SS_SR4(ret1),t4
#endif
	mtsp	t4,sr1

	/*
	 * next get the user's stack pointer into t4
	 */
	ldw	SS_R30(ret1),t4

#ifndef ROLLBACK
	/*
	 * Prepare for faults when accessing the user address space.
	 */
	ldil	L%$msg_rollback_3,t1
	ldw	ACT_THREAD(t3),r1
	ldo	R%$msg_rollback_3(t1),t1
	stw	t1,THREAD_RECOVER(r1)
#endif
	/*
	 * Check the msgh_bits field to make sure we're making a normal
	 * send/reply request.
	 */
	ldw	MSGH_BITS(sr1,arg0),t1
	ldil	L%MACH_MSGH_BITS_MIGR,t2
	ldo	R%MACH_MSGH_BITS_MIGR(t2),t2
#if COMPLEX_REQUEST
	depi	0,MACH_MSGH_COMPLEX_BIT,1,t1
#endif
#ifdef SYSCALLSTAT
	comb,=,n t1,t2,$foo2
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	20(r1),t2
	ldo	1(t2),t2
	stw	t2,20(r1)
	b,n	$abort
$foo2
#else
	comb,<>,n t1,t2,$abort
#endif

#if !COMPLEX_REPLY
	/* MIKE hack */
		/*
		 * Nix known complex return calls:
		 *	100000: generic (ugh!)
		 *	101035: bsd_task_by_pid
		 *	101057: bsd_table_get
		 */
		ldw	MSGH_ID(sr1,arg0),t1
		ldil	L%100000,t2
		ldo	R%100000(t2),t2
		comb,=,n t1,t2,$abort
		ldil	L%101035,t2
		ldo	R%101035(t2),t2
		comb,=,n t1,t2,$abort
		ldil	L%101057,t2
		ldo	R%101057(t2),t2
		comb,=,n t1,t2,$abort
	/* End MIKE hack */
#else
	/* MIKE hack */
		ldw		MSGH_ID(sr1,arg0),t1
#ifdef ACTWATCH
		/* XXX stash mig id for later */
		stw		t1,ACT_SPARE+44(t3)
#endif
		/*
		 * XXX skip bsd_exec_done for now so GDB will work
		 */
		ldil		L%101099,t2
		ldo		R%101099(t2),t2
		comb,=,n	t1,t2,$abort
		/*
		 * Skip exit for now, id=100000 with syscode=1
		 * XXX should check size
		 */
		ldil		L%100000,t2
		ldo		R%100000(t2),t2
		comb,<>,n	t1,t2,$notgeneric
		ldw		0x20(sr1,arg0),t1 /* bsd_request.syscode */
#ifdef ACTWATCH
		/* XXX bsd syscode for later */
		stw		t1,ACT_SPARE+48(t3)
#endif
		comib,=,n	1,t1,$isbad
#ifdef SYSCALLSTAT
		b,n		$notgeneric
$isbad
		ldil		L%syscallstats,r1
		ldo		R%syscallstats(r1),r1
		ldw		44(r1),t2
		ldo		1(t2),t2
		stw		t2,44(r1)
		b,n		$abort
#endif
$notgeneric
	/* End MIKE hack */
#endif

	/*
	 * Make sure the port we're receiving from
	 * is the same as the port we're making the reply right from.
	 */
	ldw	MSGH_LOCAL_PORT(sr1,arg0),t1
	ldw	VA_ARG4(sr1,t4),t2
#ifdef SYSCALLSTAT
	comb,=,n t1,t2,$foo4
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	24(r1),t2
	ldo	1(t2),t2
	stw	t2,24(r1)
	b,n	$abort
$foo4
#else
	comb,<>,n t1,t2,$abort
#endif

	/*
	 * Find the current thread (r1) and IPC space (t4)
	 */
	ldw	ACT_TASK(t3),t4
	ldw	ACT_THREAD(t3),r1
	ldw	TASK_SPACE(t4),t4

	/*
	 * Now we start peering into data that isn't part of our activation
	 * and isn't constant (like ACT_TASK and TASK_SPACE are).
	 * Therefore interruption rollback must kick us back here
	 * during this next little bit.
	 * Encapsulation rollback is still unaffected.
	 */

	/*
	 * Lookup the destination port (t4).
	 * If it isn't in the table, we have to divert and call ipc_entry_lookup.
	 */
	ldw	MSGH_REMOTE_PORT(sr1,arg0),t1
	ldw	SPACE_TABLE_SIZE(t4),t2
	ldw	SPACE_TABLE(t4),t4
	comb,>>=,n t1,t2,$divert_dest_lookup
#if IPC_ENTRY_SIZE == 16
	shd	t1,r0,28,t1
	add	t1,t4,t4
#endif
#if IPC_ENTRY_SIZE == 20
	sh2add	t1,t1,t1
	sh2add	t1,t4,t4
#endif
#if IPC_ENTRY_SIZE == 24
	sh1add	t1,t1,t1
	sh3add	t1,t4,t4
#endif
$divert_dest_lookup_back
	ldw	ENTRY_BITS(t4),t1
	ldw	ENTRY_OBJECT(t4),t4
#ifdef SYSCALLSTAT
	bb,<,n t1,MACH_PORT_TYPE_SEND_BIT,$foo5
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	32(r1),t2
	ldo	1(t2),t2
	stw	t2,32(r1)
	b,n	$abort
$foo5
#else
	bb,>=,n t1,MACH_PORT_TYPE_SEND_BIT,$abort
#endif

	/*
	 * Save the port's receiver name (r31).
	 */
	ldw	IP_RECEIVER_NAME(t4),r31

	/*
	 * XXX save a pointer to the sequence number for later
	 */
	ldo	IP_SEQNO(t4),t2
	stw	t2,ACT_SPARE+52(t3)

	/*
	 * Find the port's current target (t4).
	 * It'll be either the one in itself or the one in its port set.
	 */
	ldw	IP_CUR_TARGET(t4),t4

	/*
	 * Grab an activation from the port's pool (t2).
	 * If the port doesn't have a pool, fall back to a message.
	 * If the pool is empty, block waiting for one.
	 */
	ldw	IPT_ACTS(t4),t2
	comb,<>,n t2,r0,$haveact
#ifdef SYSCALLSTAT
	ldw	IPT_TYPE(t4),t1
	bb,<	t1,IPT_TYPE_MIGRATE_RPC_BIT,$foo6
	ldil	L%syscallstats,t1
	ldo	R%syscallstats(t1),t1
	ldw	36(t1),t2
	ldo	1(t2),t2
	stw	t2,36(t1)
	b,n	$abort
$foo6
	ldo	R%syscallstats(t1),t1
	ldw	40(t1),t2
	ldo	1(t2),t2
	stw	t2,40(t1)
#if ACT_FALLBACK
	b,n	$abort
#else
	b,n	$waitforact
#endif
#else
#if !ACT_FALLBACK
	/*
	 * No activation, see if the port/pset supports migrating RPC.
	 * If so, we wait, otherwise we fall back on a message.
	 */
	ldw	IPT_TYPE(t4),t1
	bb,<,n	t1,IPT_TYPE_MIGRATE_RPC_BIT,$waitforact
#endif
	b,n	$abort
#endif

$haveact
	ldw	ACT_IPT_NEXT(t2),t1
	stw	t1,IPT_ACTS(t4)

	/*
	 * We will no longer abort beyond this point so now we will
	 * go to the trouble of saving the callee-save registers.
	 */
	stw	r3,SS_R3(ret1)
#ifndef R4_IS_SPECIAL
	stw	r4,SS_R4(ret1)
#endif
	stw	r5,SS_R5(ret1)
	stw	r6,SS_R6(ret1)
	stw	r7,SS_R7(ret1)
	stw	r8,SS_R8(ret1)
	stw	r9,SS_R9(ret1)
	stw	r10,SS_R10(ret1)
	stw	r11,SS_R11(ret1)
	stw	r12,SS_R12(ret1)
	stw	r13,SS_R13(ret1)
	stw	r14,SS_R14(ret1)
	stw	r15,SS_R15(ret1)
	stw	r16,SS_R16(ret1)
	stw	r17,SS_R17(ret1)
	stw	r18,SS_R18(ret1)

$gotact
	/*
	 * An activation can only be in-use while on a (different) thread,
	 * so we don't need to check that here.
	 */

#ifdef SYSCALLSTAT
	ldil	L%syscallstats,t1
	ldo	R%syscallstats(t1),t1
	ldw	48(t1),ret0
	ldo	1(ret0),ret0
	stw	ret0,48(t1)
#endif

	/*
	 * Attach server activation (t2) to the current thread (r1).
	 * The old activation is still in t3.
	 */
	stw	t3,ACT_LOWER(t2)
	stw	t2,ACT_HIGHER(t3)
	stw	t2,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t2)
	stw	r0,ACT_USER_ACT(t2)

	/* XXX yick!!! stash the message pointer for the return path */
	stw	arg0,ACT_SPARE+60(t3)

	/*
	 * Find the new activation's PCB (t1),
	 * and make it officially the current one.
	 */
	ldw	ACT_PCB(t2),t1
	ldil	L%cur_pcb,arg1
	stw	t1,R%cur_pcb(arg1)

	/*
	 * At this point,
	 * the new activation is attached to and owned by the thread.
	 * We no longer need to muck with IPC data structures and such,
	 * only the client and server activations.
	 * Therefore, interruption rollback doesn't need to return past here,
	 * but it does need to mark the client activation as in-use
	 * since we're still in transition.
	 */

	/* XXX refcount, alerts */

	/*
	 * Indicate that no floating-point state exists yet in the new PCB
	 * by stuffing an illegal value in the saved fr0.
	 */
	ldi	-1,arg1
	stw	arg1,PCB_FS+FS_FR0(t1)

	/*
	 * Find the destination message buffer (arg1)
	 */
	ldw	ACT_USER_STACK(t2),arg1

	/*
	 * Find the destination space (r17, sr2) and PID (r18, pidr3)
	 */
	ldw	ACT_SPACE(t2),r17
	ldw	ACT_PID(t2),r18
	mtsp	r17,sr2
	mtctl	r18,pidr3

#ifndef ROLLBACK
	/*
	 * Prepare for faults accessing either the user or server address space.
	 * If interrupted now, we must pop off the server act and restore the
	 * callee save registers.
	 */
	ldil	L%$msg_rollback_2,t1
	ldo	R%$msg_rollback_2(t1),t1
	stw	t1,THREAD_RECOVER(r1)
#endif

	/*
	 * Copy the header.
	 * The size is still in arg2; we need to store it in msgh_size.
	 * The receiver name was stashed in r31 earlier; copy it as well.
	 */
	ldw	MSGH_BITS(sr1,arg0),t1
	stw	t1,MSGH_BITS(sr2,arg1)
	stw	arg2,MSGH_SIZE(sr2,arg1)
	stw	r31,MSGH_LOCAL_PORT(sr2,arg1)

#if COMPLEX_REPLY
	/*
	 * Save the (untranslated) reply port in the destinations remote
	 * remote port.  Needed for complex replies.
	 */
	ldw	MSGH_LOCAL_PORT(sr1,arg0),t1
	stw	t1,MSGH_REMOTE_PORT(sr2,arg1)
#endif

#if COMPLEX_REQUEST 
	/*
	 * Check for complex message
	 */
	ldw	MSGH_BITS(sr1,arg0),t1
	bb,<,n	t1,MACH_MSGH_COMPLEX_BIT,$complex_request
#endif

	/*
	 * Grab and then update the receive port sequence number
	 * (complex messages have this done elsewhere).
	 * XXX this should be done with the message queue lock held.
	 */
	ldw	ACT_LOWER(t2),t3
	ldw	ACT_SPARE+52(t3),t1
	ldw	0(t1),r4
	stw	r4,MSGH_SEQNO(sr2,arg1)
	ldo	1(r4),r4
	stw	r4,0(t1)

	/*
	 * Copy the rest of the message.
	 */
	ldo	MSGH_ID(arg0),arg0
	ldo	MSGH_ID(arg1),arg3
	ldo	-MSGH_ID(arg2),arg2

        addib,<,n -16,arg2,$copyto_movewords

	/*
	 * We can move the data in 4 word moves. We'll use 4 registers to
	 * avoid interlock and pipeline stalls.
	 */

$copyto_loop16

        ldwm	16(sr1,arg0),r4
        ldw	-12(sr1,arg0),r5
        ldw     -8(sr1,arg0),r6
        ldw     -4(sr1,arg0),r7
        stwm    r4,16(sr2,arg3)
        stw     r5,-12(sr2,arg3)
        stw     r6,-8(sr2,arg3)
        addib,>= -16,arg2,$copyto_loop16
        stw     r7,-4(sr2,arg3)

	/*
	 * We have already decremented the count by 16, add 12 to it and then 
	 * we can test if there is at least 1 word left to move.
	 */

$copyto_movewords
        addib,<,n 12,arg2,$copyto_done

	/*
	 * Clean up any remaining words that were not moved in the 16 byte
	 * moves
	 */

$copyto_loop4
        ldwm	4(sr1,arg0),r4
        addib,>= -4,arg2,$copyto_loop4
        stwm    r4,4(sr2,arg3)

$copyto_done

#ifndef ROLLBACK
	/*
	 * No more faults should happen
	 */
	stw	r0,THREAD_RECOVER(r1)
#endif

	/*
	 * Clear out the extra pidr.
	 */
	mtctl	r0,pidr3

#ifdef ACTCHECK
	ldw		ACT_RETURN_ROUTINE(t2),t1
	comb,=,n	r0,t1,$rtis0
	break		0x1f,0x1ffe
$rtis0
#endif

	/*
	 * Set the server's return_routine.
	 */
	ldil	L%$fast_return,rp
	ldo	R%$fast_return(rp),rp
	stw	rp,ACT_RETURN_ROUTINE(t2)

	/*
	 * Disable the FPU so the new user context will trap when trying to use it.
	 */
	mtctl	r0,ccr

#ifdef ACTWATCH
	/*
	 * Now that server activation is setup, check for inconsistancies.
	 */
	break	0,11
#endif

	/*
	 * Set up the user register state.
	 * We allocate a caller frame (including space for the standard
	 * arguments!) just above the message and set the user stack pointer
	 * just above that.
	 */
	ldo	FR_SIZE+63(arg3),t1
	depi	0,31,6,t1
	copy	t1,sp
	copy	arg1,arg0
	ldw	IPT_USER_ENTRYPOINT(t4),arg2
	ldw	IPT_USER_DATA(t4),dp

	/*
	 * XXX experimental: if this was the last activation in the pool,
	 * piggyback an indication to the server via an additional argument
	 * to the entry routine.  We make it arg3 to be consistent with
	 * the lrmsg path.
	 */
	ldw	ACT_IPT_NEXT(t2),t1
	comclr,<> r0,t1,arg3
	ldi	1,arg3

	/*
	 * XXX clear arg1 to indicate this is not the lrmsg path
	 */
	copy	r0,arg1

	/*
	 * Set space registers
	 * This means that we need to start explicitly saying sr1.
	 */
	mtsp	r0,sr1
	mtsp	r17,sr4
	mtsp	r17,sr5
	mtsp	r17,sr6

	/*
	 * Set protection ID registers.
	 * Wait to set pidr4 until the last minute.
	 */
	mtctl	r18,pidr1

	/*
	 * finally return to the caller and set the protection ID correctly
	 */
	.call
	be	0(sr4,arg2)
	mtctl	r0,pidr4

$fast_return
	/*
	 * Just came from the return path in the gateway page.
	 * t2 = server activation.
	 * arg0 = server msg buffer.
	 * arg1 = 1 if we should copyout message, 0 ow.
	 *
	 * dp = kernel DP
	 * sr4/5/6/ = kernel space
	 * pidr4 = kernel PID
	 */

	/*
	 * Find the thread (r1) and client activation (t3)
	 */
	ldw	ACT_THREAD(t2),r1
	ldw	ACT_LOWER(t2),t3

	/*
	 * Skip copyout if requested
	 */
	comb,=,n r0,arg1,$copyfrom_done

#if 0
	/*
	 * If client activation is dead, skip copying return info
	 */
	ldw	ACT_ACTIVE(t3),t1
	comb,=,n r0,t1,$copyfrom_done
#endif

#ifndef ROLLBACK
	/*
	 * Prepare for faults accessing either the user or server address space.
	 * If interrupted now, we must pop off the server act and restore the
	 * callee save registers.
	 */
	ldil	L%$msg_rollback_1,t1
	ldo	R%$msg_rollback_1(t1),t1
	stw	t1,THREAD_RECOVER(r1)
#endif

	/*
	 * Find the server space (sr1) XXX
	 */
	ldw	ACT_SPACE(t2),t1
	mtsp	t1,sr1

	/*
	 * Before getting any further along, check for a complex reply.
	 * If we have one, we head off to C code to deal with it.
	 */
	ldw	MSGH_BITS(sr1,arg0),t1
	bb,<	t1,MACH_MSGH_COMPLEX_BIT,$complex_reply
	ldw	MSGH_SIZE(sr1,arg0),arg2

	/*
	 * Find the client space (sr2) and pid (pidr3)
	 */
	ldw	ACT_SPACE(t3),t1
	mtsp	t1,sr2
	ldw	ACT_PID(t3),t1
	mtctl	t1,pidr3

	/* XXX yick!!! grab the client's message pointer */
	ldw	ACT_SPARE+60(t3),arg1

	/*
	 * Copy the header.
	 */
	ldw	MSGH_BITS(sr1,arg0),t1
	stw	t1,MSGH_BITS(sr2,arg1)
	stw	arg2,MSGH_SIZE(sr2,arg1)

	/*
	 * Copy the rest of the message.
	 */
	ldo	MSGH_ID(arg0),arg0
	ldo	MSGH_ID(arg1),arg3
	ldo	-MSGH_ID(arg2),arg2

        addib,<,n -16,arg2,$copyfrom_movewords

	/*
	 * We can move the data in 4 word moves. We'll use 4 registers to
	 * avoid interlock and pipeline stalls.
	 */

$copyfrom_loop16

        ldwm	16(sr1,arg0),t1
        ldw	-12(sr1,arg0),t4
        stwm    t1,16(sr2,arg3)
        stw     t4,-12(sr2,arg3)
        ldw     -8(sr1,arg0),t1
        ldw     -4(sr1,arg0),t4
        stw     t1,-8(sr2,arg3)
        addib,>= -16,arg2,$copyfrom_loop16
        stw     t4,-4(sr2,arg3)

	/*
	 * We have already decremented the count by 16, add 12 to it and then 
	 * we can test if there is at least 1 word left to move.
	 */

$copyfrom_movewords
        addib,<,n 12,arg2,$copyfrom_done

	/*
	 * Clean up any remaining words that were not moved in the 16 byte
	 * moves
	 */

$copyfrom_loop4
        ldwm	4(sr1,arg0),t1
        addib,>= -4,arg2,$copyfrom_loop4
        stwm    t1,4(sr2,arg3)

$copyfrom_done
	/*
	 * Will return success.
	 */
	copy	r0,ret0

	/*
	 * Restore the client pidr to its expected location.
	 * (no more server access at this point).
	 */
	ldw	ACT_PID(t3),t1
	mtctl	t1,pidr1

	/*
	 * Clear out the extra (client) pidr.
	 */
	mtctl	r0,pidr3

$msg_finish
#ifndef ROLLBACK
	/*
	 * No more faults should happen
	 */
	stw	r0,THREAD_RECOVER(r1)
#endif

#ifdef ACTWATCH
	/*
	 * All done, before popping server activation, perform a final check.
	 * XXX set return_routine non-zero so we don't trigger for this act.
	 * XXX put mig ID and bsd syscode in args so they get dumped.
	 */
#ifdef ACTCHECK
	stw	t2,ACT_RETURN_ROUTINE(t2)
#endif
	ldw	ACT_SPARE+44(t3),arg0
	ldw	ACT_SPARE+48(t3),arg1
	break	0,11
#ifdef ACTCHECK
	stw	r0,ACT_RETURN_ROUTINE(t2)
#endif
#endif

	/*
	 * Remove the server activation from the thread,
	 * making the client activation the current one.
	 */
	stw	r0,ACT_THREAD(t2)
	stw	t3,THREAD_TOP_ACT(r1)
	ldw	ACT_PCB(t3),t1
	ldil	L%cur_pcb,t4
	stw	t1,R%cur_pcb(t4)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t2)
	stw	r0,ACT_HIGHER(t3)
#endif

	/*
	 * Return the server activation to its ipc_target
	 */
	ldw	ACT_IPT(t2),arg0
	comb,=,n arg0,r0,$return_no_target
	ldw	IPT_ACTS(arg0),t4
	stw	t2,IPT_ACTS(arg0)
	stw	t4,ACT_IPT_NEXT(t2)

#if !ACT_FALLBACK
	/*
	 * Wakeup anyone waiting for an activation from this pool.
	 */
	ldw	IPT_WAITING(arg0),t4
	comb,<>,n r0,t4,$actwaiting
#endif

$return_no_target
	/*
	 * Restore callee-save registers.
	 */
	ldw	SS_R3(t1),r3
#ifndef R4_IS_SPECIAL
	ldw	SS_R4(t1),r4
#endif
	ldw	SS_R5(t1),r5
	ldw	SS_R6(t1),r6
	ldw	SS_R7(t1),r7
	ldw	SS_R8(t1),r8
	ldw	SS_R9(t1),r9
	ldw	SS_R10(t1),r10
	ldw	SS_R11(t1),r11
	ldw	SS_R12(t1),r12
	ldw	SS_R13(t1),r13
	ldw	SS_R14(t1),r14
	ldw	SS_R15(t1),r15
	ldw	SS_R16(t1),r16
	ldw	SS_R17(t1),r17
	ldw	SS_R18(t1),r18

	.import	$syscall_reentry,code
	ldil	L%$syscall_reentry,t2
	ldo	R%$syscall_reentry(t2),t2
	bv,n	0(t2)
	nop

#ifndef ROLLBACK
/*
 * "Rollback" recovery code
 */
$msg_rollback_1
	/*
	 * Restore the client pidr to its expected location.
	 * (no more server access at this point).
	 */
	ldw	ACT_PID(t3),t1
	mtctl	t1,pidr1
$msg_rollback_2
	/*
	 * Clear out the extra (client) pidr
	 */
	mtctl	r0,pidr3
	/*
	 * Return with an error
	 */
	b	$msg_finish
	ldi	1,ret0

$msg_rollback_3
	/*
	 * This one is particularly gruesome.
	 * We get here if we faulted very early, before allocating a
	 * server activation.  In that case, the fault will have overwritten
	 * some fields in the user activation save_state which we have already
	 * saved:
	 *	flags, r2, r4, r27, r30, r31
	 *	sr0, sr4, sr5, sr6
	 * of those, the only ones we have changed (and hence overwrote with
	 * bad values in the SS) are:
	 *	flags, sr4, sr5, sr6
	 * We know that flags should be set to SS_INSYSCALL so we restore it
	 * We know that the user's value of sr4 is now in sr1 so we restore it 
	 * NOTE: the latter only works if MULTIPLE_SPACES_PER_TASK is not
	 * defined.
	 *
	 * We then need to reset the thread recovery pointer,
	 * restore t1 (cur_pcb) for the syscall_reentry code,
	 * and return an error.
	 */
	ldi	SS_INSYSCALL,t1
	stw	t1,SS_FLAGS(ret1)
	mfsp	sr1,t1
	stw	t1,SS_SR4(ret1)
	ldw	ACT_THREAD(t3),r1
	stw	r0,THREAD_RECOVER(r1)
	ldw	ACT_PCB(t3),t1
	.import	$syscall_reentry,code
	ldil	L%$syscall_reentry,t2
	ldo	R%$syscall_reentry(t2),t2
	bv	0(t2)
	ldi	1,ret0
#endif

$complex_reply
	/*
	 * Complex reply handling.  At this point:
	 *	r1:	thread
	 *	t2:	server activation (current)
	 *	t3:	client activation
	 *	arg0:	server message buffer
	 *	arg2:	message size
	 */

#if !COMPLEX_REPLY
	/* MIKE hack */
	ldw	MSGH_ID(sr1,arg0),t4
	break	0x1f,0x1fff
	/* end MIKE hack */
#else
#ifdef SYSCALLSTAT
	ldil	L%syscallstats,t1
	ldo	R%syscallstats(t1),t1
	ldw	52(t1),t4
	ldo	1(t4),t4
	stw	t4,52(t1)
#endif

	/*
	 * Save away vestigal server state
	 * XXX really needed?
	 */
	ldw	ACT_PCB(t2),t4
	stw	sp,SS_R30(t4)

	/*
	 * Allocate a kernel activation
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4
	ldw	IPT_ACTS(t4),t1
	comb,<>,n t1,r0,$kact_available1

	/* XXX block */
	ldil	L%istackptr,t3
	ldw	R%istackptr(t3),sp
	stw	r0,R%istackptr(t3)
	.import panic_no_kstacks,code
	ldil	L%panic_no_kstacks,t2
	ldo	R%panic_no_kstacks(t2),t2
	bv	0(t2)
	ldo	FM_SIZE(sp),sp

$kact_available1

	/*
	 * Remove it from the ipc_target.
	 */
	ldw	ACT_IPT_NEXT(t1),ret0
	stw	ret0,IPT_ACTS(t4)

	/*
	 * Attach the kernel activation (t1) to the current thread (r1).
	 * The server activation is still in t2, user activation in t3.
	 */
	stw	t2,ACT_LOWER(t1)
	stw	t1,ACT_HIGHER(t2)
	stw	t1,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t1)

	/*
	 * Take a reference on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t2),t4
	ldo	1(t4),t4
	stw	t4,ACT_REF_COUNT(t2)
	stw	t2,ACT_USER_ACT(t1)

	/*
	 * Make the new activation officially the current one.
	 */
	ldw	ACT_PCB(t1),ret0
	ldil	L%cur_pcb,t4
	stw	ret0,R%cur_pcb(t4)

	/*
	 * Initialize the kernel stack pointer from the new activation
	 * (locore gateway page took care of dp).
	 */
	ldw	ACT_USER_STACK(t1),sp

	/*
	 * rv = rpc_copy_reply(smsg, stask, dmsg, dtask, size)
	 */
	stw     r0,FM_PSP(sp)
	stw	arg2,VA_ARG4(sp)
	ldw	ACT_TASK(t3),arg3
	ldw	ACT_SPARE+60(t3),arg2	/* XXX stashed */
	ldw	ACT_TASK(t2),arg1
	.import	rpc_copy_reply,code
	ldil	L%rpc_copy_reply,t1
	ldo	R%rpc_copy_reply(t1),t1
	.call
	blr	r0,rp
	bv	0(t1)
	nop

	/*
	 * Reload thread (r1), kernel activation (t1), server activation (t2).
	 */
	ldil	L%cur_pcb,t4
	ldw	R%cur_pcb(t4),t1
	ldw	PCB_ACT(t1),t1
	ldw	ACT_LOWER(t1),t2
	ldw	ACT_THREAD(t1),r1

	/*
	 * Remove the kernel activation from the thread.
	 */
	stw	r0,ACT_THREAD(t1)
	stw	t2,THREAD_TOP_ACT(r1)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t1)
	stw	r0,ACT_HIGHER(t2)
	stw	r0,ACT_USER_ACT(t1)
#endif

	/* 
	 * Drop the reference count on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t2),arg2
	ldo	-1(arg2),arg2
	stw	arg2,ACT_REF_COUNT(t2)

	/*
	 * Return the kernel activation to the kernel port.
	 */
	ldil	L%kernel_port,arg2
	ldw	R%kernel_port(arg2),arg2
	ldw	IPT_ACTS(arg2),arg3
	stw	t1,IPT_ACTS(arg2)
	stw	arg3,ACT_IPT_NEXT(t1)

	/*
	 * Return to the server activation.
	 * Reload server SP and pointer to user activation (t3).
	 */
	ldw	ACT_PCB(t2),t1
	stw	t1,R%cur_pcb(t4)
	ldw	SS_R30(t1),sp
	ldw	ACT_LOWER(t2),t3

	/*
	 * Make sure PIDRs are as we expect
	 */
	ldw	ACT_PID(t3),t1
	mtctl	t1,pidr1
	mtctl	r0,pidr3	

	/*
	 * Done.  At this point:
	 *	r1:	thread
	 *	t2:	server activation (current)
	 *	t3:	client activation
	 */
	b,n	$msg_finish
#endif

$divert_dest_lookup
#ifdef SYSCALLSTAT
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	28(r1),t2
	ldo	1(t2),t2
	stw	t2,28(r1)
#else
	/*XXX*/
#endif
	b,n	$abort

#if COMPLEX_REQUEST
$complex_request
#ifdef SYSCALLSTAT
	ldil	L%syscallstats,t1
	ldo	R%syscallstats(t1),t1
	ldw	56(t1),ret0
	ldo	1(ret0),ret0
	stw	ret0,56(t1)
#endif
	/*
	 * Complex request handling.  At this point:
	 *	r1:	thread
	 *	t2:	server activation
	 *	t3:	client activation
	 *	t4:	pointer to port entry
	 *	arg0:	client message buffer
	 *	arg1:	server message buffer
	 *	arg2:	message size
	 */

	/*
	 * State we need to keep around (moved to callee save regs):
	 *	arg1:	server message buffer
	 *	arg2:	message size
	 *	t4:	destination port entry
	 */
	copy	arg1,r14
	copy	arg2,r15
	copy	t4,r16
	 
	/*
	 * Allocate a kernel activation
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4
	ldw	IPT_ACTS(t4),t1
	comb,<>,n t1,r0,$kact_available2

	/* XXX block */
	ldil	L%istackptr,t3
	ldw	R%istackptr(t3),sp
	stw	r0,R%istackptr(t3)
	.import panic_no_kstacks,code
	ldil	L%panic_no_kstacks,t2
	ldo	R%panic_no_kstacks(t2),t2
	bv	0(t2)
	ldo	FM_SIZE(sp),sp

$kact_available2

	/*
	 * Remove it from the port.
	 */
	ldw	ACT_IPT_NEXT(t1),ret0
	stw	ret0,IPT_ACTS(t4)

	/*
	 * Attach the kernel activation (t1) to the current thread (r1).
	 * The server activation is still in t2, user activation in t3.
	 */
	stw	t2,ACT_LOWER(t1)
	stw	t1,ACT_HIGHER(t2)
	stw	t1,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t1)

	/*
	 * Take a reference on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t2),t4
	ldo	1(t4),t4
	stw	t4,ACT_REF_COUNT(t2)
	stw	t2,ACT_USER_ACT(t1)

	/*
	 * Make the new activation officially the current one.
	 */
	ldw	ACT_PCB(t1),ret0
	ldil	L%cur_pcb,t4
	stw	ret0,R%cur_pcb(t4)

	/*
	 * Initialize the kernel stack pointer from the new activation
	 * and set the kernel data pointer.
	 */
	ldw	ACT_USER_STACK(t1),sp
	ldil	L%$global$,dp
	ldo	R%$global$(dp),dp

	/*
	 * rv = rpc_copy_request(smsg, dmsg, size, stask, dtask)
	 */
	stw     r0,FM_PSP(sp)
	ldw	ACT_TASK(t2),t1
	stw	t1,VA_ARG4(sp)
	ldw	ACT_LOWER(t2),t3
	ldw	ACT_TASK(t3),arg3
	.import	rpc_copy_request,code
	ldil	L%rpc_copy_request,t1
	ldo	R%rpc_copy_request(t1),t1
	.call
	blr	r0,rp
	bv	0(t1)
	nop

	/*
	 * Reload thread (r1), kernel activation (t1), server activation (t2).
	 */
	ldil	L%cur_pcb,t4
	ldw	R%cur_pcb(t4),t1
	ldw	PCB_ACT(t1),t1
	ldw	ACT_LOWER(t1),t2
	ldw	ACT_THREAD(t1),r1

	/*
	 * Remove the kernel activation from the thread.
	 */
	stw	r0,ACT_THREAD(t1)
	stw	t2,THREAD_TOP_ACT(r1)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t1)
	stw	r0,ACT_HIGHER(t2)
	stw	r0,ACT_USER_ACT(t1)
#endif

	/* 
	 * Drop the reference count on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t2),arg2
	ldo	-1(arg2),arg2
	stw	arg2,ACT_REF_COUNT(t2)

	/*
	 * Return the kernel activation to the kernel port.
	 */
	ldil	L%kernel_port,arg2
	ldw	R%kernel_port(arg2),arg2
	ldw	IPT_ACTS(arg2),arg3
	stw	t1,IPT_ACTS(arg2)
	stw	arg3,ACT_IPT_NEXT(t1)

	/*
	 * Return to the server activation.
	 */
	ldw	ACT_PCB(t2),t1
	stw	t1,R%cur_pcb(t4)

	/*
	 * Restore needed state
	 */
	copy	r14,arg1
	add	arg1,r15,arg3
	copy	r16,t4

	/*
	 * Done.  At this point:
	 *	t2:	server activation (current)
	 *	t4:	server port entry
	 *	arg1:	server message buffer
	 *	arg3:	end of message in server buffer
	 */
	b,n	$copyto_done
#endif

#ifndef SYSCALLSTAT
$isbad
#endif
$abort
	/*
	 * Load the system call code for mach_msg_trap back into t1.
	 */
	.import	$mach_msg_trap_fast_abort,code
	ldil	L%$mach_msg_trap_fast_abort,t4
	ldo	R%$mach_msg_trap_fast_abort(t4),t4
	bv	0(t4)
	ldi	-25,t1

#if !ACT_FALLBACK
$waitforact
	/*
	 * Save callee-save registers now.
	 */
	stw	r3,SS_R3(ret1)
#ifndef R4_IS_SPECIAL
	stw	r4,SS_R4(ret1)
#endif
	stw	r5,SS_R5(ret1)
	stw	r6,SS_R6(ret1)
	stw	r7,SS_R7(ret1)
	stw	r8,SS_R8(ret1)
	stw	r9,SS_R9(ret1)
	stw	r10,SS_R10(ret1)
	stw	r11,SS_R11(ret1)
	stw	r12,SS_R12(ret1)
	stw	r13,SS_R13(ret1)
	stw	r14,SS_R14(ret1)
	stw	r15,SS_R15(ret1)
	stw	r16,SS_R16(ret1)
	stw	r17,SS_R17(ret1)
	stw	r18,SS_R18(ret1)

	/*
	 * At this point:
	 *	r1:	thread
	 *	t3:	client activation
	 *	t4:	ipc_target pointer
	 *	arg0:	client message buffer
	 *	arg2:	message size
	 *	r31:	receive port (kernel form)
	 *	sr1:	client data space
	 * All except ret0 must be preserved across the C call,
	 * so move them to caller-save registers.
	 */
	copy	r1,r11
	copy	t3,r12
	copy	t4,r13
	copy	arg0,r14
	copy	arg2,r16
	copy	r31,r17
	mfsp	sr1,r18
	 
	/*
	 * Allocate a kernel activation
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4
	ldw	IPT_ACTS(t4),t1
	comb,<>,n t1,r0,$kact_available3

	/* XXX block */
	ldil	L%istackptr,t3
	ldw	R%istackptr(t3),sp
	stw	r0,R%istackptr(t3)
	.import panic_no_kstacks,code
	ldil	L%panic_no_kstacks,t2
	ldo	R%panic_no_kstacks(t2),t2
	bv	0(t2)
	ldo	FM_SIZE(sp),sp

$kact_available3

	/*
	 * Remove it from the ipc_target.
	 */
	ldw	ACT_IPT_NEXT(t1),ret1
	stw	ret1,IPT_ACTS(t4)

	/*
	 * Attach the kernel activation (t1) to the current thread (r1).
	 * The client activation is still in t3.
	 */
	stw	t3,ACT_LOWER(t1)
	stw	t1,ACT_HIGHER(t3)
	stw	t1,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t1)

	/*
	 * Take a reference on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t4
	ldo	1(t4),t4
	stw	t4,ACT_REF_COUNT(t3)
	stw	t3,ACT_USER_ACT(t1)

	/*
	 * Make the new activation officially the current one.
	 */
	ldw	ACT_PCB(t1),ret1
	ldil	L%cur_pcb,t4
	stw	ret1,R%cur_pcb(t4)

	/*
	 * Initialize the kernel stack pointer from the new activation
	 * and set the kernel data pointer.
	 */
	ldw	ACT_USER_STACK(t1),sp
	ldil	L%$global$,dp
	ldo	R%$global$(dp),dp

	/*
	 * act = ipc_target_block(ipt)
	 */
	stw     r0,FM_PSP(sp)
	copy	t1,r10
	copy	r13,arg0
	.import	ipc_target_block,code
	ldil	L%ipc_target_block,t1
	ldo	R%ipc_target_block(t1),t1
	.call
	blr	r0,rp
	bv	0(t1)
	nop

	/*
	 * 
	 * Reload thread (r1), kernel activation (t1), client activation (t3).
	 */
	copy	r11,r1
	copy	r10,t1
	copy	r12,t3

	/*
	 * Remove the kernel activation from the thread.
	 */
	stw	r0,ACT_THREAD(t1)
	stw	t3,THREAD_TOP_ACT(r1)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t1)
	stw	r0,ACT_HIGHER(t3)
	stw	r0,ACT_USER_ACT(t1)
#endif

	/* 
	 * Drop the reference count on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),arg2
	ldo	-1(arg2),arg2
	stw	arg2,ACT_REF_COUNT(t3)

	/*
	 * Return the kernel activation to the kernel ipc_target.
	 */
	ldil	L%kernel_port,arg2
	ldw	R%kernel_port(arg2),arg2
	ldw	IPT_ACTS(arg2),arg3
	stw	t1,IPT_ACTS(arg2)
	stw	arg3,ACT_IPT_NEXT(t1)

	/*
	 * Return to the client activation.
	 * XXX sp, dp
	 */
	ldw	ACT_PCB(t3),t1
	ldil	L%cur_pcb,t4
	stw	t1,R%cur_pcb(t4)

	/*
	 * Restore remaining state
	 */
	copy	ret0,t2
	copy	r13,t4
	copy	r14,arg0
	copy	r16,arg2
	copy	r17,r31
	mtsp	r18,sr1

	/*
	 * Done.  At this point:
	 *	r1:	thread
	 *	t2:	server activation
	 *	t3:	client activation (current)
	 *	t4:	ipc_target pointer
	 *	arg0:	client message buffer
	 *	arg2:	message size
	 *	r31:	receive port (kernel form)
	 */
	b,n	$gotact

$actwaiting
	/*
	 * At this point:
	 *	r1:	thread
	 *	t1:	current PCB
	 *	t3:	client activation (current)
	 *	arg0:	ipc_target pointer
	 *	ret0:	RPC return value
	 *	ret1:	maybe RPC return value?
	 * All but t1/arg0 must be preserved across the C call,
	 * so they are moved to caller-save registers.
	 */
	copy	r1,r14
	copy	t3,r16
	copy	ret0,r17
	copy	ret1,r18
	 
	/*
	 * Allocate a kernel activation
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4
	ldw	IPT_ACTS(t4),t1
	comb,<>,n t1,r0,$kact_available4

	/* XXX block */
	ldil	L%istackptr,t3
	ldw	R%istackptr(t3),sp
	stw	r0,R%istackptr(t3)
	.import panic_no_kstacks,code
	ldil	L%panic_no_kstacks,t2
	ldo	R%panic_no_kstacks(t2),t2
	bv	0(t2)
	ldo	FM_SIZE(sp),sp

$kact_available4

	/*
	 * Remove it from the ipc_target.
	 */
	ldw	ACT_IPT_NEXT(t1),ret1
	stw	ret1,IPT_ACTS(t4)

	/*
	 * Attach the kernel activation (t1) to the current thread (r1).
	 * The user activation in still in t3.
	 */
	stw	t3,ACT_LOWER(t1)
	stw	t1,ACT_HIGHER(t3)
	stw	t1,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t1)

	/*
	 * Take a reference on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t4
	ldo	1(t4),t4
	stw	t4,ACT_REF_COUNT(t3)
	stw	t3,ACT_USER_ACT(t1)

	/*
	 * Make the new activation officially the current one.
	 */
	ldw	ACT_PCB(t1),ret1
	ldil	L%cur_pcb,t4
	stw	ret1,R%cur_pcb(t4)

	/*
	 * Initialize the kernel stack pointer from the new activation
	 * and set the kernel data pointer.
	 */
	ldw	ACT_USER_STACK(t1),sp
	ldil	L%$global$,dp
	ldo	R%$global$(dp),dp

	/*
	 * ipc_target_wakeup(ipt)
	 */
	stw     r0,FM_PSP(sp)
	copy	t1,r15
	.import	ipc_target_wakeup,code
	ldil	L%ipc_target_wakeup,t1
	ldo	R%ipc_target_wakeup(t1),t1
	.call
	blr	r0,rp
	bv	0(t1)
	nop

	/*
	 * 
	 * Reload thread (r1), kernel activation (t1), client activation (t3).
	 */
	copy	r14,r1
	copy	r15,t1
	copy	r16,t3

	/*
	 * Remove the kernel activation from the thread.
	 */
	stw	r0,ACT_THREAD(t1)
	stw	t3,THREAD_TOP_ACT(r1)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t1)
	stw	r0,ACT_HIGHER(t3)
	stw	r0,ACT_USER_ACT(t1)
#endif

	/* 
	 * Drop the reference count on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),arg2
	ldo	-1(arg2),arg2
	stw	arg2,ACT_REF_COUNT(t3)

	/*
	 * Return the kernel activation to the kernel port.
	 */
	ldil	L%kernel_port,arg2
	ldw	R%kernel_port(arg2),arg2
	ldw	IPT_ACTS(arg2),arg3
	stw	t1,IPT_ACTS(arg2)
	stw	arg3,ACT_IPT_NEXT(t1)

	/*
	 * Return to the client activation.
	 * XXX sp, dp
	 */
	ldw	ACT_PCB(t3),t1
	ldil	L%cur_pcb,t4
	stw	t1,R%cur_pcb(t4)

	/*
	 * Restore remaining state
	 */
	copy	r17,ret0
	copy	r18,ret1

	/*
	 * Done.  At this point:
	 *	r1:	thread
	 *	t1:	current PCB
	 *	t3:	client activation (current)
	 *	ret0:	RPC return value
	 *	ret1:	maybe RPC return value
	 */
	b,n	$return_no_target
#endif

	.end
