/* 
 * Copyright (c) 1991,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: pdcsprt.c 1.16 94/12/14$
 *	Author: Leigh Stoller, Jeff Forys, University of Utah CSL
 */

/*
 * PDC support routines:
 *
 * pdcboot()	- load boot code from disk and execute.
 * pdcdump()	- dump kernel memory to disk.
 */

#ifdef MACH_KERNEL
#include <sys/types.h>
#include <device/param.h>
#include <machine/machparam.h>
#include <machine/iplhdr.h>
#else
#include <sys/param.h>
#include <hp800stand/iplhdr.h>
#endif
#include <machine/pdc.h>
#include <machine/iodc.h>
#include <machine/iomod.h>


#ifdef MACH_KERNEL
#define MIN(a,b)	(((a)<(b))?(a):(b))
#define roundup(x, y)   ((((x)+((y)-1))/(y))*(y))
#define	MINIOSIZ	64		/* minimum buffer size for IODC call */
#define	MAXIOSIZ	(64 * 1024)	/* maximum buffer size for IODC call */
#endif
#define	IOREQALN	2048		/* alignment for user read requests */

#define	BT_HPA		PAGE0->mem_boot.pz_hpa
#define	BT_SPA		PAGE0->mem_boot.pz_spa
#define	BT_LAYER	PAGE0->mem_boot.pz_layers
#define	BT_IODC		PAGE0->mem_boot.pz_iodc_io
#define	BT_CLASS	PAGE0->mem_boot.pz_class

#define	READLOOP(dev, mem, tot) {					\
	register u_int rdloc = (dev), wrloc = (mem), tcnt = (tot);	\
	register u_int n;						\
	do {								\
		if ((*btiodc)(BT_HPA, IODC_IO_BOOTIN, BT_SPA, BT_LAYER,	\
		              pdcret, rdloc, wrloc, MIN(tcnt,MAXIOSIZ),	\
		              MAXIOSIZ) < 0)				\
			goto bad;					\
		n = *pdcret; rdloc += n; wrloc += n; tcnt -= n;		\
	} while ((int)tcnt > 0);					\
}

/*
 * Load a boot program at `jefboot' (using the PDC) and execute it.
 * On failure, the hardware is reset causing the old, slow reboot.
 *
 * The grand assumption we make is that the boot device listed in
 * PAGE0 also houses a boot program.  We also assume that no one has
 * screwed with PAGE0 (which should be safe).
 *
 * This routine must be called in real mode, with interrupts disabled.
 * For a complete list of constraints, check the routine that calls it.
 */
pdcboot(interactive, jefboot)
	int interactive;
	int (*jefboot)();
{
	char iplbuf[sizeof(struct ipl_image) + 64 + IOREQALN];
	struct ipl_image *iplptr;
	int *pdcret, pdcbuf[33];
	int (*pdc)(), (*btiodc)();

	fcacheall();

	/*
	 * Get our PDC entry point and pointer to IODC boot code.
	 */
	pdc = (int (*)()) PAGE0->mem_pdc;
	if ((btiodc = BT_IODC) == 0)
		goto bad;

	/*
	 * Buffer alignment.
	 */
	iplptr = (struct ipl_image *) ((((u_int)iplbuf) + 63) & ~63);
	pdcret = (int *) ((((u_int)pdcbuf) + 7) & ~7);

	/*
	 * Read the boot device initialization code into memory,
	 * Initialize the boot module/device, and
	 * Load the boot device I/O code into memory.
	 */
	if ((*pdc)(PDC_IODC, PDC_IODC_READ, pdcret, BT_HPA, IODC_INIT,
	           btiodc, IODC_MAXSIZE) < 0)
		goto bad;

	if ((*btiodc)(BT_HPA, IODC_INIT_ALL, BT_SPA, BT_LAYER, pdcret,
	              0,0,0,0) < 0)
		goto bad;

	if ((*pdc)(PDC_IODC, PDC_IODC_READ, pdcret, BT_HPA, IODC_IO,
	           btiodc, IODC_MAXSIZE) < 0)
		goto bad;

	/*
	 * Load the IPL header from the disk, followed by the boot program.
	 * Calculate the entry point into the boot code.
	 */
	READLOOP(0, (u_int)iplptr, roundup(sizeof(struct ipl_image),IOREQALN));
	READLOOP(iplptr->ipl_addr, (u_int)jefboot, iplptr->ipl_size);
	jefboot = (int (*)()) ((u_int)jefboot + iplptr->ipl_entry);

	fcacheall();		/* for split I & D cache's (e.g. hp720) */
#ifndef MACH_KERNEL		/* TLB config data not saved at boot time */
	pgtlbs();		/* emulate PDC behavior as best we can */
#endif

	/*
	 * Launch the boot code!
	 */
	(*jefboot)(interactive);
	/*NOTREACHED*/

bad:
	/*
	 * If anything went wrong above, we end up here.  Since we trashed
	 * kernel memory (starting at `jefboot'), all we can do here is a
	 * hard reboot... which is what we would have done anyway!
	 */
	BT_IODC = (int (*)()) 0;
	(*(struct iomod *)LBCAST_ADDR).io_command = CMD_RESET;
}


#ifdef PDCDUMP
/*
 * Dump memory into the swap partition of the primary boot device. The
 * config code will compute the offset from the start of the disk, rather
 * then the start of the swap partition since this code does not know
 * anything about partitions.
 */

pdcdump()
{
	extern int dumpsize, dumpoffset;
	static int pdcbuf[33];
	int *pdcret = (int *) ((((u_int)pdcbuf) + 7) & ~7);
	int (*pdc)(), (*btiodc)();
#ifdef USELEDS
	extern int inledcontrol;
#endif

	if (dumpoffset == -1)		/* initialized? */
		goto bad;

	fcacheall();

	pdc = PAGE0->mem_pdc;
	if ((btiodc = BT_IODC) == 0)
		goto bad;

#ifdef USELEDS
	if (inledcontrol == 0)
		ledcontrol(0xAA, 0x55, 0);
#endif
	/* Read the initialization code */
	if ((*pdc)(PDC_IODC, PDC_IODC_READ, pdcret, BT_HPA, IODC_INIT,
	           (int) btiodc, IODC_MAXSIZE) < 0)
		goto bad;

	/* Initialize the module and the device */
	if ((*btiodc)(BT_HPA, IODC_INIT_ALL, BT_SPA, BT_LAYER, pdcret,
	              0,0,0,0) < 0)
		goto bad;

	/* Read the the io code */
	if ((*pdc)(PDC_IODC, PDC_IODC_READ, pdcret, BT_HPA, IODC_IO,
	           (int) btiodc, IODC_MAXSIZE) < 0)
		goto bad;

	/*
	 * Write memory to disk; write it all at once if possible,
	 * o/w loop until done.
	 *
	 * PDC restrictions:
	 *	dumpsize - multiple of 2K bytes (guaranteed)
	 *	dumpoffset - 2K byte aligned (not enforced!)
	 */
	{
		register u_int rdloc = 0;
		register u_int wrloc = dumpoffset * DEV_BSIZE;
		register u_int tcnt = ctob(dumpsize);
		register u_int n;

		do {
#ifdef USELEDS
			if (inledcontrol == 0)
				ledcontrol(0, 0, 0xFF);
#endif
			if ((*btiodc)(BT_HPA,IODC_IO_BOOTOUT,BT_SPA,BT_LAYER,
			              pdcret,wrloc,rdloc,MIN(tcnt,MAXIOSIZ))<0)
				goto bad;
			n = *pdcret; rdloc += n; wrloc += n; tcnt -= n;
		} while ((int)tcnt > 0);
	}
#ifdef USELEDS
	if (inledcontrol == 0)
		ledcontrol(0, 0xFF, 0);
#endif
bad:
	/* reboot */
	(*(struct iomod *)LBCAST_ADDR).io_command = CMD_RESET;
}
#endif

#ifdef PDCPRINTF
/*
 * Use the PDC to do kernel (console) putchars.
 * XXX this used to be ifdef'ed under HPLMACH, is it still useful?
 */
#ifdef MACH_KERNEL
#include <cons.h>
#else
#include <hpdev/cons.h>
#endif

#define	CN_HPA		PAGE0->mem_cons.pz_hpa
#define	CN_SPA		PAGE0->mem_cons.pz_spa
#define	CN_LAYER	PAGE0->mem_cons.pz_layers
#define	CN_IODC		PAGE0->mem_cons.pz_iodc_io
#define	CN_CLASS	PAGE0->mem_cons.pz_class

static	int pdcbuf[33];			/* PDC return buffer */
static	int *pdcret = 0;		/* double-aligned ptr into `pdcbuf' */
static	int (*cniodc)() = (int (*)())0;	/* console IODC entry point */
static	char cnbuf[MINIOSIZ * 2];	/* char output buffer */
static	char *cndata = 0;		/* double-aligned ptr into `cnbuf' */

void
pdc_cnputc(c)
{
	if (CN_IODC == 0)
		return;
	*cndata = c;
	(void) (*cniodc)(CN_HPA, IODC_IO_CONSOUT, CN_SPA, CN_LAYER,
	                 pdcret, 0, cndata, 1, 0);
}

void
pdc_cninit(cp)
	struct consdev *cp;
{
	pdcret = (int *) ((((int) pdcbuf) + 7) & ~7);
	cndata = (char *) ((((int) &cnbuf[0]) + MINIOSIZ-1) & ~(MINIOSIZ-1));
	cniodc = (int (*)()) PAGE0->mem_free;
	cp->cn_putc = (int (*)()) pdc_cnputc;
}
#endif
