/*
 *  (c) Copyright 1989 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Copyright (c) 1990,1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * THE UNIVERSITY OF UTAH AND CSL PROVIDE THIS SOFTWARE IN ITS "AS IS"
 * CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM ITS USE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: trap.c 1.56 95/08/23$
 */

#include <mach/thread_status.h>
#include <mach/exception.h>
#include <mach/vm_param.h>
#include <kern/ast.h>
#include <kern/task.h>
#include <kern/thread.h>
#include <kern/act.h>
#include <kern/current.h>
#include <vm/vm_map.h>
#include <machine/trap.h>
#include <machine/psl.h>
#include <machine/regs.h>
#include <machine/machspl.h>
#include <machine/break.h>
#include <mach_kdb.h>
#include <machine/pdc.h>
#include <machine/iomod.h>
#include <debug.h>

#ifdef KGDB
#ifdef KGDB_LAN
#include <machine/setjmp.h>
extern jmp_buf_t *kgdb_recover;
extern boolean_t kgdb_initialized;
#define kgdb_enabled kgdb_initialized
#else
extern int kgdb_enabled;
#endif
#endif

#define DEBUG
#ifdef DEBUG
/* catch failures from server or init early */
#if 0
#define BLOWCHUNKS (exception == EXC_BAD_ACCESS && ssp->sr4 > 0 && ssp->sr4 < 8)
#else
#define BLOWCHUNKS 0
#endif
#endif

int exit_from_int = 0;		/* see interrupt.s */

/*
 * XXX don't pass VM_PROT_EXECUTE to vm_fault(), execute permission is implied
 * in either R or RW (note: the pmap module knows this).  This is done for the
 * benefit of programs that execute out of their data space (ala lisp).
 * If we didn't do this in that scenerio, the ITLB miss code would call us
 * and we would call vm_fault() with RX permission.  However, the space was
 * probably vm_allocate()ed with just RW and vm_fault would fail.  The "right"
 * solution to me is to have the un*x server always allocate data with RWX for
 * compatibility with existing binaries.
 */
#define	PROT_EXEC	(VM_PROT_READ)
#define PROT_RO		(VM_PROT_READ)
#define PROT_RW		(VM_PROT_READ|VM_PROT_WRITE)

#define FAULT_TYPE(op) (OPCODE_IS_LOAD(op) ? PROT_RO : PROT_RW)

/*
 * alignment restrictions for load and store instructions
 */

static int align_index[] = {
	0x0,		/* no alignment	    */
	0x1,		/* 2-byte alignment */
	0x3,		/* 4-byte alignment */
	0xf		/* f-byte alignment */
	};

static int align_ldst[] = {
	0x0,		/* no alignment	    */
	0x1,		/* 2-byte alignment */
	0x3,		/* 4-byte alignment */
	0x3		/* 4-byte alignment */
	};

static int align_coproc[] = {
	0x3,		/* 4-byte alignment */
	0x7		/* 8-byte alignment */
	};

#define OPCODE_UNALIGNED(opcode, offset) \
	(((((opcode) & 0xd0000000) == 0x40000000) && \
	  (offset & align_ldst[opcode >> 26 & 0x03])) || \
	 ((((opcode) & 0xf4000200) == 0x24000200) && \
	  (offset & align_coproc[opcode >> 27 & 0x01])) || \
	 ((((opcode) & 0xfc000200) == 0x0c000200) && \
	  (offset & align_index[opcode >> 6 & 0x03])))

/*
 * interrupt switch table
 */
int_illegal(int_bit, ssp)
	int int_bit;
	struct save_state *ssp;
{
	printf("Unexpected interrupt on EIRR bit %d\n",int_bit);
}
ignore_int(int_bit, ssp)
	int int_bit;
	struct save_state *ssp;
{
}
struct intrtab itab_proc[] = {
	{ ignore_int,	INTPRI_00 },	/* hard clock interrupt */
	{ int_illegal,	INTPRI_01 },
	{ int_illegal,	INTPRI_02 },
	{ int_illegal,	INTPRI_03 },
	{ int_illegal,	INTPRI_04 },
	{ int_illegal,	INTPRI_05 },
	{ int_illegal,	INTPRI_06 },
	{ int_illegal,	INTPRI_07 },
	{ int_illegal,	INTPRI_08 },
	{ int_illegal,	INTPRI_09 },
	{ int_illegal,	INTPRI_10 },
	{ int_illegal,	INTPRI_11 },
	{ int_illegal,	INTPRI_12 },
	{ int_illegal,	INTPRI_13 },
	{ int_illegal,	INTPRI_14 },
	{ int_illegal,	INTPRI_15 },
	{ int_illegal,	INTPRI_16 },
	{ int_illegal,	INTPRI_17 },
	{ int_illegal,	INTPRI_18 },
	{ int_illegal,	INTPRI_19 },
	{ int_illegal,	INTPRI_20 },
	{ int_illegal,	INTPRI_21 },
	{ int_illegal,	INTPRI_22 },
	{ int_illegal,	INTPRI_23 },
	{ int_illegal,	INTPRI_24 },
	{ int_illegal,	INTPRI_25 },
	{ int_illegal,	INTPRI_26 },
	{ int_illegal,	INTPRI_27 },
	{ int_illegal,	SPL0 },
	{ int_illegal,	SPL0 },
	{ int_illegal,	SPL0 },
	{ int_illegal,	SPL0 }
};

/*
 * names of traps for panic calls
 */

static char *trap_type[] = {
	"Non-existent trap",
	"High priority machine check",
	"Power failure trap",
	"Recover counter overflow trap",
	"External interrupt",
	"Low priority machine check",
	"Instruction page fault",
	"Instruction memory protection fault",
	"Illegal instruction trap",
	"Break instruction trap",
	"Privileged operation trap",
	"Privileged register trap",
	"Overflow",
	"Conditional trap",
	"Assist exception trap",
	"Data page fault",
	"Non-access instruction TLB miss",
	"Non-access data TLB miss",
	"Data memory protection fault",
	"Data memory break",
	"TLB dirty",
	"Page reference trap",
	"Assist Emulation trap",
	"Higher privilege transfer",
	"Lower privilege transfer",
	"Taken branch trap",
	"Data memory access fault",
	"Data memory protection ID fault",
	"Data alignment fault",
	"Interrupt stack overflow",
	"Kernel stack overflow",
        "Interrupt stack lost"
};

#define	TRAP_TYPES	(sizeof trap_type / sizeof trap_type[0])

extern vm_map_t kernel_map;

extern int mfctl();
extern void mtctl();

#ifdef FTRACE_SPACE
pa_space_t	ftspace = FTRACE_SPACE;
#endif

struct parisc_saved_state reg_state;

/*
 * Set a bit in debug_exceptions matching the exceptions you want to trace.
 */
unsigned int debug_exceptions = 0;

void
doexception(exc, code, sub)
	int exc, code, sub;
{
	exception(exc, code, sub);
}

/*
 * Locate and return the map associated with the space/offset in ssp.
 *
 * Note that up to three different levels of activations are involved here:
 *	1. the "trap" activation we are in right now,
 *	2. the "destination" activation (kernel for copyin/out, server for RPC)
 *	3. the "source" activation (user for copyin/out, client for RPC).
 * A copyin/out is just a special-case of a cross-address space copy so we
 * treat them the same.
 *
 * Note that #2 might be missing for an RPC where it faults prior to allocating
 * a server activation.
 *
 * For RPC/copy faults, the thread recovery pointer will always be set.
 * In that case the faulting space must be either current_act() (the server)
 * or current_act()->lower (the client).
 *
 * XXX gak!  An exception is the old mach_msg compatible interface which uses
 * rpc_copy_{request,reply}.  When doing copy in or out to the client we need
 * to use current_act()->lower->lower.  Since these routines set user_act we
 * can always just use current_act()->mact.user_act when it is set.
 *
 * For anything else the faulting space must be from current_act().
 */
vm_map_t
locate_user_map(ssp)
	struct parisc_saved_state *ssp;
{
	vm_map_t map;
	pa_space_t space = (pa_space_t) ssp->isr;
	Act *act = current_act();

	/*
	 * User -> server -> kernel
	 * (RPC or kernel copyin/out)
	 */
#ifndef ROLLBACK
	if (current_thread()->mthread.recover && act->lower) {
#else
	if (act->lower) {
#endif
		/*
		 * First check the source.
		 * See XXX above for explanation of user_act check.
		 */
		if (act->mact.user_act)
			map = act->mact.user_act->task->map;
		else
			map = act->lower->task->map;
		if (map->pmap->space != space) {
			/*
			 * Not the source, try the destination
			 */
			if ((map = act->task->map)->pmap->space != space) {
				/*
				 * Neither one, panic!
				 */
#if 1
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
#endif
				printf("Illegal kernel space ref #1 (%x.%x)\n",
				       space, ssp->ior);
				panic("");
			}
#if 0
			printf("locate_user_map: hit destination space\n");
			dump_act(cur_pcb->ss.act);
			dump_ssp(ssp);
#endif
		}
	}
	/*
	 * Other
	 * (can this be anything but an error now?)
	 */
	else {
		if ((map = act->task->map)->pmap->space != space) {
#if 1
			dump_act(cur_pcb->ss.act);
			dump_ssp(ssp);
#endif
			printf("Illegal kernel space ref #2 (%x.%x)\n",
			       space, ssp->ior);
			panic("");
		}
#if 1
		printf("Unexpected kernel->user space reference:\n");
		dump_act(cur_pcb->ss.act);
		dump_ssp(ssp);
#endif
	}
	return (map);
}

void trap(type, ssp)
	int type;
	struct parisc_saved_state *ssp;
{
	int exception;
	int code;
	int subcode;
	vm_offset_t offset;
	pa_space_t space;
	unsigned  opcode;
	vm_map_t map;
	int x;		/* spl level at entry */

#if 0
	if (USERMODE(ssp)) {
		/*printf("trap 0x%02x from user mode\n", type)*/;
	} else {
		extern vm_offset_t zdata;
		
		if (ssp->r30 >= (unsigned)zdata) {
			dump_ssp(ssp);
			panic("trap: are we having fun yet??");
		}
		/*printf("trap 0x%02x from kernel mode\n", type)*/;
	}
#endif
#if 0
	printf("Made it to trap. type = %d\n", type);
	dump_ssp(ssp);
#endif
#if 0
	if (USERMODE(ssp)) {
		printf("Made it to user trap. type = %d\n", type);
		dump_ssp(ssp);
	}
#endif

	/*
	 * mark the saved state structure with the trap we came in on
	 * for debugging...
	 */
	ssp->flags |= type << 16;

#ifdef DEBUG
	if ((ssp->ipsw & PSW_Q) == 0) {
		printf("type = %d\n", type);
		dump_ssp(ssp);
		panic("Unexpected trap with PSW-Q = 0");
	}

	if (!current_thread()) {
		printf("type = %d, cur_pcb = %x, act = %x, thread = %x\n",
		       type, cur_pcb, cur_pcb ? (int)cur_pcb->ss.act : -1,
		       cur_pcb ? (int)cur_pcb->ss.act->thread : -1);
		dump_act(cur_pcb->ss.act);
		dump_ssp(ssp);
		if (current_act()->lower)
			dump_ssp(&current_act()->lower->mact.pcb->ss);
		panic("trap in boot phase");	/* XXX was halt */
	}
#endif

	/*
	 * if the tail is in privileged mode make sure the head is too 
	 */

	if (!USERMODE_PC(ssp->iioq_tail))
		ssp->iioq_head &= ~PC_PRIV_USER;

	/*
	 * clear exception to indicate that we haven't found one yet
	 */
	exception = 0;

	/*
 	 * Entered at spl7. We will lower the spl level to the level it was
 	 * when we took the trap with a minimum level of spl0.
	 */
	x = splx((int)ssp->eiem);

	/*
	 * handle kernel traps first.
	 */
	if (!USERMODE(ssp)) {

		switch (type) {
			
			/*
			 * These trap types should never be seen by trap(). 
			 * Most are interrupts that should be seen by 
			 * interrupt() others just don't happen because they 
			 * are handled in locore. Some could happen but are
			 * considered to be fatal in kernel mode.
			 */
		case I_NONEXIST:      /* shouldn't happen */
		case I_HPMACH_CHK:    /* handled as interrupt */
		case I_POWER_FAIL:    /* handled as interrupt */
		case I_EXT_INTP:      /* handled as interrupt */
		case I_LPMACH_CHK:    /* handled as interrupt */
		case I_DPGFT_NA:      /* converted to DPGFT in locore */
		case I_IMEM_PROT:     /* fatal for kernel */
		case I_UNIMPL_INST:   /* fatal for kernel */
		case I_PRIV_OP:       /* fatal for kernel */
		case I_PRIV_REG:      /* fatal for kernel */
		case I_OVFLO:         /* fatal for kernel */
		case I_EXCEP:         /* fatal for kernel */
		case I_TLB_DIRTY:     /* handled in locore */
		case I_EMULAT:        /* fatal for kernel */
		case I_IS_OVFL:       /* fatal for kernel */
		case I_KS_OVFL:       /* fatal for kernel */
		case I_IS_LOST:       /* fatal for kernel */
		case I_LPRIV_XFR:     /* shouldn't happen */
		default:
			dump_act(cur_pcb->ss.act);
			dump_ssp(ssp);
			printf("Illegal trap in kernel mode\n");
			if ((unsigned)type < TRAP_TYPES)
				panic(trap_type[type]);
			else
				panic("Unknown trap type 0x%x", type);
			break;
			
		case I_RECOV_CTR:
			printf("Kernel mode R-bit trap\n");
			dump_ssp(ssp);
			ssp->ipsw &= ~PSW_R;
			break;

		case I_IPGFT:
		case I_IPGFT_NA:

			if (type == I_IPGFT_NA) {
				offset = (vm_offset_t) ssp->ior;
				space = (pa_space_t) ssp->isr;
			} else {
				offset = (vm_offset_t) ssp->iioq_head;
				space = (pa_space_t) ssp->iisq_head;
			}

			/*
			 * We don't want to run any user code as the kernel...
			 */
			
			if (space != 0)
			{
				dump_ssp(ssp);
				panic("Kernel instruction reference in user space");
			}

			code = vm_fault(kernel_map, trunc_page(offset),
					PROT_EXEC, FALSE, 
					FALSE, (void (*)()) 0);
#ifdef FTRACE_SPACE
			if (space == ftspace)
			printf("type %d: vm_fault(%x, %x(%x), 5, 0) -> %d\n",
			       type+100, kernel_map, trunc_page(offset),
			       offset, code);
#endif
			if (code == KERN_SUCCESS)
				break;

			panic("Kernel instruction page fault unresolved");
			break;

		case I_DMEM_ACC:
		case I_DMEM_PID:
		case I_UNALIGN:
			/*
			 * These traps replace I_DMEM_PROT (T-chip only).
			 */
			/* fall through */

		case I_DMEM_PROT:
			/*
			 * Unaligned kernel data access, die.
			 */
			if (OPCODE_UNALIGNED(ssp->iir, ssp->ior)) {
				if (space != PARISC_SID_KERNEL &&
				    current_thread()->mthread.recover)
					goto recover;
				printf("trap(%x): thrd %x, task %x\n",
		 		        type, current_thread(), current_task());
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
				panic("Unaligned kernel data access");
			}
			/* else fall into... */

		case I_DPGFT:
#ifdef KGDB_LAN
			if (kgdb_recover) {
			printf("trap: data page/protection fault in debugger, kva=%x.%x\n", ssp->isr, ssp->ior);
				_longjmp(kgdb_recover, 1);
			}
#endif
			offset = (vm_offset_t) ssp->ior;
			space = (pa_space_t) ssp->isr;
			opcode = ssp->iir;

			/*
			 * Copyin/copyout operations will appear as kernel
			 * faults.  They are detected by comparing the
			 * faulting space with that of the current task.
			 */
			if (space == PARISC_SID_KERNEL)
				map = kernel_map;
			else
				map = locate_user_map(ssp);

			/*
			 * Normal data page/protection fault, call vm_fault
			 * to deal with it.
			 */
			code = vm_fault(map, trunc_page(offset),
					FAULT_TYPE(opcode), FALSE,
					FALSE, (void (*)())0);
#ifdef FTRACE_SPACE
			if (space == ftspace)
			printf("type %d: vm_fault(%x, %x(%x), %x, 0) -> %d\n",
			       type+100, map, trunc_page(offset), offset,
			       FAULT_TYPE(opcode), code);
#endif
			if (code == KERN_SUCCESS)
				break;

			/*
			 * If this was a user access and there is a recovery
			 * pointer, the kernel was anticipating a problem with
			 * this user access.  Return to the recovery location.
			 *
			 * XXX we put the faulting address in ret0 for easy
			 * access by the catch routine.
			 */
			if (space != PARISC_SID_KERNEL &&
			    current_thread()->mthread.recover) {
recover:
#if 1
				printf("trap: user %s, addr=%x.%x, th_r=%x\n",
				       type == I_DPGFT ? "DPGFT" : "DMEM",
				       space, offset,
				       current_thread()->mthread.recover);
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
#endif
				ssp->iioq_head = (int) current_thread()->mthread.recover;
				ssp->iioq_tail = (int) current_thread()->mthread.recover + 4;
				current_thread()->mthread.recover = 0;
				ssp->ret0 = offset;
				break;
			}

			/*
			 * An unresolved kernel fault, die.
			 */
			dump_act(cur_pcb->ss.act);
			dump_ssp(ssp);
#if 1
			if (current_act()->lower) {
				dump_ssp(&current_act()->lower->mact.pcb->ss);
				if (current_act()->lower->lower)
					dump_ssp(current_act()->lower->lower->mact.pcb->ss);
			}
#endif
			if (type == I_DPGFT)
				panic("Kernel data page fault unresolved");
			else
				panic("Kernel data protection fault");
			/*NOTREACHED*/
			break;

		case I_COND:
			panic("Integer divide by zero in kernel");
			break;

		case I_BRK_INST:

			if (ssp->iisq_head == 0 && ssp->iioq_head == 0) {
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
				panic("kernel jump to zero");
			}

			if (break5(ssp->iir) != BREAK_KERNEL) {
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
				if (current_act()->lower)
					dump_ssp(&current_act()->lower->mact.pcb->ss);
				panic("unknown break instruction in kernel");
			}

			switch(break13(ssp->iir)) {

			case BREAK_KERNPRINT:
				kern_print(type, ssp);
				break;

#if    MACH_KDB
			case BREAK_MACH_DEBUGGER:
				kdb_trap(ssp, 1);
				break;
#endif MACH_KDB

#ifdef KGDB
			/* In kernel mode, send both the standard
			   GDB breakpoint and KGDB breakpoint to
			   kgdb_trap.  */
			case BREAK_GDB:
				/*
				 * Debugging.
				 * Dump state and continue.
				 */
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
				ssp->iioq_head = ssp->iioq_tail;
				ssp->iioq_tail = ssp->iioq_head + 4;
				break;

			case BREAK_KGDB:
				kgdb_trap(type, ssp);
				break;
#else
			case BREAK_GDB:
				/*
				 * Debugging.
				 * Dump state and continue.
				 */
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
				ssp->iioq_head = ssp->iioq_tail;
				ssp->iioq_tail = ssp->iioq_head + 4;
				break;
#endif
#ifdef ACTWATCH
			case 11:
				actwatch(ssp);
				ssp->iioq_head = ssp->iioq_tail;
				ssp->iioq_tail = ssp->iioq_head + 4;
				splx(x);
				return;
#endif
			default:
				dump_ssp(ssp);
				panic("unknown break instruction in kernel");
			}

			break;


		case I_TAKEN_BR:
#ifdef KGDB
			if (kgdb_enabled) {
				kgdb_trap(type, ssp);
				break;
			}
			/* fall into... */
#endif KGDB
		case I_DMEM_BREAK:
		case I_PAGE_REF:
		case I_HPRIV_XFR:
#ifdef DEBUG
			printf("unexpected trap %d, ignored\n", type);
			dump_ssp(ssp);
#endif
			break;
		}
	} else {
		/*
		 * The thread was running in user mode when exception occured
		 */

		switch (type) {

			/*
			 * These trap types should never be seen by trap(). 
			 * Most are interrupts that should be seen by 
			 * interrupt() others just don't happen because they 
			 * are handled in locore.
			 */

		case I_NONEXIST:	/* shouldn't happen */
		case I_HPMACH_CHK:	/* handled as interrupt */
		case I_POWER_FAIL:	/* handled as interrupt */
		case I_EXT_INTP:	/* handled as interrupt */
		case I_LPMACH_CHK:	/* handled as interrupt */
		case I_DPGFT_NA:	/* converted to DPGFT in locore */
		case I_IS_OVFL:		/* handled as interrupt */
		case I_KS_OVFL:		/* handled as interrupt */
		case I_IS_LOST:		/* handled as interrupt */

		default:
			dump_ssp(ssp);
			printf("Illegal trap in user mode\n");
			if (type < TRAP_TYPES)
				panic(trap_type[type]);
			else
				panic("Unknown trap type 0x%x", type);
			break;
			
		case I_RECOV_CTR:	/* single step */
		{
			if (current_act()->mact.sstep) {
#if 0
				printf("Warning: sstep = %d\n",
				       current_act()->mact.sstep);
#endif
				ssp->cr0 = 0;
				exception = EXC_BREAKPOINT;
				code = EXC_PARISC_BREAKPOINT;
				subcode = ssp->iir;
				break;
			}
			printf("User mode R-bit trap\n");
			dump_ssp(ssp);
			ssp->ipsw &= ~PSW_R;
			break;
		}
			/* fall into... */
		case I_LPRIV_XFR:	/* AST delivery from syscall */
			/*
			 * just need a hook, this will drop into AST delivery
			 * below
			 */
#ifdef DEBUG
			if ((ssp->ipsw & PSW_L) == 0 &&
			    ((ssp->flags & 0xffff) != SS_INTRAP ||
			     ssp->iisq_head == 0 ||
			     (ssp->iisq_tail == 0 &&
			      (ssp->iioq_tail & 0xFFFFF000) != 0xC0000000) ||
			     (ssp->iioq_head & 3) != 3 ||
			     (ssp->iioq_tail & 3) != 3)) {
				printf("interrupt AST: bad state\n");
				dump_ssp(ssp);
			} else
#endif
			ssp->ipsw &= ~PSW_L;
			break;

		case I_IPGFT:
		case I_IPGFT_NA:
			if (type == I_IPGFT_NA) {
				offset = (vm_offset_t) ssp->ior;
				space = (pa_space_t) ssp->isr;
			} else {
				offset = (vm_offset_t) ssp->iioq_head;
				space = (pa_space_t) ssp->iisq_head;
			}

			map = current_map();
			if (space != map->pmap->space) 
				code = EXC_PARISC_BADSPACE;
			else {
				
				/*
				 * normal instruction page fault
				 */
				
				code = vm_fault(map, trunc_page(offset),
						PROT_EXEC, FALSE,
						FALSE, (void (*)()) 0);
#ifdef FTRACE_SPACE
				if (space == ftspace)
					printf("type %d: vm_fault(%x, %x(%x), 5, 0) -> %d\n",
					       type, map, trunc_page(offset), offset, code);
#endif
				if (code == KERN_SUCCESS)
					break;
			}
			
			exception = EXC_BAD_ACCESS;
			subcode = offset;
			break;

		case I_IMEM_PROT:
			offset = (vm_offset_t) ssp->iioq_head;
			space = (pa_space_t) ssp->iisq_head;
			
			map = current_map();

			if (space != map->pmap->space) 
				code = EXC_PARISC_BADSPACE;
			else {
				
				code = vm_fault(map, trunc_page(offset),
						PROT_EXEC, FALSE,
						FALSE, (void (*)()) 0);
#ifdef FTRACE_SPACE
				if (space == ftspace)
					printf("type %d: vm_fault(%x, %x(%x), 5, 0) -> %d\n",
					       type, map, trunc_page(offset), offset, code);
#endif
				if (code == KERN_SUCCESS)
					break;
			}
			
			exception = EXC_BAD_ACCESS;
			subcode = offset;
			break;

		case I_UNIMPL_INST:
			exception = EXC_BAD_INSTRUCTION;
			code = EXC_PARISC_UNIPL_INST;
			subcode = ssp->iioq_head;
			break;

		case I_PRIV_OP:
			exception = EXC_BAD_INSTRUCTION;
			code = EXC_PARISC_PRIVINST;
			subcode = ssp->iioq_head;
			break;
		
		case I_PRIV_REG:
			exception = EXC_BAD_INSTRUCTION;
			code = EXC_PARISC_PRIVREG;
			subcode = ssp->iioq_head;
			break;

		case I_OVFLO:
			exception = EXC_ARITHMETIC;
			code = EXC_PARISC_OVERFLOW;
			subcode = ssp->iioq_head;
			fpu_flush();
			break;
			
		case I_EXCEP:
			exception = EXC_ARITHMETIC;
			code = EXC_PARISC_FLT_INEXACT;
			subcode = ssp->iioq_head;
			fpu_flush();
			/*
			 * XXX - fill in the right type of coprocessor 
			 * exception in code (so that UNIX can set them all to 
			 * floating point exception :-) )
			 */
			break;

		case I_COND:
			exception = EXC_ARITHMETIC;
			code = EXC_PARISC_ZERO_DIVIDE;
			subcode = ssp->iioq_head;
			fpu_flush();
			break;

		case I_DMEM_ACC:
		case I_DMEM_PID:
		case I_UNALIGN:
			/*
			 * These traps replace I_DMEM_PROT (T-chip only).
			 */
			/* fall through */

		case I_DMEM_PROT:
			space = (pa_space_t) ssp->isr;
			opcode = ssp->iir;
			if (OPCODE_UNALIGNED(ssp->iir, ssp->ior)) {
				code = EXC_PARISC_UNALIGNED;
				exception = EXC_BAD_ACCESS;
				subcode = (int) ssp->ior;
				break;
			}
			/* else fall into... */

		case I_DPGFT:
			offset = (vm_offset_t) ssp->ior;
			space = (pa_space_t) ssp->isr;
			opcode = ssp->iir;

			map = current_map();

			if (space != map->pmap->space)
				code = EXC_PARISC_BADSPACE;
			else {
				/*
				 * Normal data page/protection fault, 
				 * call vm_fault to deal with it.
				 */
				code = vm_fault(map, trunc_page(offset),
						FAULT_TYPE(opcode), FALSE,
						FALSE, (void (*)()) 0);
#ifdef FTRACE_SPACE
				if (space == ftspace)
					printf("type %d: vm_fault(%x, %x(%x), %x, 0) -> %d\n",
					       type, map, trunc_page(offset), offset,
					       FAULT_TYPE(opcode), code);
#endif
				if (code == KERN_SUCCESS)
					break;
			}
			exception = EXC_BAD_ACCESS;
			subcode = offset;
			break;

		case I_EMULAT:
			exception = EXC_ARITHMETIC;
			code = EXC_PARISC_NOEMULATION;
			subcode = ssp->iioq_head;
			fpu_flush();
			break;

		case I_TLB_DIRTY:
			panic("TLB dirty trap");
			break;

		case I_BRK_INST:

			/*
			 * if this isn't a kernel break then pass it up to 
			 * the server
			 */

			if (break5(ssp->iir) != BREAK_KERNEL) {
				exception = EXC_BREAKPOINT;
				code = EXC_PARISC_BREAKPOINT;
				subcode = ssp->iir;
				break;
			}

			switch(break13(ssp->iir)) {

			case BREAK_KERNPRINT:
				kern_print(type, ssp);
				break;

#ifdef KGDB
			/* Insert a kernel breakpoint in the server
			   and you'll come here.  Insert a standard
			   breakpoint and you'll send an exception
			   to the server.  */
			case BREAK_KGDB:
				if (kgdb_enabled)
					kgdb_trap(type, ssp);
				else {
					exception = EXC_BREAKPOINT;
					code = EXC_PARISC_BREAKPOINT;
					subcode = ssp->iir;
				}
				break;
#endif KGDB
#ifdef ACTWATCH
			case 11:
				actwatch(ssp);
				ssp->iioq_head = ssp->iioq_tail;
				ssp->iioq_tail = ssp->iioq_head + 4;
				splx(x);
				return;
#endif
			case 8:
				/*
				 * hook for mach kdb
				 *
				 * fall through for now...
				 */
				dump_act(cur_pcb->ss.act);
				dump_ssp(ssp);
				ssp->iioq_head = ssp->iioq_tail;
				ssp->iioq_tail = ssp->iioq_head + 4;
				break;

#if    MACH_KDB
			case BREAK_MACH_DEBUGGER:
				kdb_trap(ssp, 1);
				break;
#endif MACH_KDB

			case BREAK_THREAD_SETSTATE:
			{
				struct parisc_thread_state tss;

				if (copyin(ssp->arg0, &tss, sizeof tss) == 0) {
					(void)act_machine_set_state(
						current_act(),
						PARISC_THREAD_STATE,
						(void*)&tss,
						PARISC_THREAD_STATE_COUNT);
				} else {
					/*
					 * XXX advance over the break and
					 * keep going.
					 */
					ssp->iioq_head = ssp->iioq_tail;
					ssp->iioq_tail = ssp->iioq_head + 4;
				}
				break;
			}

			default:
				exception = EXC_BREAKPOINT;
				code = EXC_PARISC_BREAKPOINT;
				subcode = ssp->iir;
				break;
			}

			break;

		case I_DMEM_BREAK:
		case I_PAGE_REF:
		case I_HPRIV_XFR:
#ifdef DEBUG
			printf("unexpected trap %d, ignored\n", type);
			dump_ssp(ssp);
#endif
			break;

		case I_TAKEN_BR:
			exception = EXC_BREAKPOINT;
			code = EXC_PARISC_TAKENBRANCH;
			subcode = ssp->iir;
			break;
		}
	}

	if (exception) {
#ifdef DEBUG
if (BLOWCHUNKS) {
	printf("trap(%x): thread %x, task %x, exc %x, code %x, sub %x, ssp\n",
	       type, current_thread(), current_task(), exception, code, subcode);
	dump_act(cur_pcb->ss.act);
	dump_ssp(ssp);
	if (current_act()->lower)
		dump_ssp(&current_act()->lower->mact.pcb->ss);
#if 0
	panic("yow!");
#endif
}
#endif
		/*
		 * If proper bit set in debug_exceptions, show \'em what happened.
		 */
		if ((1 << type) & debug_exceptions) {
			printf("trap(%x): thread %x, task %x, exc %x, code %x, sub %x, ssp\n",
			       type, current_thread(), current_task(), exception, code, subcode);
			dump_ssp(ssp);
			if (exception == EXC_ARITHMETIC)
				dump_fp((struct parisc_fp_state *)&ssp[1]);
		}

		doexception(exception, code, subcode);
	}
	splx(x);	/* restore entry spl level */

	if (!exception) {

		/*
		 * ast delivery 
		 * Check if we need an AST and if so then we'll take care of it here...
		 */
		if (ast_needed(cpu_number()) && USERMODE(ssp))
			ast_taken();

	}

#if 0
	printf("leaving trap %02x from %s mode in pcb %08x thread %08x\n",
		type, USERMODE(ssp) ? "user" : "kernel", cur_pcb, current_thread());
#endif
#ifdef TRAP_STACK_CHECK
	{
		extern int trap_stack_check(int ret);
		trap_stack_check(0);
	}
#endif TRAP_STACK_CHECK
}

/*
 * interrupt handler 
 */

void interrupt(type, ssp)
	int type;
	struct parisc_saved_state *ssp;
{
#if 0
	if (USERMODE(ssp)) {
		/*printf("interrupt %08x from user mode\n", type)*/;
	} else {
		extern vm_offset_t zdata, end;
		
		if (ssp->r30 >= (unsigned)zdata && ssp->r30 < (unsigned)&end) {
			dump_ssp(ssp);
			printf("kernel bss: [%x-%x]\n", zdata, &end);
			panic("interrupt: are we having fun yet??");
		}
	}
#endif
	if ((ssp->ipsw & PSW_Q) == 0) {
		printf("\ninterrupt(type = %x)\n", type);
		printf("eiem = %x, eirr = %x\n", 
		       mfctl(CR_EIEM), mfctl(CR_EIRR));
		panic("Unexpected interrupt with PSW-Q = 0");
	}

	/*
	 * mark the saved state structure with the interrupt we came in on
	 * for debugging...
	 */
	ssp->flags |= type << 16;

	switch (type) {
			
		/*
		 * These trap types should never be seen by interrupt(). 
		 * Most are traps that should be seen by trap(). Others just 
		 * don't happen because they are handled in locore. Some 
		 * could happen but are considered to be fatal in kernel mode.
		 */

	case I_DPGFT:         /* fatal on interrupt stack */
	case I_DMEM_PROT:     /* fatal on interrupt stack */
	case I_DMEM_ACC:      /* fatal on interrupt stack */
	case I_DMEM_PID:      /* fatal on interrupt stack */
	case I_UNALIGN:       /* fatal on interrupt stack */
#ifdef KGDB_LAN
		if (kgdb_recover) {
			printf("interrupt: data page/protection fault in debugger, kva=%x.%x\n", ssp->isr, ssp->ior);
			_longjmp(kgdb_recover, 1);
		}
		/* fall into... */
#endif
	case I_NONEXIST:      /* shouldn't happen */
	case I_HPMACH_CHK:    /* fatal */
	case I_POWER_FAIL:    /* fatal */
	case I_LPMACH_CHK:    /* fatal */
	case I_DPGFT_NA:      /* converted to DPGFT in locore */
	case I_IMEM_PROT:     /* fatal on interrupt stack */
	case I_UNIMPL_INST:   /* fatal on interrupt stack */
	case I_PRIV_OP:       /* fatal on interrupt stack */
	case I_PRIV_REG:      /* fatal on interrupt stack */
	case I_OVFLO:         /* fatal on interrupt stack */
	case I_COND:          /* handled in locore */
	case I_EXCEP:         /* fatal on interrupt stack */
	case I_TLB_DIRTY:     /* handled in locore */
	case I_EMULAT:        /* fatal on interrupt stack */
	case I_IPGFT:	      /* fatal on interrupt stack */
	case I_IPGFT_NA:      /* fatal on interrupt stack */
	case I_IS_OVFL:       /* fatal on interrupt stack */
	case I_KS_OVFL:       /* fatal on interrupt stack */
	case I_IS_LOST:       /* fatal on interrupt stack */
	case I_LPRIV_XFR:     /* shouldn't happen */
	case I_RECOV_CTR:     /* shouldn't happen */
	default:
		printf("Illegal trap on the interrupt stack\n");
		printf("interrupt(type = %x)\n", type);

		dump_ssp(ssp);

		if (type < TRAP_TYPES)
			panic(trap_type[type]);
		else
			panic("Unknown interrupt type");
		break;
		

	case I_EXT_INTP:
	{
		int ibit;
		extern int ffset();

		/*
		 * keep looping as long as the interrupt level we came in on
		 * allows us to service a pending interrupt
		 */
		if /* XXX while*/ ((ibit = ffset(mfctl(CR_EIRR) & ssp->eiem)) >= 0) {
			/*
			 * Clear this interrupt pending bit in EIRR
			 */
			mtctl(CR_EIRR, 1 << 31 - ibit);
			splx(itab_proc[ibit].intpri);
			(*itab_proc[ibit].handler)(itab_proc[ibit].arg0, ssp);
			/*
			 * We must notify the ASP/LASI/WAX that we have dealt
			 * with this interrupt so it can unmask it again.
			 *
			 * XXX right now, any post-handler routine is recorded
			 * in the otherwise unused arg1 field.
			 */
			if (ibit && itab_proc[ibit].arg1)
				(*((int (*)())itab_proc[ibit].arg1))(ibit);
			/*
			 * To avoid a race condition when reading the EIRR,
			 * we must prevent interrupts again.
			 */
			(void) splhigh();
		}
		break;
	}

	case I_BRK_INST:
		if (break5(ssp->iir) != BREAK_KERNEL) {
			dump_ssp(ssp);
			panic("unknown break instruction in interrupt");
		}

		switch(break13(ssp->iir)) {

		case BREAK_KERNPRINT:
			kern_print(type, ssp);
			break;

#ifdef KGDB
		/* Both the standard GDB breakpoint and KGDB breakpoint
		   are sent to kgdb_trap if they happen in kernel mode.  */
		case BREAK_KGDB:
		case BREAK_GDB:
			kgdb_trap(type, ssp);
			break;
#endif KGDB

#if    MACH_KDB
		case BREAK_MACH_DEBUGGER:
			kdb_trap(ssp, 1);
			break;
#endif MACH_KDB

		default:
			dump_ssp(ssp);
			panic("unknown break instruction in interrupt");
			break;
		}
		break;

	case I_DMEM_BREAK:
	case I_PAGE_REF:
	case I_HPRIV_XFR:
	case I_TAKEN_BR:
#ifdef KGDB
		if (kgdb_enabled) {
			kgdb_trap(type, ssp);
			break;
		}
		/* fall into... */
#endif KGDB
#ifdef DEBUG
		printf("unexpected trap %d, ignored\n", type);
		dump_ssp(ssp);
#endif
		break;
	}
#if 0
	if (ast_needed(cpu_number()) && USERMODE(ssp))
	{
		static int done = 0;
		if (done == 0)
		{
			done = 1;
			printf("!!!about to take ast from interrupt!!!\n");
		}
	}
#endif
}

/*
 * called from syscall if there is an error
 */

int syscall_error(exception, code, subcode, ssp)
	int exception;
	int code;
	int subcode;
	struct parisc_saved_state *ssp;
{
	register thread_t thread;

	thread = current_thread();

	if (thread == 0)
	    panic("syscall error in boot phase");

	if (!USERMODE_PC(ssp->r31))
		panic("system call called from kernel");

	doexception(exception, code, subcode);

	return 0;
}

set_interrupt(level)
{
	/*
	 * XXX - this assumes that all soft interrupts occure through the
	 * monarch processor.
	 */
	PAGE0->mem_hpa->io_eir = level;
}

/*
 * This is the default routine used if interrupts are not configured
 * for a specific eirr bit. Going to this routine means that we got
 * an eirr on a bit that has no IO device.
 */

/*
 * interrupt service routine for power failure
 */
void
powerfail(int_bit, ssp)
	int int_bit;
	struct parisc_saved_state *ssp;
{
        panic("remote power failure");
}

machine_dump_act(mact)
	MachineAct *mact;
{
	printf("\tpcb=%x, uact=%x, ustack=%x, urbuf=%x, urbufsz=%x\n",
	       mact->pcb, mact->user_act, mact->user_stack,
	       mact->user_rbuf, mact->user_rbuf_size);
	printf("\treturn=%x, reentry=%x\n",
	       mact->return_routine, mact->reentry_routine);
	printf("\tspace=%x, pid=%x, sstep=%x\n",
	       mact->space, mact->pid, mact->sstep);
}


dump_ssp(ssp)
	struct parisc_saved_state *ssp;
{
	printf("saved state @ %8.8x:\n", ssp);
	if (ssp == 0)
		return;

	printf("flg %8.8x	r16  %8.8x	sr0   %8.8x	iisqh %8.8x\n",
	       ssp->flags, ssp->r16, ssp->sr0, ssp->iisq_head);
	printf("r1  %8.8x	r17  %8.8x	sr1   %8.8x	iisqt %8.8x\n",
	       ssp->r1, ssp->r17, ssp->sr1, ssp->iisq_tail);
	printf("rp  %8.8x	r18  %8.8x	sr2   %8.8x	iioqh %8.8x\n",
	       ssp->r2, ssp->r18, ssp->sr2, ssp->iioq_head);
	printf("r3  %8.8x	t4   %8.8x	sr3   %8.8x	iioqt %8.8x\n",
	       ssp->r3, ssp->r19, ssp->sr3, ssp->iioq_tail);
	printf("r4  %8.8x	t3   %8.8x	sr4   %8.8x	iir   %8.8x\n",
	       ssp->r4, ssp->r20, ssp->sr4, ssp->iir);
	printf("r5  %8.8x	t2   %8.8x	sr5   %8.8x	isr   %8.8x\n",
	       ssp->r5, ssp->r21, ssp->sr5, ssp->isr);
	printf("r6  %8.8x	t1   %8.8x	sr6   %8.8x	ior   %8.8x\n",
	       ssp->r6, ssp->r22, ssp->sr6, ssp->ior);
	printf("r7  %8.8x	arg3 %8.8x	sr7   %8.8x	ipsw  %8.8x\n",
	       ssp->r7, ssp->r23, ssp->sr7, ssp->ipsw);
	printf("r8  %8.8x	arg2 %8.8x\n",
	       ssp->r8, ssp->r24);
	printf("r9  %8.8x	arg1 %8.8x	cr0   %8.8x	goto  %8.8x\n",
	       ssp->r9, ssp->r25, ssp->cr0, ssp->bsd_goto);
	printf("r10 %8.8x	arg0 %8.8x	pidr1 %8.8x\n",
	       ssp->r10, ssp->r26, ssp->cr8);
	printf("r11 %8.8x	dp   %8.8x	pidr2 %8.8x\n",
	       ssp->r11, ssp->r27, ssp->cr9);
	printf("r12 %8.8x	ret0 %8.8x	sar   %8.8x\n",
	       ssp->r12, ssp->r28, ssp->cr11);
	printf("r13 %8.8x	ret1 %8.8x	pidr3 %8.8x\n",
	       ssp->r13, ssp->r29, ssp->cr12);
	printf("r14 %8.8x	sp   %8.8x	pidr4 %8.8x\n",
	       ssp->r14, ssp->r30, ssp->cr13);
	printf("r15 %8.8x	r31  %8.8x	eiem  %8.8x\n",
	       ssp->r15, ssp->r31, ssp->eiem);
}

dump_fp(fssp)
	struct parisc_fp_state *fssp;
{
	int *fp = (int *)&fssp->fr0;

	printf("fp saved state @ %8.8x:\n", fssp);

	printf("fr0  %8.8x%8.8x		fr16 %8.8x%8.8x\n",
	       fp[0], fp[1], fp[32], fp[33]);
	printf("fr1  %8.8x%8.8x		fr17 %8.8x%8.8x\n",
	       fp[2], fp[3], fp[34], fp[35]);
	printf("fr2  %8.8x%8.8x		fr18 %8.8x%8.8x\n",
	       fp[4], fp[5], fp[36], fp[37]);
	printf("fr3  %8.8x%8.8x		fr19 %8.8x%8.8x\n",
	       fp[6], fp[7], fp[38], fp[39]);
	printf("fr4  %8.8x%8.8x		fr20 %8.8x%8.8x\n",
	       fp[8], fp[9], fp[40], fp[41]);
	printf("fr5  %8.8x%8.8x		fr21 %8.8x%8.8x\n",
	       fp[10], fp[11], fp[42], fp[43]);
	printf("fr6  %8.8x%8.8x		fr22 %8.8x%8.8x\n",
	       fp[12], fp[13], fp[44], fp[45]);
	printf("fr7  %8.8x%8.8x		fr23 %8.8x%8.8x\n",
	       fp[14], fp[15], fp[46], fp[47]);
	printf("fr8  %8.8x%8.8x		fr24 %8.8x%8.8x\n",
	       fp[16], fp[17], fp[48], fp[49]);
	printf("fr9  %8.8x%8.8x		fr25 %8.8x%8.8x\n",
	       fp[18], fp[19], fp[50], fp[51]);
	printf("fr10 %8.8x%8.8x		fr26 %8.8x%8.8x\n",
	       fp[20], fp[21], fp[52], fp[53]);
	printf("fr11 %8.8x%8.8x		fr27 %8.8x%8.8x\n",
	       fp[22], fp[23], fp[54], fp[55]);
	printf("fr12 %8.8x%8.8x		fr28 %8.8x%8.8x\n",
	       fp[24], fp[25], fp[56], fp[57]);
	printf("fr13 %8.8x%8.8x		fr29 %8.8x%8.8x\n",
	       fp[26], fp[27], fp[58], fp[59]);
	printf("fr14 %8.8x%8.8x		fr30 %8.8x%8.8x\n",
	       fp[28], fp[29], fp[60], fp[61]);
	printf("fr15 %8.8x%8.8x		fr31 %8.8x%8.8x\n",
	       fp[30], fp[31], fp[62], fp[63]);
}

#ifdef SYSCALL_PRINT

int syscall_print_enable = 0;
int last_syscall_num;

int syscall_pre_print(int arg0, int arg1, int arg2, int arg3)
{
	/* Keep the syscall routine pointer safe */
	register int r21 __asm("r21");
	int sysrt = r21;

	printf("entering syscall %d (pc %08x), args %08x,%08x,%08x,%08x\n",
		last_syscall_num, sysrt, arg0, arg1, arg2, arg3);


	asm volatile("
		copy %0,%%r26
		copy %1,%%r25
		copy %2,%%r24
		copy %3,%%r23
	" : : "r" (arg0), "r" (arg1), "r" (arg2), "r" (arg3)
	  : "r26", "r25", "r24", "r23");
	return sysrt;
}

int syscall_print(int ret)
{
	printf("finished syscall %d, ret %d\n", last_syscall_num, ret);
	return ret;
}

#endif

#ifdef ACTWATCH
actwatch(ssp)
	struct parisc_saved_state *ssp;
{
	Act *act, *sact;
	int dumped = 0;
	
	for (act = get_next_act(4); act != 0; act = get_next_act(4)) {
		/*
		 * A Lites server activation which is in use (act->lower)
		 * should always have a return_routine set.
		 */
		if (act->lower && act->mact.return_routine == 0) {
			sact = act;
			while (act->higher)
				act = act->higher;
			printf("Screwed server act, chain:\n");
			dump_act(act);
			printf("Server state:\n");
			dump_ssp(&sact->mact.pcb->ss);
			printf("User state:\n");
			dump_ssp(&sact->lower->mact.pcb->ss);
			printf("Current state:\n");
			dump_act(cur_pcb->ss.act);
			dump_ssp(ssp);
			dumped++;
		}
	}
	if (dumped)
		panic("yow!");
}	
#endif
