/* 
 * Copyright (c) 1991, 1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: cs80.h 1.2 91/09/25$
 *	Author: Bob Wheeler, University of Utah CSL
 */

/*
 *	CS80 command macros
 */

#define hpib_initialize(bp)				\
	do {						\
		*(bp)++ = 0x1d;				\
	   } while(0)

#define amigo_identify(bp, device)			\
	do {						\
		*(bp)++ = 0x1c;				\
		*(bp)++ = (device);			\
	   } while(0)


/*
 * macros to build CS80 commands in a buffer...
 */

/*
 * Preamble routines for the 27110B
 * Concurrent CS/80
 */
#define cs80_nodata(bp, device, timeout) 		\
	do {						\
		*(bp)++ = 0x40;				\
		*(bp)++ = (device) & 0x07;		\
		*(bp)++ = (timeout) >> 8 & 0xff;	\
		*(bp)++ = (timeout) & 0xff;		\
	   } while(0)

#define cs80_write(bp, device, timeout) 		\
	do {						\
		*(bp)++ = 0x48;				\
		*(bp)++ = (device) & 0x07;		\
		*(bp)++ = (timeout) >> 8 & 0xff;	\
		*(bp)++ = (timeout) & 0xff;		\
	   } while(0)
	
#define cs80_read(bp, device, timeout) 			\
	do {						\
		*(bp)++ = 0x4c;				\
		*(bp)++ = (device) & 0x07;		\
		*(bp)++ = (timeout) >> 8 & 0xff;	\
		*(bp)++ = (timeout) & 0xff;		\
	   } while(0)

/*
 * Transparent Concurrent CS/80
 */
#define tcs80_nodata(bp, device, timeout) 		\
	do {						\
		*(bp)++ = 0x60;				\
		*(bp)++ = (device) & 0x07;		\
		*(bp)++ = (timeout) >> 8 & 0xff;	\
		*(bp)++ = (timeout) & 0xff;		\
	   } while(0)

#define tcs80_write(bp, device, timeout) 		\
	do {						\
		*(bp)++ = 0x68;				\
		*(bp)++ = (device) & 0x07;		\
		*(bp)++ = (timeout) >> 8 & 0xff;	\
		*(bp)++ = (timeout) & 0xff;		\
	   } while(0)
	
#define tcs80_read(bp, device, timeout) 		\
	do {						\
		*(bp)++ = 0x6c;				\
		*(bp)++ = (device) & 0x07;		\
		*(bp)++ = (timeout) >> 8 & 0xff;	\
		*(bp)++ = (timeout) & 0xff;		\
	   } while(0)

/*
 * non-transparent commands
 */
#define locate_and_read(bp)				\
	do {						\
		*(bp)++ = 0;				\
	   } while(0)

#define cold_load_read(bp)				\
	do {						\
		*(bp)++ = 0x0a;				\
	   } while(0)

#define locate_and_write(bp)				\
	do {						\
		*(bp)++ = 0x02;				\
	   } while(0)

#define write_file_mark(bp)				\
	do {						\
		*(bp)++ = 0x49;				\
	   } while(0)

#define set_unit(bp, unit)				\
	do {						\
		*(bp)++ = 0x20 | (unit) & 0x0f;		\
	   } while(0)

#define set_volume(bp, volume)				\
	do {						\
		*(bp)++ = 0x40 | (volume) & 0x07;	\
	   } while(0)

#define set_address(bp, high, low)			\
	do {						\
		*(bp)++ = 0x10;				\
		*(bp)++ = (high) >> 8 & 0xff;		\
		*(bp)++ = (high) & 0xff;		\
		*(bp)++ = (low) >> 24 & 0xff;		\
		*(bp)++ = (low) >> 16 & 0xff;		\
		*(bp)++ = (low) >> 8 & 0xff;		\
		*(bp)++ = (low) & 0xff;			\
	   } while(0)

#define set_block_disp(bp, high, low)			\
	do {						\
		*(bp)++ = 0x12;				\
		*(bp)++ = (high) >> 8 & 0xff;		\
		*(bp)++ = (high) & 0xff;		\
		*(bp)++ = (low) >> 24 & 0xff;		\
		*(bp)++ = (low) >> 16 & 0xff;		\
		*(bp)++ = (low) >> 8 & 0xff;		\
		*(bp)++ = (low) & 0xff;			\
	   } while(0)

#define set_length(bp, count)				\
	do {						\
		*(bp)++ = 0x18;				\
		*(bp)++ = (count) >> 24 & 0xff;		\
		*(bp)++ = (count) >> 16 & 0xff;		\
		*(bp)++ = (count) >> 8 & 0xff;		\
		*(bp)++ = (count) & 0xff;		\
	   } while(0)

#define set_burst(bp, tag, count)			\
	do {						\
		*(bp)++ = 0x3c | (tag) & 1;		\
		*(bp)++ = (count) & 0xff;		\
	   } while(0)

#define set_rps(bp, time1, time2)			\
	do {						\
		*(bp)++ = 0x39;				\
		*(bp)++ = (time1) & 0xff;		\
		*(bp)++ = (time2) & 0xff;		\
	   } while(0)

#define set_retry_time(bp, time)			\
	do {						\
		*(bp)++ = 0x3a;				\
		*(bp)++ = (time) >> 8 & 0xff;		\
		*(bp)++ = (time) & 0xff;		\
	   } while(0)

#define set_status_mask(bp, p1, p2, p3, p4, p5, p6, p7, p8) \
	do {						\
		*(bp)++ = 0x3e;				\
		*(bp)++ = (p1);				\
		*(bp)++ = (p2);				\
		*(bp)++ = (p3);				\
		*(bp)++ = (p4);				\
		*(bp)++ = (p5);				\
		*(bp)++ = (p6);				\
		*(bp)++ = (p7);				\
		*(bp)++ = (p8);				\
	   } while(0)

#define no_op(bp)					\
	do {						\
		*(bp)++ = 0x34;				\
	   } while(0)

#define set_release(bp, suppress, automatic)		\
	do {						\
		*(bp)++ = 0x3b;				\
		*(bp)++ = (suppress) << 7 & 0x80 | 	\
			  (automatic) << 6 & 0x40;	\
	   } while(0)

#define set_options(bp, auto_sparing, jump_sparing, char_count, imm_report) \
	do {						\
		*(bp)++ = 0x38;				\
		*(bp)++ = (imm_report) << 3 & 0x08 |    \
	        	  (auto_sparing) << 2 & 0x04 |  \
		          (jump_sparing) << 1 & 0x02 |  \
		          (imm_report) & 0x01;		\
	   } while(0)

#define set_ret_addr_mode(bp, mode)			\
	do {						\
		*(bp)++ = 0x48;				\
		*(bp)++ = (mode) & 0x07;		\
	   } while(0)

#define describe(bp)					\
	do {						\
		*(bp)++ = 0x35;				\
	   } while(0)

#define initialize_media(bp, mode, interleave)		\
	do {						\
		*(bp)++ = 0x37;				\
		*(bp)++ = (mode) & 0x07;		\
		*(bp)++ = (interleave);			\
	   } while(0)

#define spare_block(bp, skip_jump, retain_data)		\
	do {						\
		*(bp)++ = 0x06;				\
		*(bp)++ = (skip_jump) << 2 & 0x04 |	\
			  (retain_data) & 0x01;		\
	   } while(0)

#define locate_and_verify(bp)				\
	do {						\
		*(bp)++ = 0x04;				\
	   } while(0)

#define copy_data(bp, src_unit, src_vol, src_high, src_low, dst_unit, dst_vol, dst_high, dst_low) 	\
	do {						\
		*(bp)++ = 0x08;				\
		*(bp)++ = (src_vol) << 4 & 0x70 | 	\
			  (src_unit) & 0x07;		\
		set_address(bp, src_high, src_low);	\
		*(bp)++ = (dst_vol) << 4 & 0x70 | 	\
			  (dst_unit) & 0x07;		\
		set_address(bp, dst_high, dst_low);	\
	   } while(0)

#define release(bp)					\
	do {						\
		*(bp)++ = 0x0e;				\
	   } while(0)

#define release_denied(bp)				\
	do {						\
		*(bp)++ = 0x0f;				\
	   } while(0)

#define unload(bp)					\
	do {						\
		*(bp)++ = 0x4a;				\
	   } while(0)

#define initiate_utility(bp, mode, utility)		\
	do {						\
		*(bp)++ = 0x30 | (mode) & 0x03;		\
		*(bp)++ = (utility)			\
	   } while(0)

#define initiate_diag(bp, loop, diag)			\
	do {						\
		*(bp)++ = 0x33;				\
		*(bp)++ = (loop) >> 8 & 0xff;		\
		*(bp)++ = (loop) & 0xff;		\
		*(bp)++ = (diag);			\
	   } while(0)

#define request_status(bp)				\
	do {						\
		*(bp)++ = 0x0d;				\
	   } while(0)

/*
 * transparent commands (odd parity)
 */
#define univ_dev_clear(bp)				\
	do {						\
		*(bp)++ = 0x94;				\
	   } while(0)

#define selected_dev_clear(bp, addrs)			\
	do {						\
		extern char odd_parity[];		\
		*(bp)++ = odd_parity[(addrs) & 0x1f | 0x20]; \
		*(bp)++ = 0x84;				\
	   } while(0)

#define chan_indep_clear(bp, unit)			\
	do {						\
		*(bp)++ = 0x20 | (unit) & 0x0f;		\
		*(bp)++ = 0x08;				\
	   } while(0)

#define cancel(bp, unit)				\
	do {						\
		*(bp)++ = 0x20 | (unit) & 0x0f;		\
	   } while(0)

#define loopback(bp, write)				\
	do {						\
		*(bp)++ = 0x02 | (write) & 1;		\
	   } while(0)

#define set_parity(bp, srq, parity)			\
	do {						\
		*(bp)++ = 0x01;				\
		*(bp)++ = (srq) << 1 & 0x02 | 		\
			  (parity) & 1;			\
	   } while(0)

#define identify(bp, addrs)				\
	do {						\
		extern char odd_parity[];		\
		*(bp)++ = (addrs) & 0x1f | 0x60; 	\
	   } while(0)


