/* 
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: hpibreg.h 1.4 92/06/27$
 *	Author: Bob Wheeler, University of Utah CSL
 */

/*
 *	HPIB and CS80 structures
 */

struct hpib_request {
	struct hpib_request *next;	/* pointer to next request */
	struct hpib_request *prev;	/* pointer to previous request */
	int	read;			/* direction of data transfer */
	caddr_t	command;		/* CS/80 command string */
	int	cmdcnt;			/* size of cs80 command string */
	caddr_t	buffer;			/* data block address */
	int	bufcnt;			/* size of data block */
	caddr_t qstatus;		/* address of status block */
	int	qstatcnt;		/* size of status block */
	int	unit;			/* unit that started this request */
	int	hpibunit;		/* HPIB that started this request */
	void	(*done)();		/* routine to call when done */
	int	status;			/* final status of request */
};

/*
 * per device structure
 */
struct hpib_dev {
	int	slave;			/* slave number on HPIB */
	int	hpibunit;		/* HPIB unit */
	int	hwid;			/* Hardware ID */
	int	alive;			/* device configured? */
	struct bus *device;		/* device selected by autoconf */
};

/*
 * alignment requirements for buffers
 */
#define	HPIB_BUFFER_ALIGN	64
#define	HPIB_MIN_BUFFER		64

/*
 * any tstat less than 9 is not considered an error
 */
#define TSTAT_ERROR	9

/*
 * maximum number of slaves on HPIB, This is a limitation of PPOLL, the IEEE
 * standard allows 32 slaves
 */
#define MAX_HPIB_SLAVES	8

struct cs80_status {
	char  	tstat;
	char  	qstat;
	char  	volunit;		/* volume and unit */
	char  	pending;		/* lowest unit with status pending */
	short 	reject;			/* reject errors */
	short 	fault;			/* fault errors */
	short 	access;			/* access errors */
	short 	info;			/* information errors */
	u_char  params[10];		/* parameters */
	char  	statqstat;		/* qstat of request status */
};
