/* 
 * Copyright (c) 1991,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: smb.c 1.2 91/09/25$
 *	Author: Bob Wheeler, University of Utah CSL, 8/91
 */

/*
 *	Device driver for system memory bus
 */

#include <smb.h>
#if NSMB > 0

#include <sys/types.h>
#include <device/errno.h>
#include <hpcio/device.h>
#include <machine/iodc.h>
#include <machine/iomod.h>
#include <machine/pdc.h>

extern struct modtab *smb_bus;		/* from busconf */

int	smbattach(), smbprobe(), smbscan(), smblayer();
void	smbdisplay();
struct	driver smbdriver = {
	"smb", (int (*)()) 0, (int (*)()) 0, smbscan, smbdisplay, smblayer
};

int
smbscan()
{
	struct modtab *m;
	extern struct bus bus_devices[];
	struct bus *device;

	for (device = bus_devices; device->driver != (struct driver *)0; 
	     device++) {
		/*
		 * if this device isn't on the SMB then skip it
		 */
		if (device->bustype != &smbdriver)
			continue;

		/*
		 * look in the configuration table and try to identify 
		 * the other modules
		 */
		for (m = smb_bus; m != (struct modtab *)0; m = m->m_next) {
			/*
			 * see if this module has already been configured 
			 * or is in the wrong slot
			 */
			if (m->m_alive || (device->busslot != m->m_fixed && 
					   device->busslot != -1))
				continue;

			/*
			 * see if we recognize this module as what we want...
			 */
			if ((*device->driver->probe)(m) != 0)
				continue;
				
			/*
			 * this looks like it... attach it and associate it 
			 * with the right unit
			 */
			if ((*device->driver->attach)(m, device->unit) != 0)
				continue;

			/*
			 * Now scan it for devices or controllers below it
			 */
			if ((*device->driver->scan)(device->unit) != 0)
				continue;

			/*
			 * found it and initialized it
			 */
			m->m_device = device;
			m->m_alive = 1;
			break;
		}
	}

	return 0;
}

void
smbdisplay()
{
	struct modtab *m;
	extern int indent;

	indent = 5;
	for (m = smb_bus; m != (struct modtab *)0; m = m->m_next) {
		if (m->m_type.iodc_type == IODC_TP_NPROC) {
			iprintf("Procesor in smb slot %d\n",
				m->m_fixed, m->m_hpa);
			continue;
		}

		if (m->m_type.iodc_type == IODC_TP_MEMORY) {
			iprintf("%d Mbytes of memory in smb slot %d\n",
			m->m_memsize / 0x100000, m->m_fixed);
			continue;
		}

		if (!m->m_alive) {
			iprintf("unknown module (type:%x svers:%x hvers:%x) in smb slot %d\n",
				m->m_type.iodc_type, m->m_type.iodc_sv_model, 
				m->m_type.iodc_model, m->m_fixed);
		} else {
			iprintf("%s%d in smb slot %d\n", 
				m->m_device->driver->name,
				m->m_device->unit, m->m_fixed);
			indent += 5;
			(*m->m_device->driver->display)
				(m->m_device->unit);
			indent -= 5;
		}
	}
	indent -= 5;
}

/*
 * XXX - this needs work if it is to run on a machine with a bus convertor.
 * We're assuming that the boot controller (the CIO) is on the SMB bus.
 * We should decend down busconvertors and find the HPA on any bus.
 */
smblayer(hpa, layer, devname)
	struct iomod *hpa;
	int *layer;
	char *devname;
{
	struct modtab *m;

	for (m = smb_bus; m != (struct modtab *)0; m = m->m_next) {
		if (m->m_hpa != hpa)
			continue;

		if (m->m_type.iodc_type == IODC_TP_NPROC) {
			*devname++ = 'c';
			*devname++ = 'p';
			*devname++ = 'u';
			*devname++ = '\0';
			return 0;
		}

		if (m->m_type.iodc_type == IODC_TP_MEMORY) {
			*devname++ = 'm';
			*devname++ = 'e';
			*devname++ = 'm';
			*devname++ = '\0';
			return 0;
		}

		if (!m->m_alive)
			return ENXIO;

		return (*m->m_device->driver->layer)
			(m->m_device->unit, layer, devname);
	}

	return ENXIO;
}
#endif	NSMB




