/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Common ATbus control declarations
 */

/*
 * DMA controllers
 */
#define ATBUS_NCHANS	8

union dma0 {
	volatile struct {
		volatile struct {
			volatile unsigned address:8;
			volatile unsigned count:8;
		} channel[4];
		volatile unsigned status:8;
		volatile unsigned :8;
		volatile unsigned :8;
		volatile unsigned :8;
		volatile unsigned :8;
		volatile unsigned temporary:8;
		volatile unsigned :8;
		volatile unsigned :8;
	} read;
	volatile struct {
		volatile struct {
			volatile unsigned address:8;
			volatile unsigned count:8;
		} channel[4];
		volatile unsigned command:8;
		volatile unsigned request:8;
		volatile unsigned mask_channel:8;
		volatile unsigned mode:8;
		volatile unsigned clear_byte_pointer:8;
		volatile unsigned master_clear:8;
		volatile unsigned mask_clear:8;
		volatile unsigned mask_write:8;
	} write;
	volatile unsigned char	all[0x20];
};

union dma1 {
	volatile struct {
		volatile struct {
			volatile unsigned address:8;
			volatile unsigned :8;	/* pad */
			volatile unsigned count:8;
			volatile unsigned :8;	/* pad */
		} channel[4];
		volatile unsigned status:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned :8;
		volatile unsigned :8;	/* pad */
		volatile unsigned :8;
		volatile unsigned :8;	/* pad */
		volatile unsigned :8;
		volatile unsigned :8;	/* pad */
		volatile unsigned :8;
		volatile unsigned :8;	/* pad */
		volatile unsigned temporary:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned :8;
		volatile unsigned :8;	/* pad */
		volatile unsigned :8;
		volatile unsigned :8;	/* pad */
	} read;
	volatile struct {
		volatile struct {
			volatile unsigned address:8;
			volatile unsigned :8;	/* pad */
			volatile unsigned count:8;
			volatile unsigned :8;	/* pad */
		} channel[4];
		volatile unsigned command:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned request:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned mask_channel:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned mode:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned clear_byte_pointer:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned master_clear:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned mask_clear:8;
		volatile unsigned :8;	/* pad */
		volatile unsigned mask_write:8;
		volatile unsigned :8;	/* pad */
	} write;
	volatile unsigned char	all[0x20];
};

#define MODE(chan)		((chan)&0x3)
#define MODE_VERIFY		0x00
#define MODE_WRITE		0x04
#define MODE_READ		0x08
#define MODE_AUTOINIT		0x10
#define MODE_DECREMENT		0x20
#define MODE_DEMAND		0x00
#define MODE_SINGLE		0x40
#define MODE_BLOCK		0x80
#define MODE_CASCADE		0xc0

/*
 * Programmable interrupt controller (PIC)
 */

#define ATBUS_PIC_NINTS		8	/* number of interrupt levels per pic */

/* PIC commands */
#define ATBUS_PIC_EOI		0x20	/* write to pic r0 to eoi interrupt */
#define ATBUS_PIC_READ_IREQ	0x0a	/* write to pic r0 to prepare to read irr */
#define ATBUS_PIC_READ_INSERV	0x0b	/* write to pic r0 to read in-service register */

union pic {
        volatile struct {
                volatile unsigned char  irr;            /* interrupt request register */
                volatile unsigned char  imr;            /* interrupt mask register */
        } p;
        volatile struct {
                volatile unsigned char  r0;
                volatile unsigned char  r1;
        } r;
};

/*
 * Timers
 */
struct timer {
	volatile unsigned char sysclk;
	volatile unsigned char refresh;
	volatile unsigned char speaker;
	volatile unsigned char control;
	volatile unsigned char pad[4];
};

/*
 * Page registers
 */
union page_regs {
        volatile struct {
                volatile unsigned char  pad0;
                volatile unsigned char  ch2;
                volatile unsigned char  ch3;
                volatile unsigned char  ch1;
                volatile unsigned char  pad1;
                volatile unsigned char  pad2;
                volatile unsigned char  pad3;
                volatile unsigned char  ch0;
                volatile unsigned char  pad4;
                volatile unsigned char  ch6;
                volatile unsigned char  ch7;
                volatile unsigned char  ch5;
        } r;
        volatile unsigned char all[0x20];
};

/*
 * ATbus control page
 */
struct at_control {
        union dma0      dma0;           /* 0x000: dma channels 0-3 */
        union pic       pic0;           /* 0x020: pic, irq 0-7 */
        volatile unsigned char iack;    /* 0x022: trailways - 2 back to back reads generate atbus iack cycle and return int level */
        unsigned char   pad0[0x3f-0x23+1];
        struct timer    timers[2];      /* 0x040: timers */
        unsigned char   pad1[0x5f-0x50+1];
        volatile unsigned short nmi;    /* 0x060: non-maskable interrupt control */
        unsigned char   pad2[0x7f-0x62+1];
        union page_regs pr;             /* 0x080: page registers */
        union pic       pic1;           /* 0x0a0: pic, irq 8-15 */
        unsigned char   pad3[0xbf-0xa2+1];
        union dma1      dma1;           /* 0x0c0: dma channels 4-7 */
        volatile unsigned char master_req;     /* 0x0e0: master request register */
        unsigned char   pad4[0xff-0xe1+1];
};                                      /* 0x100 */
