/* 
 * Copyright (c) 1991, 1992, 1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: bsd_glue.h 1.14 94/12/14$
 */
/*
 * Glue to make OSF-1 or Mach 3.0 drivers work in BSD
 * with minimal damage to driver code.
 */

#define DBG	1	/* XXX */

#define _NO_PROTO	/* XXX for gcc -traditional that we use */

#undef	FALSE
#undef	TRUE
#define FALSE	0
#define TRUE	1

typedef	u_long		vm_offset_t;
typedef	u_long		vm_size_t;
typedef int		boolean_t;

#define KERN_SUCCESS	0

#define	kmem_alloc(m,bp,s)	\
({ \
	*(bp) = (vm_offset_t)malloc(s, M_DEVBUF, M_NOWAIT); \
	*(bp) ? KERN_SUCCESS : 1; \
})
#define kmem_free(m,a,s)	free(a, M_DEVBUF)
#define trunc_page(v)		((vm_offset_t)(v) &~ PGOFSET)
#define round_page(v)		(((vm_offset_t)(v)+NBPG-1) &~ PGOFSET)

/*
 * HP OSF1 stuff
 * - map/pmap pointer is actually a buffer pointer or NULL for kernel space
 */
typedef	struct buf	*vm_map_t;
#define	kernel_map	((vm_map_t)0)
#define	vm_map_pmap(m)	(m)
#define pmaptosid(bp, v) \
	(((bp) && ((bp)->b_flags & (B_PAGET|B_PHYS)) == B_PHYS) \
		? bvtospace(bp, v) : KERNSPC)
#define pmap_extract(p, v)		\
	(vm_offset_t) ltor(pmaptosid(p, v), v)
#define pmap_flush_range(p, v, s)	\
{ \
	space_t sid = pmaptosid(p, v); \
	if ((((int)(v) & 31) + (s)) <= 32) \
		fcacheline(sid, (caddr_t)(v)); \
	else if (sid == KERNSPC) \
		fdcache(sid, v, s); \
	else if (ltovalidr(sid, v)) { \
		fdcache(sid, v, s); \
		ficache(sid, v, s); \
	} \
}
#define pmap_purge_range(p, v, s)	\
{ \
	space_t sid = pmaptosid(p, v); \
	if (sid == KERNSPC) \
		pdcache(sid, v, s); \
	else if (ltovalidr(sid, v)) { \
		pdcache(sid, v, s); \
		ficache(sid, v, s); \
	} \
}

#ifdef USELEDS
/*
 * LED control
 */
#define PANEL_LED_DISK		0x04
#define PANEL_LED_TOGGLE	2
#define panel_led(led, how) \
	if (how == PANEL_LED_TOGGLE) { \
		extern int inledcontrol; \
		if (inledcontrol == 0) ledcontrol(0, 0, led); \
	}
#else
#define	panel_led(led, how)
#endif

/*
 * For EISA
 */
#define RESOLVER_ADD_HANDLER 1
#define RESOLVER_DEL_HANDLER 2

#define	IO_EISA_IACK_PHYS	0xfc01f000
#define	IO_EISA_IACK_LEN	0x1000

#define io_map(pa, sz, nm)	(pa)
#define io_unmap(va, nm)

/*
 * Missing disk stuff
 */
#define dkunit(d)	(minor(d)>>3)
#define dkpart(d)	(minor(d)&0x07)
#define dkminor(u,p)	(((u)<<3)|((p)&0x07))
#define b_cylin		b_resid

/*
 * Skeletal structs for attach routines
 */
struct hp_businfo {
	int	bustype;
};

#define	MAXCSRS	1
typedef struct hp_ctlrinfo {
	int		   ctlrnum;
	struct hp_ctlrinfo *parent_ctlrinfop;
	struct hp_businfo  *ctlr_businfop;
	int                irq;
	int		   numcsrs;
	vm_offset_t        csrs[MAXCSRS];
	vm_offset_t	   vcsrs[MAXCSRS];
} hp_ctlrinfo_t;

#define BUSTYPE_NOSUCHBUS	0
#define BUSTYPE_SYS		1
#define BUSTYPE_EISA		2
