/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/* 
 * $RCSfile: ec596reg.h,v $
 *
 * $Revision: 1.2 $
 * $Date: 1994/07/21 18:17:39 $
 */

#ifndef ec596reg_h
#define ec596reg_h

#define EC596_NIL_PTR		((void*)(-1))

typedef unsigned char dcache_line_t[DCACHE_LINE_SIZE];

/*
 * BIG FAT NOTICE:
 *
 * Due to PA-RISC's Cache management strategy, several main-memory
 * Command Blocks managed by this driver must be very carefully laid
 * out in order to avoid overwriting Status bits when issuing commands
 * to the chip.  The problem is that a Command Block may have command
 * bits (written by the CPU) and Status bits (written by the chip) in
 * adjacent memory locations.  Since a "write" to main memory updates
 * an entire Data Cache line, we must insure that these bits occupy
 * different lines.  The intended Cache line boundaries are indicated
 * by comments in the data-structure declarations below.
 */
#define ALIGNED(x,y)((int)(x)+(y)&~((y)-1))
#define CACHE_ALIGNED(x)ALIGNED((x),DCACHE_LINE_SIZE)

/*
 * Note that we must place some 32-bit words on halfword boundaries,
 * due to the data cache line requirements discussed above.  Since the
 * compiler aligns data naturally, we must use the following data type
 * in order to write a word on a halfword boundary, using 2 halfword
 * "write"s.  This is OK since, in most cases, the halfwords must be
 * swapped anyway in order to satisfy the 82596.
 */
typedef struct {
    unsigned short low;
    unsigned short high;
} swap_t;

/*
 * Transmit Buffer Descriptor:
 */
typedef struct tbd {
    unsigned short eof	:1,  /* <=> Last buffer/End of file */
			:1,  /* (Unused; must be 0) */
		   count:14, /* # bytes in buffer */ 
			:16; /* (Unused; must be 0) */
    swap_t link,	     /* ->Next buffer (halfword-swapped) */
	   buffer,	     /* ->Current buffer (halfword-swapped) */
	   pa;		     /* Phys. addr. of this TBD (halfword-swapped) */

    /*
     * This is a hack so that "Transmit" buffers are always aligned on
     * a even (halfword) storage boundary.
     */
    char oddchar[2];
#ifdef DEBUG
    short dummy;	     /* (Unused) */
    struct tbd *next;	     /* Virtual address of next TBD */
#endif
} tbd_t;

/*
 * Command Block command codes:
 */
typedef unsigned short cb_cuc_t;
#define cb_cuc_nop			0	/* nop */
#define cb_cuc_ia_setup			1	/* setup individual address */
#define cb_cuc_configure		2	/* configure */
#define cb_cuc_multicast_setup		3	/* setup multicast */
#define cb_cuc_transmit			4	/* transmit */
#define cb_cuc_time_domain_reflectometer 5	/* line analysis */
#define cb_cuc_dump			6	/* dump internal status */
#define cb_cuc_diagnose			7	/* diagnosis */

/*
 * Command Block header:
 */
typedef union cb_command {
    struct {
	unsigned char pad[DCACHE_LINE_SIZE-2]; /* Cache-line alignment pad */
	unsigned short completion:1,  /* <=> Command complete? */
		       busy	 :1,  /* <=> Chip busy? */
		       ok	 :1,  /* <=> Completion successful? */
		       abnormal	 :1,  /* <=> 'scb_cuc_abort' command stop? */
		       fail	 :1,  /* <=> Self-test failed? */
				 :11; /* (Unused; must be 0) */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	unsigned short el	 :1,  /* End of list */
		       suspend	 :1,  /* Suspend after completion */
		       interrupt :1,  /* Interrupt after completion */
				 :8,  /* (Unused; must be 0) */
		       no_crc	 :1,  /* Disable CRC insertion */
		       sf	 :1;  /* 0/1 <=> Simplified/Flexible mode */
		     cb_cuc_t cuc:3;  /* Command Block control unit command */
	swap_t link;	/* ->Next header (halfword-swapped) */
    } cmd;
    struct {
	unsigned char pad[DCACHE_LINE_SIZE-2]; /* Cache-line alignment pad */
	unsigned short			  :4, /* (Unused; must be 0) */
		       late_collision	  :1,
		       no_carrier_sense	  :1,
		       lost_clear_to_send :1,
		       dma_underrun	  :1,
		       deferred		  :1,
		       heartbeat	  :1,
		       too_many_collisions:1,
					  :1, /* (Unused; must be 0) */
		       maxcoll		  :4;
    } xmit;
    struct {
	unsigned char pad[DCACHE_LINE_SIZE-2]; /* Cache-line alignment pad */
	unsigned short skip		       :3,
		       length_error	       :1,
		       crc_error	       :1,
		       alignment_error	       :1,
		       no_resources	       :1,
		       dma_overrun	       :1,
		       individual_address_match:1,
		       no_eop_flag	       :1,
		       zeros		       :5,
		       collision_detected      :1;
    } recv;
    struct {
	swap_t link; /* ->Next CB (halfword-swapped) */
    } save;
    struct {
	unsigned char pad[DCACHE_LINE_SIZE-2]; /* Pad to reach Status bits */
	unsigned short cb_stat; /* Bits set by chip */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	unsigned short cb_cmd; /* Bits set by CPU */
    } cacheline;
} cb_command_t;

/*
 * "No operation" Command Block:
 */
typedef union {
    struct {
	cb_command_t command;
    } nop;
    struct {
	dcache_line_t stat; /* 1st Data cache line w/bits set by chip */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	dcache_line_t cmd; /* 2nd Data cache line w/bits set by CPU */
    } cacheline;
} cb_cuc_nop_t;

typedef cb_cuc_nop_t	nop_t;		/* (A convenient alias) */
typedef unsigned char	eaddr_t[6];	/* Ethernet address */

/*
 * "Individual Address Setup" Command Block:
 */
typedef struct {
    cb_command_t	command;
    eaddr_t		address;	
    unsigned short	pad;
} cb_cuc_ia_setup_t;

/*
 * "Configure" Command Block:
 */
typedef struct {
    cb_command_t command;
    union {
	struct {
	    /* begins at byte 8 */
	    unsigned char prefetched	   :1, /* Wrt pref. bit in pref RBD */
					   :3, /* (Unused; must be 0) */
			  byte_cnt	   :4, /* Parameter byte count */
    
	 /* Byte 9 */	  monitor0	   :2, /* Receive monitor options */
					   :2,
			  fifo_limit	   :4, /* FIFO limit */
	 /* Byte 10 */	  sav_bf	   :1, /* Save bad frames in memory */
					   :7,
	 /* Byte 11 */	  loopback_mode	   :2, /* Loopback mode */
			  preamble_length  :2, /* Preamble length */
			  no_src_addr_ins  :1, /* No source addr. insertion */
			  addr_len	   :3, /* Address length */
	 /* Byte 12 */	  bof_metd	   :1, /* Exponential backoff method */
			  exp_prio	   :3, /* Exponential priority */
					   :1,
			  lin_prio	   :3, /* Linear priority */
	 /* Byte 13 */	  interframe_spacing,  /* Interframe spacing */
	 /* Byte 14 */	  slot_time_lo,	       /* Slot time, low byte */
	 /* Byte 15 */	  max_retry_num	   :4, /* Max. no. of retries */
					   :1,
			  slot_time_hi	   :3, /* Slot time, high bits */
	 /* Byte 16 */	  padding	   :1, /* Padding requested */
			  bit_stuffing	   :1, /* Bit-stuffing */
			  crc_16	   :1, /* 0/1 <=> 32/16-bit CRC */
			  no_crc_insertion :1, /* No CRC insertion */
			  xmit_on_no_crs   :1, /* Xmit on no carrier sense */
			  manchester_nrz   :1, /* Manchester or NRZ encoding */
			  broadcast_disable:1, /* Broadcast disable */
			  promiscious_mode :1, /* Promiscuous mode */
	 /* Byte 17 */	  cdt_src	   :1, /* Collision detect source */
			  cdt_filter	   :3, /* Coll. detect filter len */
			  crs_src	   :1, /* Carrier sense source */
			  crs_filter	   :3, /* Carr. sense filter length */
	 /* Byte 18 */	  min_frame_len,       /* Min. frame length */
	 /* Byte 19 */	  monitor1	   :2, /* Receive monitor options */
			  disable_mc_all   :1, /* Receive all MC frames */
			  coll_detect_bsac :1, /* Col. det. by source addr. */
			  auto_retransmit  :1, /* Auto-retransmit */
		     disable_crc_in_memory :1, /* App. rcv. CRC to mem frame */
			  length_field	   :1, /* Enable length field */
			  precrs	   :1, /* Preamble'til carrier sense */
	 /* Byte 20 */	  dcr		   :1, /* Deterministic col. res. */
			  fdx		   :1, /* Full-duplex resolution */
			  dcr_slot_address :6, /* Station index in DCR mode */
	 /* Byte 21 */	  disable_backoff  :1, /* disable back-off algorithm */
			  multiple_ia	   :1, /* Multiple individual addr. */
			  dcr_no_of_stations:6;/* # stations in DCR mode */
	    unsigned short clear;	       /* bytes 22-23 must be 0 */
	} field;
	unsigned char bytes[16];
    } cfg;
} cb_cuc_configure_t;

/*
 * "Multi-cast Setup" Command Block:
 */
typedef struct {
    cb_command_t command;
    unsigned short    :2,   /* (Unused; must be 0) */
		   cnt:14;  /* # of bytes in multicast list */
    eaddr_t mc[NMULTICAST]; /* Multicast list */
} cb_cuc_multicast_setup_t;

/*
 * "Transmit" Command Block:
 */
typedef union tcb {
    struct {
	cb_command_t command;
	swap_t tbd_ptr; /* ->Transmit buffer (halfword-swapped) */
	unsigned short eof	:1,  /* <=> Whole frame is in TCB? */
				:1,  /* (Unused; must be 0) */
		       tcb_count:14; /* byte count to transmit */
	unsigned short zeros;	     /* (Unused; must be 0) */
	eaddr_t	da;		     /* (Unused) Destination address */
	unsigned short length;	     /* (Unused) User-def len in IEEE 802.3 */
	unsigned short pad;	     /* (Unused) */
	/*
	 * The following are optional (user-defined) bytes that are
	 * ignored by the controller.
	 */
	io_req_t ior; /* ->I/O request we are transmitting */
	union tcb *prev;	     /* Virtual address of next TCB */
	union tcb *next;	     /* Virtual address of previous TCB */
    } tcb;
    struct {
	dcache_line_t stat; /* 1st Data cache line w/bits set by chip */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	dcache_line_t cmd; /* 2nd Data cache line w/bits set by CPU * */
    } cacheline;
} cb_cuc_transmit_t;

typedef cb_cuc_transmit_t tcb_t;	/* (A convenient abbreviation) */

/*
 * "Time-Domain Reflectometer" Command Block:
 */
typedef struct {
    cb_command_t command;
    unsigned short link_ok	    :1,
		   transceiver_cable:1,
		   ethernet_open    :1,
		   ethernet_short   :1,
				    :1,
		   time		    :11;
    unsigned short clear;
} cb_cuc_time_domain_reflectometer_t;

/*
 * "Dump" Command Block:
 */
typedef struct {
    cb_command_t command;
    /* You need it?  You code it! */
} cb_cuc_dump_t;

/*
 * "Diagnose" Command Block:
 */
typedef struct {
    cb_command_t command;
    /* You need it?  You code it! */
} cb_cuc_diagnose_t;

/*
 * (Composite) Command Block:
 */
typedef union {
    cb_cuc_nop_t		       nop;
    cb_cuc_ia_setup_t		       ia_setup;
    cb_cuc_configure_t		       configure;
    cb_cuc_multicast_setup_t	       multicast_setup;
    cb_cuc_transmit_t		       transmit;
    cb_cuc_time_domain_reflectometer_t time_domain_reflectometer;
    cb_cuc_dump_t		       dump;
    cb_cuc_diagnose_t		       diagnose;
} cb_t;

/*
 * Receive Buffer Descriptor:
 */
typedef union rbd {
    struct {
	unsigned char pad[DCACHE_LINE_SIZE-2];
	unsigned short eof	 :1,  /* End of file */
		       filled	 :1,  /* Filled by adapter */
		       count	 :14; /* # of bytes in buffer */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	unsigned short zeros1;	      /* (Unused; must be 0) */
	swap_t link;   /* ->Next RBD (halfword-swapped) */
	swap_t buffer; /* ->Buffer (halfword-swapped) */
	unsigned short el	 :1,  /* End of FB list */
		       prefetched:1,  /* Data in this RBD already prefetched */
		       size	 :14; /* byte count (should be EVEN!) */
	unsigned short zeros2;	      /* (Unused; must be 0) */
	char dummy[6];	 /* (Unused) */
	swap_t pa;	 /* (Debugging) Phy addr of this RBD (h.w.-swapped) */
	ipc_kmsg_t kmsg; /* ->IPC message buffer we're receiving into */
	union rbd *next; /* Virtual address of next RBD */
    } rbd;
    struct {
	unsigned char  pad1[DCACHE_LINE_SIZE-2];
	unsigned short stat; /* Flags and count */
    } cmd;
    struct {
	swap_t link; /* ->Next RBD (halfword-swapped) */
    } save;
    struct {
	dcache_line_t stat;	/* 1st data cache line w/bits set by chip */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	dcache_line_t cmd;	/* 2nd data cache line w/bits set by CPU */
    } cacheline;
} rbd_t;

/*
 * Receive Frame Descriptor:
 */
typedef union rfd {
    struct {
	cb_command_t command;
	swap_t rbd_ptr;	/* ->RBD (halfword-swapped) */
	unsigned short eof   :1,  /* End of file */
		       filled:1,  /* Filled by adapter */
		       count :14; /* byte count */
	unsigned short	     :2,  /* (Unused; must be 0) */
			size :14; /* Buffer size (must be EVEN!) */
	eaddr_t da;		  /* Destination address */
	eaddr_t sa;		  /* Source address */
	unsigned short length;	  /* User-defined length in IEEE 802.3 */
	union rfd *next;	  /* Virtual address of next RFD */
    } rfd;
    struct {
	dcache_line_t stat;	/* 1st Data cache line w/bits set by chip */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE)
	 */
	dcache_line_t cmd;	/* 2nd Data cache line w/bits set by CPU */
    } cacheline;
} rfd_t;

typedef unsigned short scb_int_t; /* SCB interrupt bits */
# define SCB_RNR	1	/* Receive Unit left the Ready State */
# define SCB_CNA	2	/* The Command Unit left the Active State */
# define SCB_FR		4	/* The Recv. Unit finished receiving a frame */
# define SCB_CX		8	/* The CU finished a cmd with int-bit on */

/*
 * Control Unit command codes:
 */
typedef unsigned short scb_cuc_t;
#define scb_cuc_nop			0
#define scb_cuc_start			1
#define scb_cuc_resume			2
#define scb_cuc_suspend			3
#define scb_cuc_abort			4
#define scb_cuc_throttle		5
#define scb_cuc_throttle_immediate	6
#define scb_cuc_reserved		7

/*
 * Receive Unit command codes:
 */
typedef unsigned short scb_ruc_t;
#define scb_ruc_nop			0
#define scb_ruc_start			1
#define scb_ruc_resume			2
#define scb_ruc_suspend			3
#define scb_ruc_abort			4
#define scb_ruc_reserved0		5
#define scb_ruc_reserved1		6
#define scb_ruc_reserved2		7

/*
 * SCB command word:
 */
typedef union {
    struct {
	scb_int_t ack		:4, /* Misc. command bits */
				:1;
	scb_cuc_t cuc		:3; /* Control Unit commands */
	unsigned short reset	:1; /* Reset */
	scb_ruc_t ruc		:3; /* Receive Unit commands */
	unsigned short		:4;
    } cmd;
    unsigned short all;
} scb_command_t;

/*
 * Control Unit status codes:
 */
typedef unsigned short scb_cus_t;
#define scb_cus_idle			0
#define scb_cus_suspended		1
#define scb_cus_active			2
#define scb_cus_not_used0		3
#define scb_cus_not_used1		4
#define scb_cus_not_used2		5
#define scb_cus_not_used3		6
#define scb_cus_not_used4		7

/*
 * Receive Unit status codes:
 */
typedef unsigned short scb_rus_t;
#define scb_rus_idle			 0
#define scb_rus_suspended		 1
#define scb_rus_no_resources		 2
#define scb_rus_not_used0		 3
#define scb_rus_ready			 4
#define scb_rus_not_used1		 5
#define scb_rus_not_used2		 6
#define scb_rus_not_used3		 7
#define scb_rus_no_rbd			 8
#define scb_rus_not_used4		 9
#define scb_rus_no_rbd_resources	10
#define scb_rus_not_used5		11
#define scb_rus_not_used6		12
#define scb_rus_not_used7		13
#define scb_rus_not_used8		14
#define scb_rus_not_used9		15

/*
 * SCB Status word:
 */
typedef struct {
    scb_int_t stat	 :4; /* Misc. status bits */
    unsigned short	 :1;
    scb_cus_t cus	 :3; /* Command Unit status */
    scb_rus_t rus	 :4; /* Receive Unit status */
    unsigned short throad:1, /* Bus throttle timers loaded */
			 :3;
} scb_status_t;

/*
 * System Control Block, padded to occupy 3 cache lines:
 */
typedef union {
    struct {
	unsigned char pad[DCACHE_LINE_SIZE-2];
	scb_status_t status; /* Status word */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	scb_command_t command; /* Command word */
	swap_t cbl_ptr;	 /* ->Command Block list (halfword-swapped) */
	swap_t rfa_ptr;	 /* ->Receive Frame Area (halfword-swapped) */
	swap_t crcerrs;	 /* # frames discarded w/CRC errors (h.w.-swapped) */
	swap_t alnerrs;	 /* # frames misaligned AND w/CRC errors (swapped) */
	swap_t rscerrs;	 /* # frames discarded only for lack of resources */
	swap_t ovrnerrs; /* # frames lost because local bus not available */
	swap_t rcvcdt;	 /* # collisions detected during frame reception */
	swap_t shrtfrm;	 /* # frames received shorter than min. length */
	unsigned short toff; /* Min. time to stay off the bus */
	unsigned short ton;  /* Max. time to stay on  the bus */
    } scb;
    struct {
	dcache_line_t stat;	/* 1st Data cache line w/bits set by chip */
	/*
	 * The PA-RISC data cache line boundary must fall here
	 * (cf. BIG FAT NOTICE).
	 */
	dcache_line_t cmd;	/* 2nd Data cache line w/bits set by CPU */
	dcache_line_t pad;	/* 3rd Data cache line with more bits */
    } cacheline;
} scb_t;

/*
 * Intermediate System Configuration Pointer:
 */
typedef struct {
    unsigned short	busy;
    unsigned short	dummy; /* (Unused) */
    swap_t		scb_ptr;
} iscp_t;


typedef unsigned short sysbus_mode_t;
#define mode_82586		0
#define mode_32_bit_segmented	1
#define mode_linear		2
#define mode_reserved		3
 
/*
 * Interrupt polarity:
 */
#define ACTIVE_HIGH	0
#define ACTIVE_LOW	1

typedef struct {
    unsigned short		:9, /* (Unused; must be 0) */
		   csw		:1,
		   int_polarity :1,
		   lock		:1,
		   trg		:1;
    sysbus_mode_t mode:2;
    unsigned short x:1;
} sysbus_t;

/*
 * System Configuration Pointer:
 */
typedef struct {
    unsigned short dummy0; /* (Unused; must be 0) */
    sysbus_t sysbus;
    unsigned short dummy1; /* (Unused; must be 0) */
    unsigned short dummy2; /* (Unused; must be 0) */
    swap_t iscp_ptr; /* ->ISCP (halfword-swapped) */
} scp_t;

/*
 * Self-test format:
 */
typedef struct {
    /*
     * This must be 16-byte aligned, and padded to a data cache line
     * boundary:
     */
    unsigned int signature; /* ROM signature */
    union {
	struct {
	    unsigned int selftest      :1,  /* Self-test result */
				       :9,
			 diagnosetest  :1,  /* Diagnose result */
			 bustimertest  :1,  /* Bus timer result */
			 registertest  :1,  /* Register result */
			 romcontenttest:1,  /* ROM content result */
			 zeros	       :18; /* Set to zero when done */
	} of;
	unsigned int all; /* Aggregate pass/fail */
    } result;
    unsigned char pad[DCACHE_LINE_SIZE-8]; /* Pad to data cache line */
} selftest_t;

/*
 * Consult the INTEL handbook for the layout of this, but it must be
 * 16-byte aligned and padded to a data cache line boundary.
 */
typedef unsigned char dump_t[CACHE_ALIGNED(306)]; /* Dump format */

typedef enum {port_reset, port_selftest, port_scp, port_dump} iport_t;

/*
 * Cache-aligned Ethernet controller data:
 */
typedef struct {
    nop_t nop;		     /* NOP CB (see BIG FAT NOTICE) */
    scb_t scb;		     /* System Control Block */
    rfd_t rfd_list[NRFDBUF]; /* Receive Frame Descriptor list */
    rbd_t rbd_list[NRBDBUF]; /* Receive Buffer Descriptor list */
    tcb_t tcb_list[NTCBBUF]; /* Transmit Command Block list */
    dump_t dump;	     /* 82596 Dump data */
} ecsc_t;

typedef enum {
    warn_link_ok,
    warn_transceiver_cable,
    warn_ethernet_open,
    warn_ethernet_short,
    warn_cb_timeout,
} warning_t;

#endif /* ec596reg_h */
