/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * HP_M68K high level SCSI disk driver header file for NCR 53C90 SCSI controller (DN2500)
 */

#ifndef _HP_SCSI_DSK_H_
#define _HP_SCSI_DSK_H_ 1

/*
 * The following piece of code was taken (translated) from /os/kins/win_scsi.pvt.pas
 */
typedef struct {
        u_int   reads;
        u_int   writes;
        u_short not_ready;
        u_short ctl_busy;
        u_short seek_error;
        u_short equip_check;
        u_short drv_time_out;
        u_short crcerr;
        u_short overrun;
} win_cnt_t;

/*
 * The following piece of code was taken (translated) from /os/kins/win_scsi.pvt.pas
 */

/* artificially limit lun to 0 to save data space, we've never seen a target with > 1 lun.      */
#define WIN_FIXED_LUN           0
#define WIN_MAX_LUN             WIN_FIXED_LUN
#define WIN_DESIRED_CORR_SPAN   11                      /* desired, first choice value for correction span.     */

typedef char    *win_csrs_ptr_t;

#define MAX_GROUP0_LEN          0xFF                    /* maximum transfer length expressable in len field     */
                                                        /* ^this value is assumed less than buffer of 1024!     */
#define MAX_GROUP1_LEN          0xFFFF                  /* maximum transfer length expressable in len field     */
#define WIN_TIMEOUT             (2 * 60 * 1000)         /* 2 minutes in milliseconds for general commands       */
#define WIN_SENSE_TIMEOUT       (2 * 60 * 1000)         /* 2 minutes in milliseconds for a request sense wait   */
#define WIN_START_STOP_TIMEOUT  (45 * 1000)             /* starts and stops can take up to 30 seconds, wait 45 to be safe.      */
#define WIN_FORMAT_TIMEOUT      (90 * 60 * 1000)        /* wait up to an hour and a half for the format to complete.    */

#define MAX_TRANSFER_TRIES      2                       /* current maximum  #of read/write retries.             */

/* better constants for fudging geometry values when a device will not          */
/* sense the rigid disk geometry page or the format page to help us out.        */
#define MAX_FLOP_B_PER_PVOL     (5 * 1024)
#define MIN_100MB_B_PER_PVOL    (90 * 1024)
#define MIN_200MB_B_PER_PVOL    (190 * 1024)
#define MAX_200MB_B_PER_PVOL    (220 * 1024)

#define MAX_FLOPPY_CAPACITY     (MAX_FLOP_B_PER_PVOL * 1024)
#define MIN_100MB_CAPACITY      (MIN_100MB_B_PER_PVOL * 1024)
#define MIN_200MB_CAPACITY      (MIN_200MB_B_PER_PVOL * 1024)
#define MAX_200MB_CAPACITY      (MAX_200MB_B_PER_PVOL * 1024)

#define FLOP_B_PER_TRK          8
#define FLOP_TRK_PER_CYL        2

#define F_100MB_B_PER_TRK       18
#define F_100MB_TRK_PER_CYL     6

#define F_200MB_B_PER_TRK       (F_100MB_B_PER_TRK * 2)
#define F_200MB_TRK_PER_CYL     (F_100MB_TRK_PER_CYL * 2)

#define OPTICAL_B_PER_TRK       17
#define OPTICAL_TRK_PER_CYL     1

#define WIN_PHYSICAL_SECTOR_SIZE        512

/* define some values for direct access mode sense/select pcf or page control field.    */
#define MODE_PCF_CURRENT        0x00
#define MODE_PCF_CHANGEABLE     0x01
#define MODE_PCF_DEFAULT        0x02
#define MODE_PCF_SAVED          0x03

/* define some values for direct access mode sense/select page_codes.   */
#define MODE_PC_RECOVERY        0x01
#define MODE_PC_CONNECT         0x02
#define MODE_PC_FORMAT          0x03
#define MODE_PC_GEOMETRY        0x04
#define MODE_PC_FLEX            0x05
#define MODE_PC_ALL             0x3F

typedef char    *scsi_buffer_ptr_t;

typedef struct {
        u_int   ps              : 1;    /* (0:7)        parameters saveable     */
        u_int   r               : 1;    /* (0:6)                                */
        u_int   page_code       : 6;    /* (0:5-0)      page code               */
        u_int   page_length     : 8;    /* (1)          page length             */
} page_hdr_t;

typedef page_hdr_t      *page_hdr_ptr_t;

typedef struct {
        page_hdr_t      hdr;                            /* (0-1)    page header, code and length        */
        u_int           num_cylinders           : 24;   /* (2-4)    number of cylinders                 */
        u_int           num_heads               : 8;    /* (5)      number of heads                     */
        u_int           precomp_start           : 24;   /* (6-8)    starting cylinder write  precomp    */
        u_int           write_current_start     : 24;   /* (9-B)    start cylinder reduced wrt current  */
        u_int           drive_step              : 16;   /* (C-D)    drive step rate                     */
        u_int           landing_zone            : 24;   /* (E-10)   landing zone cylinder               */
        u_int           res2                    : 8;    /* (11)     reserved                            */
        u_int           rotational_offset       : 8;    /* (12)     rotational offset                   */
        u_int           res3                    : 8;    /* (13)     reserved                            */
} geometry_page_data_t;
                                                                                                        
typedef geometry_page_data_t    *geometry_page_ptr_t;

typedef struct {
        page_hdr_t      hdr;                            /* (0-1)    page header, code and length        */
        u_int           tracks_per_zone         : 16;   /* (2-3)    number of tracks per zone           */
        u_int           alt_sects_per_zone      : 16;   /* (4-5)    alternate sectors per zone          */
        u_int           alt_tracks_per_zone     : 16;   /* (6-7)    alternate tracks per zone           */
        u_int           alt_tracks_per_unit     : 16;   /* (8-9)    alternate tracks per unit           */
        u_int           sectors_per_track       : 16;   /* (A-B)    sectors per track                   */
        u_int           data_bytes_per_sect     : 16;   /* (C-D)    data bytes per physical sector      */
        u_int           interleave              : 16;   /* (E-F)    interleave                          */
        u_int           track_skew              : 16;   /* (10-11)  track skew                          */
        u_int           cylinder_skew           : 16;   /* (12-13)  cylinder skew                       */
        u_int           ssec                    : 1;    /* (14:7)   soft sector formatting              */
        u_int           hsec                    : 1;    /* (14:6)   hard sector formatting              */
        u_int           rmb                     : 1;    /* (14:5)   removable                           */
        u_int           surf                    : 1;    /* (14:4)   surf, allocate w/in cyl first       */
        u_int           res2                    : 4;    /* (14:3-0) reserved                            */ 
        u_int           res3                    : 16;   /* (15-16)  reserved                            */
} format_page_data_t;

typedef format_page_data_t      *format_page_ptr_t;

typedef struct {
        page_hdr_t      hdr;                            /* (0-1)    page header, code and length        */
        u_int           transfer_rate           : 16;   /* (2-3)                                        */
        u_int           num_heads               : 8;    /* (4)                                          */
        u_int           sectors_per_track       : 8;    /* (5)                                          */
        u_int           data_bytes_per_sect     : 16;   /* (6-7)                                        */
        u_int           num_cylinders           : 16;   /* (8-9)                                        */
        u_int           precomp_start           : 16;   /* (10-11)  starting cylinder write  precomp    */
        u_int           write_current_start     : 16;   /* (12-13)  start cylinder reduced wrt current  */
        u_int           drive_step_rate         : 16;   /* (14-15)                                      */
        u_int           drive_step_width        : 8;    /* (16)                                         */
        u_int           head_settle_delay       : 8;    /* (17)                                         */
        u_int           motor_on_delay          : 8;    /* (18)                                         */
        u_int           motor_off_delay         : 8;    /* (19)                                         */
        u_int           TRDY                    : 1;    /* (20:7)                                       */
        u_int           reserved                : 7;    /* (20:6-0)                                     */
        u_int           head_load_delay         : 8;    /* (21)                                         */ 
        u_int           start_sector_0          : 8;    /* (22)     starting sector number, side 0      */
        u_int           start_sector_1          : 8;    /* (23)     starting sector number, side 1      */
} flex_page_data_t;

typedef flex_page_data_t        *flex_page_ptr_t;

typedef struct {
        page_hdr_t      hdr;                            /* (0-1)    page header, code and length        */
        u_int           awre            : 1;            /* (2:7)                                        */
        u_int           arre            : 1;            /* (2:6)                                        */
        u_int           tb              : 1;            /* (2:5)                                        */
        u_int           rc              : 1;            /* (2:4)                                        */
        u_int           eer             : 1;            /* (2:3)                                        */
        u_int           per             : 1;            /* (2:2)                                        */
        u_int           dte             : 1;            /* (2:1)                                        */
        u_int           dcr             : 1;            /* (2:0)                                        */
        u_int           retry           : 8;            /* (3)      retry count                         */
        u_int           correction_span : 8;            /* (4)      correction span                     */
        u_int           head_offset     : 8;            /* (5)      head offset                         */
        u_int           data_offset     : 8;            /* (6)      data strobe offest                  */
        u_int           recover_limit   : 8;            /* (7)      recovery time limit                 */
} recovery_page_data_t;                                                                                 

typedef recovery_page_data_t    *recovery_page_ptr_t;

typedef struct {
        page_hdr_t      hdr;                            /* (0-1)    page header, code and length        */
        u_char          buf_full_ratio;                 /* (2)      buffer full ratio                   */
        u_char          buf_empty_ratio;                /* (3)      buffer empty ratio                  */
        u_short         bus_inactive_limit;             /* (4-5)    bus inactivity limit                */
        u_short         disconnect_limit;               /* (6-7)    disconnect time limit               */
        u_short         connect_limit;                  /* (8-9)    connect time limit                  */
        u_short         reserved;                       /* (10-11)  reserved                            */
} connect_page_data_t;

typedef connect_page_data_t     *connect_page_ptr_t;

typedef struct {
        u_int   density_code    : 8;                    /* (0)      density code                        */
        u_int   n_blocks        : 24;                   /* (1-3)    number of blocks                    */
        u_int   rsv1            : 8;                    /* (4)      reserved                            */
        u_int   block_len       : 24;                   /* (5-7)    block length                        */
} block_descriptor_t;

typedef block_descriptor_t      *block_desc_ptr_t;

typedef struct {
        u_char  rsv1;                                   /* (0)                                          */
        u_char  medium;                                 /* (1)      medium type                         */
        u_char  rsv2;                                   /* (2)                                          */
        u_char  bd_len;                                 /* (3) block descriptor length, 0 or 8          */
} mode_select_hdr_t;

typedef mode_select_hdr_t       mode_sel_hdr_ptr_t;

typedef union {
        scsi_nonext_sense_data_t        nonext;
        scsi_ext_sense_data_t           ext;
} utility_sense_data_t;

/* the following MUST be the largest data area needed for any group2 command used.      */
#define WIN_MAX_USED_GROUP1_BYTES       sizeof(scsi_capacity_t)

/* we will need a buffer area large enough to accomodate all group 0 commands           */
/* or the largest group 1 command that is actually used by the driver.                  */
#define WIN_MAX_UTIL_BYTES              (MAX(WIN_MAX_USED_GROUP1_BYTES, MAX_GROUP0_LEN))

typedef struct {
        u_int   cmd             : 8;                    /* (0)      command: 8 bits                     */
        u_int   lun             : 3;                    /* (1:7-5)  logical unit: 3 bits                */
        u_int   log_block_addr  : 21;                   /* (1:4-0)  logical block address, if required  */
                                                        /* (2-3)                                        */
        u_int   len             : 8;                    /* (4)      transfer length                     */
        u_int   control_byte    : 8;                    /* (5)      control byte                        */
} gen_group0_cdb_t;

typedef struct {
        u_int   cmd             : 8;                    /* (0)      command: 8 bits                     */
        u_int   lun             : 3;                    /* (1:7-5)  logical unit: 3 bits                */
        u_int   res1            : 4;                    /* (1:4-1)  reserved                            */
        u_int   reladdr         : 1;                    /* (1:0)    relative address                    */
        u_int   log_block_addr;                         /* (2-5)    logical block address, if required  */
        u_int   res2            : 8;                    /* (6)      reserved                            */
        u_int   len             : 16;                   /* (7-8)    transfer length                     */
        u_int   control_byte    : 8;                    /* (9)      control byte                        */
} gen_group1_cdb_t;

typedef struct {
        u_int   cmd             : 8;                    /* (0)      command: 8 bits                     */
        u_int   lun             : 3;                    /* (1:7-5)  logical unit: 3 bits                */
        u_int   pf              : 1;                    /* (1:4)    page format, 1 if scsi 2 format     */
        u_int   dbd             : 1;                    /* (1:3)    disable block descriptors on sense, 0 on select     */
        u_int   res1            : 2;                    /* (1:2-1)  reserved                            */
        u_int   sp              : 1;                    /* (1:0)    save pareters on select             */
        u_int   pcf             : 2;                    /* (2:7-6)  page control field on sense, 0 on select    */
        u_int   page_code       : 6;                    /* (2:5-0)  page code on sense, 0 on select     */
        u_int   res2            : 8;                    /* (3)      reserved                            */
        u_int   len             : 8;                    /* (4)      transfer length                     */
        u_int   control_byte    : 8;                    /* (5)      control byte                        */
} gen_mode_0_cdb_t;

typedef struct {
        u_int   cmd             : 8;                    /* (0)      command: 8 bits                     */
        u_int   lun             : 3;                    /* (1:7-5)  logical unit: 3 bits                */
        u_int   res1            : 4;                    /* (1:4-1)  reserved                            */
        u_int   reladdr         : 1;                    /* (1:0)    relative address                    */
        u_int   log_block_addr;                         /* (2-5)    logical block address, if required  */
        u_int   res             : 16;                   /* (6-7)    reserved                            */
        u_int   res2            : 7;                    /* (8:7-1)  reserved                            */
        u_int   pmi             : 1;                    /* (8:0)    partial medium detector             */
        u_int   control_byte    : 8;                    /* (9)      control byte                        */
} read_cap_cdb_t;

typedef struct {
        u_int   cmd             : 8;                    /* (0)      command: 8 bits                     */
        u_int   lun             : 3;                    /* (1:7-5)  logical unit: 3 bits                */
        u_int   fmtdata         : 1;                    /* (1:4)    def. block data supplied            */
        u_int   cmplist         : 1;                    /* (1:3)    def. list is complete               */
        u_int   deflist         : 3;                    /* (1:2-0)  supplies additional                 */
        u_int   vendor_unique_1 : 8;                    /* (2)                                          */
        u_int   interleave      : 16;                   /* (3-4)    interleave                          */
        u_int   control_byte    : 8;                    /* (5)      control byte                        */
} format_cdb_t;

typedef struct {
        u_int   cmd             : 8;                    /* (0)      command: 8 bits                     */
        u_int   lun             : 3;                    /* (1:7-5)  logical unit: 3 bits                */
        u_int   res1            : 4;                    /* (1:4-1)  reserved                            */
        u_int   immed           : 1;                    /* (1:0)    immed, return upon initiation       */
        u_int   res2            : 16;                   /* (2-3)    reserved                            */
        u_int   res3            : 6;                    /* (4:7-2)  reserved                            */
        u_int   loaj            : 1;                    /* (4:1)    no action, load/eject               */
        u_int   start           : 1;                    /* (4:0)    start                               */
        u_int   control_byte    : 8;                    /* (5)      control byte                        */
} start_stop_cdb_t;

typedef struct {
        u_int   cmd             : 8;                    /* (0)      command: 8 bits                     */
        u_int   lun             : 3;                    /* (1:7-5)  logical unit: 3 bits                */
        u_int   res1            : 5;                    /* (1:4-0)  reserved                            */
        u_int   res2            : 16;                   /* (2-3)    reserved                            */
        u_int   res3            : 7;                    /* (4:7-1)  reserved                            */
        u_int   prevent         : 1;                    /* (4:0)    prevent media removal               */
        u_int   res4            : 8;                    /* (5)      reserved                            */
} media_ctl_cdb_t;

typedef union {
        scsi_cdb_t              formal;
        gen_group0_cdb_t        gen0;
        gen_group1_cdb_t        gen1;
        gen_mode_0_cdb_t        mode;
        start_stop_cdb_t        start_stop;
        media_ctl_cdb_t         media_ctl;
} utility_cdb_t;

typedef enum {
        apollo_100,  
        apollo_200,  
        unknown
} apollo_disk_t;

typedef struct {
        u_int                   read_cap_blk_len;       /* logical sector size from read capacity command.      */
        u_int                   log_blk_len;            /* logical sector size from block descriptor.           */
        u_int                   phys_blk_len;           /* physical sector size from format page.               */
        u_int                   sectors_per_k;          /* number of transfers needed to get 1K                 */
        scsi_vendor_id_t        vendor_id;              /* saved vendor identification from inquiry             */
        scsi_product_id_t       product_id;             /* saved product identification from inquiry            */
        scsi_prod_rev_t         drive_rev_level;        /* saved product revision level                         */
        win_cnt_t               win_cnt;
        scsi_sense_info_t       err_sense_info;         /* information bytes from extended sense of error.      */
        scsi_ansi_rev_t         scsi_rev;
        u_short                 pad;
        u_int                   rmb             : 1;
        u_int                   err_sense_valid : 1;    /* true when err_sense_info is valid.                   */
        u_int                   recovered_error : 1;    /* true on any flavor of recovered error.               */
        u_int                   log_len_valid   : 1;    /* logical sector size was available on a mode sense of all pages.      */
        u_int                   phys_len_valid  : 1;    /* physical sector size was available and from format page.     */
        u_int                   not_valid       : 1;    /* door has been opened and not revalidated.            */
} ucb_t;

typedef ucb_t   ucb_array_t[WIN_MAX_LUN + 1];
typedef ucb_t   *ucb_ptr_t;

typedef struct {
        char                    utility_buffer_area[WIN_MAX_UTIL_BYTES];
        scsi_buffer_ptr_t       utility_bufferp;
        scsi_handle_t           handle;         
        u_int                   max_xfer;               /* maximum number of blocks we can transfer at once.    */
        ucb_array_t             ucbs;           
        u_int                   acquired_ok : 1;    
        u_int                   is_a_floppy : 1;   
} win_ctlr_control_blk_t;

typedef win_ctlr_control_blk_t  ccb_array_t[SCSI_MAX_TARGETS];

typedef scsi_capacity_t         *scsi_capacity_ptr_t;
typedef scsi_inquiry_data_t     *scsi_inquiry_ptr_t;
typedef utility_sense_data_t    *scsi_sense_data_ptr_t;

#endif /* _HP_SCSI_DSK_H_ */

